@@ -19,7 +19,11 @@ from bypy.freeze import (
 from bypy.utils import get_dll_path, mkdtemp, py_compile, walk
 
 j = os.path.join
-arch = 'x86_64' if is64bit else 'i686'
+machine = (os.uname()[4] or '').lower()
+if machine.startswith('arm64') or machine.startswith('aarch64'):
+    arch = 'arm64'
+else:
+    arch = 'x86_64' if is64bit else 'i686'
 self_dir = os.path.dirname(os.path.abspath(__file__))
 py_ver = '.'.join(map(str, python_major_minor_version()))
 iv = globals()['init_env']
@@ -29,7 +33,7 @@ kitty_constants = iv['kitty_constants']
 def binary_includes():
     return tuple(map(get_dll_path, (
             'expat', 'sqlite3', 'ffi', 'z', 'lzma', 'png16', 'lcms2', 'crypt',
-            'iconv', 'pcre', 'graphite2', 'glib-2.0', 'freetype',
+            'iconv', 'pcre', 'graphite2', 'glib-2.0', 'freetype', 'rsync',
             'harfbuzz', 'xkbcommon', 'xkbcommon-x11',
             'ncursesw', 'readline', 'brotlicommon', 'brotlienc', 'brotlidec'
         ))) + (
@@ -115,8 +119,7 @@ def copy_python(env):
             shutil.copy2(y, env.py_dir)
 
     srcdir = j(srcdir, 'site-packages')
-    site_packages_dir = j(env.py_dir, 'site-packages')
-    import_site_packages(srcdir, site_packages_dir)
+    import_site_packages(srcdir, env.py_dir)
 
     pdir = os.path.join(env.lib_dir, 'kitty-extensions')
     os.makedirs(pdir, exist_ok=True)
