@@ -454,12 +454,30 @@ void* _glfwLoadLocalVulkanLoaderNS(void)
 //////////////////////////////////////////////////////////////////////////
 
 static bool
-is_ctrl_tab(NSEvent *event, NSEventModifierFlags modifierFlags) {
+is_modified_tab(NSEvent *event, NSEventModifierFlags modifierFlags) {
+    switch ((NSUInteger)modifierFlags) {
+        // No need to handle shift+tab, [shift]+option+tab
+        case NSEventModifierFlagShift:
+        case NSEventModifierFlagOption:
+        case (NSEventModifierFlagShift | NSEventModifierFlagOption):
+        // Do not intercept cmd+tab, shift+cmd+tab
+        case NSEventModifierFlagCommand:
+        case (NSEventModifierFlagShift | NSEventModifierFlagCommand):
+            return false;
+        default:
+            break;
+    }
+    // ctrl+whatever+tab, option+cmd+tab
     if (
-            (modifierFlags == NSEventModifierFlagControl &&
-                [event.charactersIgnoringModifiers isEqualToString:@"\t"]) ||
-            (modifierFlags == (NSEventModifierFlagControl | NSEventModifierFlagShift) &&
-                [event.charactersIgnoringModifiers isEqualToString:@"\x19"])
+            (
+                (modifierFlags & NSEventModifierFlagControl) ||
+                    modifierFlags == (NSEventModifierFlagOption | NSEventModifierFlagCommand)
+            ) && [event.charactersIgnoringModifiers isEqualToString:@"\t"]
+       ) return true;
+    // shift+whatever+tab
+    if (
+            (modifierFlags & NSEventModifierFlagShift) &&
+                [event.charactersIgnoringModifiers isEqualToString:@"\x19"]
        ) return true;
     return false;
 }
@@ -471,6 +489,38 @@ is_cmd_period(NSEvent *event, NSEventModifierFlags modifierFlags) {
     return false;
 }
 
+static bool
+is_modified_special_key(NSEvent *event, NSEventModifierFlags modifierFlags) {
+    // really one should use [[NSUserDefaults standardUserDefaults] valueForDefaultsDomain:@"com.apple.symbolichotkeys" key:@"AppleSymbolicHotKeys"]
+    // to get the list of global shortcuts and pass through the important ones,
+    // see https://stackoverflow.com/questions/21878482/what-do-the-parameter-values-in-applesymbolichotkeys-plist-dict-represent
+    // however given that in order to know which integers are which actions in that dict one needs reverse engineering
+    // see https://stackoverflow.com/questions/866056/how-do-i-programmatically-get-the-shortcut-keys-reserved-by-mac-os-x
+    // it's too much effort.
+    if ([event.charactersIgnoringModifiers length] != 1) return false;
+    const unichar ch = [event.charactersIgnoringModifiers characterAtIndex:0];
+    if (modifierFlags == (NSEventModifierFlagControl | NSEventModifierFlagShift)) {
+        switch (ch) {
+            case 0x1b:  // Esc
+            case NSF1FunctionKey: case NSF2FunctionKey: case NSF3FunctionKey: case NSF4FunctionKey:
+            case NSF5FunctionKey: case NSF6FunctionKey: case NSF7FunctionKey: case NSF8FunctionKey:
+            case NSF9FunctionKey: case NSF10FunctionKey: case NSF11FunctionKey: case NSF12FunctionKey:
+            case NSF13FunctionKey: case NSF14FunctionKey: case NSF15FunctionKey: case NSF16FunctionKey:
+            case NSF17FunctionKey: case NSF18FunctionKey: case NSF19FunctionKey:
+                return true;
+        }
+    }
+    switch (ch) {
+        case 0x1b:  // Esc
+            if (modifierFlags & (NSEventModifierFlagCommand | NSEventModifierFlagControl)) return true;
+            break;
+        case NSHelpFunctionKey:  // For some reason keyboards with an insert key have it mapped to help
+            if (!modifierFlags || modifierFlags == NSEventModifierFlagShift) return true;
+            break;
+    }
+    return false;
+}
+
 GLFWAPI GLFWapplicationshouldhandlereopenfun glfwSetApplicationShouldHandleReopen(GLFWapplicationshouldhandlereopenfun callback) {
     GLFWapplicationshouldhandlereopenfun previous = handle_reopen_callback;
     handle_reopen_callback = callback;
@@ -510,10 +560,11 @@ int _glfwPlatformInit(void)
 
     NSEvent* (^keydown_block)(NSEvent*) = ^ NSEvent* (NSEvent* event)
     {
-        NSEventModifierFlags modifierFlags = [event modifierFlags] & NSEventModifierFlagDeviceIndependentFlagsMask;
-        if (is_ctrl_tab(event, modifierFlags) || is_cmd_period(event, modifierFlags)) {
-            // Cocoa swallows Ctrl+Tab to cycle between views
+        NSEventModifierFlags modifierFlags = [event modifierFlags] & (NSEventModifierFlagShift | NSEventModifierFlagOption | NSEventModifierFlagCommand | NSEventModifierFlagControl);
+        if (is_modified_special_key(event, modifierFlags) || is_modified_tab(event, modifierFlags) || is_cmd_period(event, modifierFlags)) {
+            // Cocoa swallows various key presses, so route them explicitly
             [[NSApp keyWindow].contentView keyDown:event];
+            return nil;
         }
 
         return event;
@@ -521,16 +572,17 @@ int _glfwPlatformInit(void)
 
     NSEvent* (^keyup_block)(NSEvent*) = ^ NSEvent* (NSEvent* event)
     {
-        NSEventModifierFlags modifierFlags = [event modifierFlags] & NSEventModifierFlagDeviceIndependentFlagsMask;
+        NSEventModifierFlags modifierFlags = [event modifierFlags] & (NSEventModifierFlagShift | NSEventModifierFlagOption | NSEventModifierFlagCommand | NSEventModifierFlagControl);
         if (modifierFlags & NSEventModifierFlagCommand) {
-            // From http://cocoadev.com/index.pl?GameKeyboardHandlingAlmost
+            // From https://cocoadev.github.io/GameKeyboardHandlingAlmost/
             // This works around an AppKit bug, where key up events while holding
             // down the command key don't get sent to the key window.
             [[NSApp keyWindow] sendEvent:event];
         }
-        if (is_ctrl_tab(event, modifierFlags) || is_cmd_period(event, modifierFlags)) {
-            // Cocoa swallows Ctrl+Tab to cycle between views
+        if (is_modified_special_key(event, modifierFlags) || is_modified_tab(event, modifierFlags) || is_cmd_period(event, modifierFlags)) {
+            // Cocoa swallows various key presses, so route them explicitly
             [[NSApp keyWindow].contentView keyUp:event];
+            return nil;
         }
 
         return event;
