@@ -6,7 +6,7 @@ import json
 import os
 import re
 import sys
-from typing import Callable, Dict, List, Optional, Tuple
+from typing import Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple
 
 _plat = sys.platform.lower()
 is_linux = 'linux' in _plat
@@ -14,9 +14,23 @@ is_openbsd = 'openbsd' in _plat
 base = os.path.dirname(os.path.abspath(__file__))
 
 
+class CompileKey(NamedTuple):
+    src: str
+    dest: str
+
+
+class Command(NamedTuple):
+    desc: str
+    cmd: Sequence[str]
+    is_newer_func: Callable[[], bool]
+    on_success: Callable[[], None] = lambda: None
+    key: Optional[CompileKey] = None
+    keyfile: Optional[str] = None
+
+
 class Env:
 
-    cc: str = ''
+    cc: List[str] = []
     cppflags: List[str] = []
     cflags: List[str] = []
     ldflags: List[str] = []
@@ -33,11 +47,12 @@ class Env:
     wayland_protocols: Tuple[str, ...] = ()
 
     def __init__(
-        self, cc: str = '', cppflags: List[str] = [], cflags: List[str] = [], ldflags: List[str] = [],
+        self, cc: List[str] = [], cppflags: List[str] = [], cflags: List[str] = [], ldflags: List[str] = [],
         library_paths: Dict[str, List[str]] = {}, ldpaths: Optional[List[str]] = None, ccver: Tuple[int, int] = (0, 0)
     ):
         self.cc, self.cppflags, self.cflags, self.ldflags, self.library_paths = cc, cppflags, cflags, ldflags, library_paths
-        self.ldpaths, self.ccver = [] if ldpaths is None else ldpaths, ccver
+        self.ldpaths = ldpaths or []
+        self.ccver = ccver
 
     def copy(self) -> 'Env':
         ans = Env(self.cc, list(self.cppflags), list(self.cflags), list(self.ldflags), dict(self.library_paths), list(self.ldpaths), self.ccver)
@@ -55,7 +70,14 @@ def wayland_protocol_file_name(base: str, ext: str = 'c') -> str:
     return 'wayland-{}-client-protocol.{}'.format(base, ext)
 
 
-def init_env(env: Env, pkg_config: Callable, pkg_version: Callable, at_least_version: Callable, test_compile: Callable, module: str = 'x11') -> Env:
+def init_env(
+    env: Env,
+    pkg_config: Callable[..., List[str]],
+    pkg_version: Callable[[str], Tuple[int, int]],
+    at_least_version: Callable[..., None],
+    test_compile: Callable[..., bool],
+    module: str = 'x11'
+) -> Env:
     ans = env.copy()
     ans.cflags.append('-fPIC')
     ans.cppflags.append('-D_GLFW_' + module.upper())
@@ -117,7 +139,13 @@ def init_env(env: Env, pkg_config: Callable, pkg_version: Callable, at_least_ver
     return ans
 
 
-def build_wayland_protocols(env: Env, Command: Callable, parallel_run: Callable, emphasis: Callable, newer: Callable, dest_dir: str) -> None:
+def build_wayland_protocols(
+    env: Env,
+    parallel_run: Callable[[List[Command]], None],
+    emphasis: Callable[[str], str],
+    newer: Callable[..., bool],
+    dest_dir: str
+) -> None:
     items = []
     for protocol in env.wayland_protocols:
         src = os.path.join(env.wayland_packagedir, protocol)
@@ -130,8 +158,7 @@ def build_wayland_protocols(env: Env, Command: Callable, parallel_run: Callable,
                 q = 'client-header' if ext == 'h' else env.wayland_scanner_code
                 items.append(Command(
                     'Generating {} ...'.format(emphasis(os.path.basename(dest))),
-                    [env.wayland_scanner, q, src, dest],
-                    lambda: True, None, None, None))
+                    [env.wayland_scanner, q, src, dest], lambda: True))
     if items:
         parallel_run(items)
 
