@@ -49,6 +49,70 @@ enum Capabilities {
     IBUS_CAP_SURROUNDING_TEXT   = 1 << 5
 };
 
+typedef enum
+{
+    IBUS_SHIFT_MASK    = 1 << 0,
+    IBUS_LOCK_MASK     = 1 << 1,
+    IBUS_CONTROL_MASK  = 1 << 2,
+    IBUS_MOD1_MASK     = 1 << 3,
+    IBUS_MOD2_MASK     = 1 << 4,
+    IBUS_MOD3_MASK     = 1 << 5,
+    IBUS_MOD4_MASK     = 1 << 6,
+    IBUS_MOD5_MASK     = 1 << 7,
+    IBUS_BUTTON1_MASK  = 1 << 8,
+    IBUS_BUTTON2_MASK  = 1 << 9,
+    IBUS_BUTTON3_MASK  = 1 << 10,
+    IBUS_BUTTON4_MASK  = 1 << 11,
+    IBUS_BUTTON5_MASK  = 1 << 12,
+
+    /* The next few modifiers are used by XKB, so we skip to the end.
+     * Bits 15 - 23 are currently unused. Bit 29 is used internally.
+     */
+
+    /* ibus mask */
+    IBUS_HANDLED_MASK  = 1 << 24,
+    IBUS_FORWARD_MASK  = 1 << 25,
+    IBUS_IGNORED_MASK  = IBUS_FORWARD_MASK,
+
+    IBUS_SUPER_MASK    = 1 << 26,
+    IBUS_HYPER_MASK    = 1 << 27,
+    IBUS_META_MASK     = 1 << 28,
+
+    IBUS_RELEASE_MASK  = 1 << 30,
+
+    IBUS_MODIFIER_MASK = 0x5f001fff
+} IBusModifierType;
+
+
+static uint32_t
+ibus_key_state_from_glfw(unsigned int glfw_modifiers, int action) {
+    uint32_t ans = action == GLFW_RELEASE ? IBUS_RELEASE_MASK : 0;
+#define M(g, i) if(glfw_modifiers & GLFW_MOD_##g) ans |= i
+    M(SHIFT, IBUS_SHIFT_MASK);
+    M(CAPS_LOCK, IBUS_LOCK_MASK);
+    M(CONTROL, IBUS_CONTROL_MASK);
+    M(ALT, IBUS_MOD1_MASK);
+    M(NUM_LOCK, IBUS_MOD2_MASK);
+    M(SUPER, IBUS_MOD4_MASK);
+    /* To do: figure out how to get super/hyper/meta */
+#undef M
+    return ans;
+}
+
+static unsigned int
+glfw_modifiers_from_ibus_state(uint32_t ibus_key_state) {
+    unsigned int ans = 0;
+#define M(g, i) if(ibus_key_state & i) ans |= GLFW_MOD_##g
+    M(SHIFT, IBUS_SHIFT_MASK);
+    M(CAPS_LOCK, IBUS_LOCK_MASK);
+    M(CONTROL, IBUS_CONTROL_MASK);
+    M(ALT, IBUS_MOD1_MASK);
+    M(NUM_LOCK, IBUS_MOD2_MASK);
+    M(SUPER, IBUS_MOD4_MASK);
+    /* To do: figure out how to get super/hyper/meta */
+#undef M
+    return ans;
+}
 
 static bool
 test_env_var(const char *name, const char *val) {
@@ -106,6 +170,29 @@ get_ibus_text_from_message(DBusMessage *msg) {
     return text;
 }
 
+static void
+handle_ibus_forward_key_event(DBusMessage *msg) {
+    uint32_t keysym, keycode, state;
+    DBusMessageIter iter;
+    dbus_message_iter_init(msg, &iter);
+
+    if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_UINT32) return;
+    dbus_message_iter_get_basic(&iter, &keysym);
+    dbus_message_iter_next(&iter);
+
+    if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_UINT32) return;
+    dbus_message_iter_get_basic(&iter, &keycode);
+    dbus_message_iter_next(&iter);
+
+    if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_UINT32) return;
+    dbus_message_iter_get_basic(&iter, &state);
+
+    int mods = glfw_modifiers_from_ibus_state(state);
+
+    debug("IBUS: ForwardKeyEvent: keysym=%x, keycode=%x, state=%x, glfw_mods=%x\n", keysym, keycode, state, mods);
+    glfw_xkb_forwarded_key_from_ime(keysym, mods);
+}
+
 static void
 send_text(const char *text, GLFWIMEState ime_state) {
     _GLFWwindow *w = _glfwFocusedWindow();
@@ -126,7 +213,7 @@ message_handler(DBusConnection *conn UNUSED, DBusMessage *msg, void *user_data)
     _GLFWIBUSData *ibus = (_GLFWIBUSData*)user_data;
     (void)ibus;
     const char *text;
-    switch(glfw_dbus_match_signal(msg, IBUS_INPUT_INTERFACE, "CommitText", "UpdatePreeditText", "HidePreeditText", "ShowPreeditText", NULL)) {
+    switch(glfw_dbus_match_signal(msg, IBUS_INPUT_INTERFACE, "CommitText", "UpdatePreeditText", "HidePreeditText", "ShowPreeditText", "ForwardKeyEvent", NULL)) {
         case 0:
             text = get_ibus_text_from_message(msg);
             debug("IBUS: CommitText: '%s'\n", text ? text : "(nil)");
@@ -134,8 +221,8 @@ message_handler(DBusConnection *conn UNUSED, DBusMessage *msg, void *user_data)
             break;
         case 1:
             text = get_ibus_text_from_message(msg);
-            send_text(text, GLFW_IME_PREEDIT_CHANGED);
             debug("IBUS: UpdatePreeditText: '%s'\n", text ? text : "(nil)");
+            send_text(text, GLFW_IME_PREEDIT_CHANGED);
             break;
         case 2:
             debug("IBUS: HidePreeditText\n");
@@ -143,10 +230,43 @@ message_handler(DBusConnection *conn UNUSED, DBusMessage *msg, void *user_data)
         case 3:
             debug("IBUS: ShowPreeditText\n");
             break;
+        case 4:
+            handle_ibus_forward_key_event(msg);
+            break;
     }
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
+static DBusHandlerResult
+ibus_on_owner_change(DBusConnection* conn UNUSED, DBusMessage* msg, void* user_data) {
+    if (dbus_message_is_signal(msg, "org.freedesktop.DBus", "NameOwnerChanged")) {
+        const char* name;
+        const char* old_owner;
+        const char* new_owner;
+
+        if (!dbus_message_get_args(msg, NULL,
+            DBUS_TYPE_STRING, &name,
+            DBUS_TYPE_STRING, &old_owner,
+            DBUS_TYPE_STRING, &new_owner,
+            DBUS_TYPE_INVALID
+        )) {
+            return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+        }
+
+        if (strcmp(name, "org.freedesktop.IBus") != 0) {
+            return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+        }
+
+        _GLFWIBUSData* ibus = (_GLFWIBUSData*) user_data;
+        ibus->name_owner_changed = true;
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+
+    }
+
+    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
 static const char*
 get_ibus_address_file_name(void) {
     const char *addr;
@@ -243,6 +363,8 @@ input_context_created(DBusMessage *msg, const char* errmsg, void *data) {
     free((void*)ibus->input_ctx_path);
     ibus->input_ctx_path = _glfw_strdup(path);
     if (!ibus->input_ctx_path) return;
+    dbus_bus_add_match(ibus->conn, "type='signal',interface='org.freedesktop.DBus', member='NameOwnerChanged'", NULL);
+    dbus_connection_add_filter(ibus->conn, ibus_on_owner_change, ibus, free);
     dbus_bus_add_match(ibus->conn, "type='signal',interface='org.freedesktop.IBus.InputContext'", NULL);
     DBusObjectPathVTable ibus_vtable = {.message_function = message_handler};
     dbus_connection_try_register_object_path(ibus->conn, ibus->input_ctx_path, &ibus_vtable, ibus, NULL);
@@ -285,6 +407,7 @@ glfw_connect_to_ibus(_GLFWIBUSData *ibus) {
     if (ibus->inited) return;
     if (!test_env_var("GLFW_IM_MODULE", "ibus")) return;
     ibus->inited = true;
+    ibus->name_owner_changed = false;
     setup_connection(ibus);
 }
 
@@ -306,13 +429,12 @@ glfw_ibus_terminate(_GLFWIBUSData *ibus) {
 static bool
 check_connection(_GLFWIBUSData *ibus) {
     if (!ibus->inited) return false;
-    if (ibus->conn && dbus_connection_get_is_connected(ibus->conn)) {
-        return ibus->ok;
-    }
+    if (ibus->conn && dbus_connection_get_is_connected(ibus->conn) && !ibus->name_owner_changed) return ibus->ok;
     struct stat s;
+    ibus->name_owner_changed = false;
     if (stat(ibus->address_file_name, &s) != 0 || s.st_mtime != ibus->address_file_mtime) {
         if (!read_ibus_address(ibus)) return false;
-        setup_connection(ibus);
+        return setup_connection(ibus);
     }
     return false;
 }
@@ -344,56 +466,6 @@ glfw_ibus_set_cursor_geometry(_GLFWIBUSData *ibus, int x, int y, int w, int h) {
     }
 }
 
-typedef enum
-{
-    IBUS_SHIFT_MASK    = 1 << 0,
-    IBUS_LOCK_MASK     = 1 << 1,
-    IBUS_CONTROL_MASK  = 1 << 2,
-    IBUS_MOD1_MASK     = 1 << 3,
-    IBUS_MOD2_MASK     = 1 << 4,
-    IBUS_MOD3_MASK     = 1 << 5,
-    IBUS_MOD4_MASK     = 1 << 6,
-    IBUS_MOD5_MASK     = 1 << 7,
-    IBUS_BUTTON1_MASK  = 1 << 8,
-    IBUS_BUTTON2_MASK  = 1 << 9,
-    IBUS_BUTTON3_MASK  = 1 << 10,
-    IBUS_BUTTON4_MASK  = 1 << 11,
-    IBUS_BUTTON5_MASK  = 1 << 12,
-
-    /* The next few modifiers are used by XKB, so we skip to the end.
-     * Bits 15 - 23 are currently unused. Bit 29 is used internally.
-     */
-
-    /* ibus mask */
-    IBUS_HANDLED_MASK  = 1 << 24,
-    IBUS_FORWARD_MASK  = 1 << 25,
-    IBUS_IGNORED_MASK  = IBUS_FORWARD_MASK,
-
-    IBUS_SUPER_MASK    = 1 << 26,
-    IBUS_HYPER_MASK    = 1 << 27,
-    IBUS_META_MASK     = 1 << 28,
-
-    IBUS_RELEASE_MASK  = 1 << 30,
-
-    IBUS_MODIFIER_MASK = 0x5f001fff
-} IBusModifierType;
-
-
-static uint32_t
-ibus_key_state(unsigned int glfw_modifiers, int action) {
-    uint32_t ans = action == GLFW_RELEASE ? IBUS_RELEASE_MASK : 0;
-#define M(g, i) if(glfw_modifiers & GLFW_MOD_##g) ans |= i
-    M(SHIFT, IBUS_SHIFT_MASK);
-    M(CAPS_LOCK, IBUS_LOCK_MASK);
-    M(CONTROL, IBUS_CONTROL_MASK);
-    M(ALT, IBUS_MOD1_MASK);
-    M(NUM_LOCK, IBUS_MOD2_MASK);
-    M(SUPER, IBUS_MOD4_MASK);
-    /* To do: figure out how to get super/hyper/meta */
-#undef M
-    return ans;
-}
-
 void
 key_event_processed(DBusMessage *msg, const char* errmsg, void *data) {
     uint32_t handled = 0;
@@ -422,7 +494,7 @@ ibus_process_key(const _GLFWIBUSKeyEvent *ev_, _GLFWIBUSData *ibus) {
     // Put the key's text in a field IN the structure, for proper serialization.
     if (ev->glfw_ev.text) strncpy(ev->__embedded_text, ev->glfw_ev.text, sizeof(ev->__embedded_text) - 1);
     ev->glfw_ev.text = NULL;
-    uint32_t state = ibus_key_state(ev->glfw_ev.mods, ev->glfw_ev.action);
+    uint32_t state = ibus_key_state_from_glfw(ev->glfw_ev.mods, ev->glfw_ev.action);
     if (!glfw_dbus_call_method_with_reply(
             ibus->conn, IBUS_SERVICE, ibus->input_ctx_path, IBUS_INPUT_INTERFACE, "ProcessKeyEvent",
             3000, key_event_processed, ev,
