@@ -402,11 +402,6 @@ struct _GLFWcontext
     _GLFWcontextOSMesa osmesa;
 };
 
-typedef struct GLFWKeyState {
-    uint32_t key;
-    char action;
-} GLFWKeyState;
-
 // Window and context structure
 //
 struct _GLFWwindow
@@ -437,7 +432,7 @@ struct _GLFWwindow
     bool                lockKeyMods;
     int                 cursorMode;
     char                mouseButtons[GLFW_MOUSE_BUTTON_LAST + 1];
-    GLFWKeyState        activated_keys[16];
+    GLFWkeyevent        activated_keys[16];
     // Virtual cursor position when cursor is disabled
     double              virtualCursorPosX, virtualCursorPosY;
     bool                rawMouseMotion;
@@ -717,6 +712,7 @@ void _glfwPlatformSetWindowMonitor(_GLFWwindow* window, _GLFWmonitor* monitor,
                                    int xpos, int ypos, int width, int height,
                                    int refreshRate);
 bool _glfwPlatformToggleFullscreen(_GLFWwindow *w, unsigned int flags);
+bool _glfwPlatformIsFullscreen(_GLFWwindow *w, unsigned int flags);
 int _glfwPlatformWindowFocused(_GLFWwindow* window);
 int _glfwPlatformWindowOccluded(_GLFWwindow* window);
 int _glfwPlatformWindowIconified(_GLFWwindow* window);
