@@ -413,9 +413,15 @@ static void setFullscreen(_GLFWwindow* window, _GLFWmonitor* monitor, bool on)
     setIdleInhibitor(window, on);
 }
 
+
+bool
+_glfwPlatformIsFullscreen(_GLFWwindow *window, unsigned int flags UNUSED) {
+    return window->wl.toplevel_states & TOPLEVEL_STATE_FULLSCREEN;
+}
+
 bool
 _glfwPlatformToggleFullscreen(_GLFWwindow *window, unsigned int flags UNUSED) {
-    bool already_fullscreen = window->wl.toplevel_states & TOPLEVEL_STATE_FULLSCREEN;
+    bool already_fullscreen = _glfwPlatformIsFullscreen(window, flags);
     setFullscreen(window, NULL, !already_fullscreen);
     return !already_fullscreen;
 }
@@ -890,9 +896,9 @@ void _glfwPlatformSetWindowSize(_GLFWwindow* window, int width, int height)
     if (width != window->wl.width || height != window->wl.height) {
         window->wl.user_requested_content_size.width = width;
         window->wl.user_requested_content_size.height = height;
-        int32_t width = 0, height = 0;
-        set_csd_window_geometry(window, &width, &height);
-        window->wl.width = width; window->wl.height = height;
+        int32_t w = 0, h = 0;
+        set_csd_window_geometry(window, &w, &h);
+        window->wl.width = w; window->wl.height = h;
         resizeFramebuffer(window);
         ensure_csd_resources(window);
         wl_surface_commit(window->wl.surface);
@@ -1561,11 +1567,6 @@ static void mark_primary_selection_offer(void *data UNUSED, struct zwp_primary_s
 
 static void
 set_offer_mimetype(_GLFWWaylandDataOffer* offer, const char* mime) {
-    if (strcmp(mime, "text/plain;charset=utf-8") == 0) {
-        offer->plain_text_mime = "text/plain;charset=utf-8";
-    } else if (!offer->plain_text_mime && strcmp(mime, "text/plain")) {
-        offer->plain_text_mime = "text/plain";
-    }
     if (strcmp(mime, clipboard_mime()) == 0) {
         offer->is_self_offer = true;
     }
@@ -1666,9 +1667,9 @@ static void drag_enter(void *data UNUSED, struct wl_data_device *wl_data_device
             while (window)
             {
                 if (window->wl.surface == surface) {
-                    for (size_t x = 0; x < d->mimes_count; x++) {
-                        int prio = _glfwInputDrop(window, d->mimes[x], NULL, 0);
-                        if (prio > format_priority) d->mime_for_drop = d->mimes[x];
+                    for (size_t j = 0; j < d->mimes_count; j++) {
+                        int prio = _glfwInputDrop(window, d->mimes[j], NULL, 0);
+                        if (prio > format_priority) d->mime_for_drop = d->mimes[j];
                     }
                     break;
                 }
@@ -1694,8 +1695,8 @@ static void drop(void *data UNUSED, struct wl_data_device *wl_data_device UNUSED
     for (size_t i = 0; i < arraysz(_glfw.wl.dataOffers); i++) {
         if (_glfw.wl.dataOffers[i].offer_type == DRAG_AND_DROP && _glfw.wl.dataOffers[i].mime_for_drop) {
             size_t sz = 0;
-            char *data = read_data_offer(_glfw.wl.dataOffers[i].id, _glfw.wl.dataOffers[i].mime_for_drop, &sz);
-            if (data) {
+            char *d = read_data_offer(_glfw.wl.dataOffers[i].id, _glfw.wl.dataOffers[i].mime_for_drop, &sz);
+            if (d) {
                 // We dont do finish as this requires version 3 for wl_data_device_manager
                 // which then requires more work with calling set_actions for drag and drop to function
                 // wl_data_offer_finish(_glfw.wl.dataOffers[i].id);
@@ -1704,13 +1705,13 @@ static void drop(void *data UNUSED, struct wl_data_device *wl_data_device UNUSED
                 while (window)
                 {
                     if (window->wl.surface == _glfw.wl.dataOffers[i].surface) {
-                        _glfwInputDrop(window, _glfw.wl.dataOffers[i].mime_for_drop, data, sz);
+                        _glfwInputDrop(window, _glfw.wl.dataOffers[i].mime_for_drop, d, sz);
                         break;
                     }
                     window = window->next;
                 }
 
-                free(data);
+                free(d);
             }
             destroy_data_offer(&_glfw.wl.dataOffers[i]);
             break;
@@ -1762,7 +1763,7 @@ void _glfwSetupWaylandPrimarySelectionDevice() {
     if (_glfw.wl.primarySelectionDevice) zwp_primary_selection_device_v1_add_listener(_glfw.wl.primarySelectionDevice, &primary_selection_device_listener, NULL);
 }
 
-static inline bool _glfwEnsureDataDevice(void) {
+static bool _glfwEnsureDataDevice(void) {
     if (!_glfw.wl.dataDeviceManager)
     {
         _glfwInputError(GLFW_PLATFORM_ERROR,
@@ -1814,16 +1815,38 @@ void _glfwPlatformSetClipboardString(const char* string)
     wl_callback_add_listener(callback, &clipboard_copy_callback_listener, _glfw.wl.dataSourceForClipboard);
 }
 
+static bool
+offer_has_mime(const _GLFWWaylandDataOffer *d, const char *mime) {
+    for (unsigned i = 0; i < d->mimes_count; i++) {
+        if (strcmp(d->mimes[i], mime) == 0) return true;
+    }
+    return false;
+}
+
+static const char*
+plain_text_mime_for_offer(const _GLFWWaylandDataOffer *d) {
+#define A(x) if (offer_has_mime(d, x)) return x;
+    A("text/plain;charset=utf-8");
+    A("text/plain");
+    A("UTF8_STRING");
+    A("STRING");
+#undef A
+    return NULL;
+}
+
 const char* _glfwPlatformGetClipboardString(void)
 {
     for (size_t i = 0; i < arraysz(_glfw.wl.dataOffers); i++) {
         _GLFWWaylandDataOffer *d = _glfw.wl.dataOffers + i;
-        if (d->id && d->offer_type == CLIPBOARD && d->plain_text_mime) {
+        if (d->id && d->offer_type == CLIPBOARD) {
             if (d->is_self_offer) return _glfw.wl.clipboardString;
-            free(_glfw.wl.pasteString);
-            size_t sz = 0;
-            _glfw.wl.pasteString = read_data_offer(d->id, d->plain_text_mime, &sz);
-            return _glfw.wl.pasteString;
+            const char *mime = plain_text_mime_for_offer(d);
+            if (mime) {
+                free(_glfw.wl.pasteString);
+                size_t sz = 0;
+                _glfw.wl.pasteString = read_data_offer(d->id, mime, &sz);
+                return _glfw.wl.pasteString;
+            }
         }
     }
     return NULL;
@@ -1871,13 +1894,14 @@ const char* _glfwPlatformGetPrimarySelectionString(void)
 
     for (size_t i = 0; i < arraysz(_glfw.wl.dataOffers); i++) {
         _GLFWWaylandDataOffer *d = _glfw.wl.dataOffers + i;
-        if (d->id && d->is_primary && d->offer_type == PRIMARY_SELECTION && d->plain_text_mime) {
-            if (d->is_self_offer) {
-                return _glfw.wl.primarySelectionString;
+        if (d->id && d->is_primary && d->offer_type == PRIMARY_SELECTION) {
+            if (d->is_self_offer) return _glfw.wl.primarySelectionString;
+            const char *mime = plain_text_mime_for_offer(d);
+            if (mime) {
+                free(_glfw.wl.pasteString);
+                _glfw.wl.pasteString = read_primary_selection_offer(d->id, mime);
+                return _glfw.wl.pasteString;
             }
-            free(_glfw.wl.pasteString);
-            _glfw.wl.pasteString = read_primary_selection_offer(d->id, d->plain_text_mime);
-            return _glfw.wl.pasteString;
         }
     }
     return NULL;
