@@ -1,18 +1,18 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import warnings
-from functools import lru_cache
 from gettext import gettext as _
 from itertools import repeat, zip_longest
 from math import ceil
 from typing import Callable, Dict, Generator, Iterable, List, Optional, Tuple
 
-from kitty.fast_data_types import truncate_point_for_length, wcswidth
 from kitty.cli_stub import DiffCLIOptions
+from kitty.fast_data_types import truncate_point_for_length, wcswidth
+from kitty.types import run_once
 from kitty.utils import ScreenSize
 
+from ..tui.images import ImageManager, can_display_images
 from .collect import (
     Collection, Segment, data_for_path, highlights_for_path, is_image,
     lines_for_path, path_name_map, sanitize
@@ -20,14 +20,13 @@ from .collect import (
 from .config import formats
 from .diff_speedup import split_with_highlights as _split_with_highlights
 from .patch import Chunk, Hunk, Patch
-from ..tui.images import ImageManager, can_display_images
 
 
 class ImageSupportWarning(Warning):
     pass
 
 
-@lru_cache(maxsize=2)
+@run_once
 def images_supported() -> bool:
     ans = can_display_images()
     if not ans:
@@ -42,7 +41,7 @@ class Ref:
 
     def __repr__(self) -> str:
         return '{}({})'.format(self.__class__.__name__, ', '.join(
-            '{}={}'.format(n, getattr(self, n)) for n in self.__slots__ if n != '_hash'))
+            f'{n}={getattr(self, n)}' for n in self.__slots__ if n != '_hash'))
 
 
 class LineRef(Ref):
@@ -141,6 +140,7 @@ removed_format = format_func('removed')
 removed_margin_format = format_func('removed_margin')
 added_margin_format = format_func('added_margin')
 filler_format = format_func('filler')
+margin_filler_format = format_func('margin_filler')
 hunk_margin_format = format_func('hunk_margin')
 hunk_format = format_func('hunk')
 highlight_map = {'remove': ('removed_highlight', 'removed'), 'add': ('added_highlight', 'added')}
@@ -209,12 +209,12 @@ def truncate_points(line: str, width: int) -> Generator[int, None, None]:
             break
 
 
-def split_with_highlights(line: str, width: int, highlights: List, bg_highlight: Optional[Segment] = None) -> List:
+def split_with_highlights(line: str, width: int, highlights: List[Segment], bg_highlight: Optional[Segment] = None) -> List[str]:
     truncate_pts = list(truncate_points(line, width))
     return _split_with_highlights(line, truncate_pts, highlights, bg_highlight)
 
 
-margin_bg_map = {'filler': filler_format, 'remove': removed_margin_format, 'add': added_margin_format, 'context': margin_format}
+margin_bg_map = {'filler': margin_filler_format, 'remove': removed_margin_format, 'add': added_margin_format, 'context': margin_format}
 text_bg_map = {'filler': filler_format, 'remove': removed_format, 'add': added_format, 'context': text_format}
 
 
@@ -263,14 +263,14 @@ def render_diff_pair(
 
 def hunk_title(hunk_num: int, hunk: Hunk, margin_size: int, available_cols: int) -> str:
     m = hunk_margin_format(' ' * margin_size)
-    t = '@@ -{},{} +{},{} @@ {}'.format(hunk.left_start + 1, hunk.left_count, hunk.right_start + 1, hunk.right_count, hunk.title)
+    t = f'@@ -{hunk.left_start + 1},{hunk.left_count} +{hunk.right_start + 1},{hunk.right_count} @@ {hunk.title}'
     return m + hunk_format(place_in(t, available_cols))
 
 
 def render_half_line(
     line_number: int,
     line: str,
-    highlights: List,
+    highlights: List[Segment],
     ltype: str,
     margin_size: int,
     available_cols: int,
@@ -538,7 +538,7 @@ class RenderDiff:
                 assert other_path is not None
                 ans = yield_lines_from(rename_lines(path, other_path, args, columns, margin_size), item_ref)
             else:
-                raise ValueError('Unsupported item type: {}'.format(item_type))
+                raise ValueError(f'Unsupported item type: {item_type}')
             yield from ans
             if i < last_item_num:
                 yield Line('', item_ref)
