@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -16,10 +15,13 @@ from typing import (
 
 from kitty.cli import parse_args
 from kitty.cli_stub import HintsCLIOptions
-from kitty.fast_data_types import set_clipboard_string
+from kitty.constants import website_url
+from kitty.fast_data_types import get_options, set_clipboard_string
 from kitty.key_encoding import KeyEvent
 from kitty.typing import BossType, KittyCommonOpts
-from kitty.utils import ScreenSize, screen_size_function, set_primary_selection
+from kitty.utils import (
+    ScreenSize, resolve_custom_file, screen_size_function, set_primary_selection
+)
 
 from ..tui.handler import Handler, result_handler
 from ..tui.loop import Loop
@@ -94,6 +96,11 @@ def highlight_mark(m: Mark, text: str, current_input: str, alphabet: str, colors
     )
 
 
+def debug(*a: Any, **kw: Any) -> None:
+    from ..tui.loop import debug as d
+    d(*a, **kw)
+
+
 def render(text: str, current_input: str, all_marks: Sequence[Mark], ignore_mark_indices: Set[int], alphabet: str, colors: Dict[str, str]) -> str:
     for mark in reversed(all_marks):
         if mark.index in ignore_mark_indices:
@@ -102,8 +109,7 @@ def render(text: str, current_input: str, all_marks: Sequence[Mark], ignore_mark
         text = text[:mark.start] + mtext + text[mark.end:]
 
     text = text.replace('\0', '')
-
-    return text.replace('\n', '\r\n').rstrip()
+    return re.sub('[\r\n]', '\r\n', text).rstrip()
 
 
 class Hints(Handler):
@@ -212,7 +218,7 @@ class Hints(Handler):
         self.write(self.current_text)
 
 
-def regex_finditer(pat: Pattern, minimum_match_length: int, text: str) -> Generator[Tuple[int, int, Dict], None, None]:
+def regex_finditer(pat: 'Pattern[str]', minimum_match_length: int, text: str) -> Generator[Tuple[int, int, Dict[str, str]], None, None]:
     has_named_groups = bool(pat.groupindex)
     for m in pat.finditer(text):
         s, e = m.span(0 if has_named_groups else pat.groups)
@@ -265,9 +271,13 @@ def brackets(text: str, s: int, e: int) -> Tuple[int, int]:
     # Remove matching brackets
     if s < e <= len(text):
         before = text[s]
-        if before in '({[<' and text[e-1] == closing_bracket_map[before]:
-            s += 1
-            e -= 1
+        if before in '({[<':
+            q = closing_bracket_map[before]
+            if text[e-1] == q:
+                s += 1
+                e -= 1
+            elif text[e:e+1] == q:
+                s += 1
     return s, e
 
 
@@ -276,9 +286,12 @@ def quotes(text: str, s: int, e: int) -> Tuple[int, int]:
     # Remove matching quotes
     if s < e <= len(text):
         before = text[s]
-        if before in '\'"' and text[e-1] == before:
-            s += 1
-            e -= 1
+        if before in '\'"':
+            if text[e-1] == before:
+                s += 1
+                e -= 1
+            elif text[e:e+1] == before:
+                s += 1
     return s, e
 
 
@@ -306,7 +319,7 @@ def mark(pattern: str, post_processors: Iterable[PostprocessorFunc], text: str,
         except InvalidMatch:
             continue
 
-        mark_text = text[s:e].replace('\n', '').replace('\0', '')
+        mark_text = re.sub('[\r\n\0]', '', text[s:e])
         yield Mark(idx, s, e, mark_text, groupdict)
 
 
@@ -342,12 +355,12 @@ def functions_for(args: HintsCLIOptions) -> Tuple[str, List[PostprocessorFunc]]:
         )
         post_processors.append(url)
     elif args.type == 'path':
-        pattern = r'(?:\S*/\S+)|(?:\S+[.][a-zA-Z0-9]{2,7})'
+        pattern = r'(?:\S*?/[\r\S]+)|(?:\S[\r\S]*\.[a-zA-Z0-9\r]{2,7})'
         post_processors.extend((brackets, quotes))
     elif args.type == 'line':
         pattern = '(?m)^\\s*(.+)[\\s\0]*$'
     elif args.type == 'hash':
-        pattern = '[0-9a-f]{7,128}'
+        pattern = '[0-9a-f][0-9a-f\r]{6,127}'
     elif args.type == 'ip':
         pattern = (
             # # IPv4 with no validation
@@ -361,7 +374,7 @@ def functions_for(args: HintsCLIOptions) -> Tuple[str, List[PostprocessorFunc]]:
         chars = args.word_characters
         if chars is None:
             chars = kitty_common_opts()['select_by_word_characters']
-        pattern = r'(?u)[{}\w]{{{},}}'.format(escape(chars), args.minimum_match_length)
+        pattern = fr'(?u)[{escape(chars)}\w]{{{args.minimum_match_length},}}'
         post_processors.extend((brackets, quotes))
     else:
         pattern = args.regex
@@ -370,16 +383,22 @@ def functions_for(args: HintsCLIOptions) -> Tuple[str, List[PostprocessorFunc]]:
 
 def convert_text(text: str, cols: int) -> str:
     lines: List[str] = []
-    empty_line = '\0' * cols
+    empty_line = '\0' * cols + '\n'
     for full_line in text.split('\n'):
         if full_line:
             if not full_line.rstrip('\r'):  # empty lines
                 lines.extend(repeat(empty_line, len(full_line)))
                 continue
+            appended = False
             for line in full_line.split('\r'):
                 if line:
                     lines.append(line.ljust(cols, '\0'))
-    return '\n'.join(lines)
+                    lines.append('\r')
+                    appended = True
+            if appended:
+                lines[-1] = '\n'
+    rstripped = re.sub('[\r\n]+$', '', ''.join(lines))
+    return rstripped
 
 
 def parse_input(text: str) -> str:
@@ -404,7 +423,6 @@ def load_custom_processor(customize_processing: str) -> Any:
         return {k: getattr(m, k) for k in dir(m)}
     if customize_processing == '::linenum::':
         return {'mark': linenum_marks, 'handle_result': linenum_handle_result}
-    from kitty.constants import resolve_custom_file
     custom_path = resolve_custom_file(customize_processing)
     import runpy
     return runpy.run_path(custom_path, run_name='__main__')
@@ -434,7 +452,7 @@ def process_hyperlinks(text: str) -> Tuple[str, Tuple[Mark, ...]]:
         active_hyperlink_start_offset = 0
         idx += 1
 
-    def process_hyperlink(m: 're.Match') -> str:
+    def process_hyperlink(m: 're.Match[str]') -> str:
         nonlocal removed_size, active_hyperlink_url, active_hyperlink_id, active_hyperlink_start_offset
         raw = m.group()
         start = m.start() - removed_size
@@ -542,10 +560,12 @@ choices=self,window,tab,os_window,background
 Where to perform the action on matched errors. :code:`self` means the current
 window, :code:`window` a new kitty window, :code:`tab` a new tab,
 :code:`os_window` a new OS window and :code:`background` run in the background.
-The action to perform on the matched errors. The actual action is whatever
-arguments are provided to the kitten, for example: :code:`kitty + kitten hints
---type=linenum --linenum-action=tab vim +{line} {path}` will open the matched
-path at the matched line number in vim in a new kitty tab.
+The actual action is whatever arguments are provided to the kitten, for
+example:
+:code:`kitty + kitten hints --type=linenum --linenum-action=tab vim +{line} {path}`
+will open the matched path at the matched line number in vim in
+a new kitty tab. Note that only when using :code:`self` are the special values for
+:option:`kitty +kitten hints --program` to copy/paste the text respected.
 
 
 --url-prefixes
@@ -587,7 +607,7 @@ indices into the list of selections. You can use 0 for the first selection and
 default=auto
 choices=auto,always,never
 Add trailing space after matched text. Defaults to auto, which adds the space
-when used together with --multiple.
+when used together with :option:`--multiple`.
 
 
 --hints-offset
@@ -630,7 +650,7 @@ The foreground color for text pointed to by the hints
 --customize-processing
 Name of a python file in the kitty config directory which will be imported to provide
 custom implementations for pattern finding and performing actions
-on selected matches. See https://sw.kovidgoyal.net/kitty/kittens/hints.html
+on selected matches. See {hints_url}
 for details. You can also specify absolute paths to load the script from elsewhere.
 
 
@@ -639,7 +659,8 @@ The window title for the hints window, default title is selected based on
 the type of text being hinted.
 '''.format(
     default_regex=DEFAULT_REGEX,
-    line='{{line}}', path='{{path}}'
+    line='{{line}}', path='{{path}}',
+    hints_url=website_url('kittens/hints'),
 ).format
 help_text = 'Select text from the screen using the keyboard. Defaults to searching for URLs.'
 usage = ''
@@ -675,6 +696,7 @@ def main(args: List[str]) -> Optional[Dict[str, Any]]:
         import traceback
         traceback.print_exc()
         input(_('Press Enter to quit'))
+    return None
 
 
 def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_id: int, boss: BossType, extra_cli_args: Sequence[str], *a: Any) -> None:
@@ -693,9 +715,19 @@ def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_i
 
     if action == 'self':
         if w is not None:
-            import shlex
-            text = ' '.join(shlex.quote(arg) for arg in cmd)
-            w.paste_bytes(text + '\r')
+            is_copy_action = cmd[0] in ('-', '@', '*')
+            if is_copy_action:
+                text = ' '.join(cmd[1:])
+                if cmd[0] == '-':
+                    w.paste_bytes(text)
+                elif cmd[0] == '@':
+                    set_clipboard_string(text)
+                elif cmd[0] == '*':
+                    set_primary_selection(text)
+            else:
+                import shlex
+                text = ' '.join(shlex.quote(arg) for arg in cmd)
+                w.paste_bytes(text + '\r')
     elif action == 'background':
         import subprocess
         subprocess.Popen(cmd, cwd=data['cwd'])
@@ -754,7 +786,7 @@ def handle_result(args: List[str], data: Dict[str, Any], target_window_id: int,
             set_primary_selection(joined_text())
         else:
             cwd = data['cwd']
-            program = None if program == 'default' else program
+            program = get_options().open_url_with if program == 'default' else program
             if text_type == 'hyperlink':
                 w = boss.window_id_map.get(target_window_id)
                 for m in matches:
