@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import contextlib
@@ -23,7 +22,7 @@ from kitty.guess_mime_type import guess_type
 from kitty.types import run_once
 from kitty.typing import GRT_f, GRT_t
 from kitty.utils import (
-    TTYIO, ScreenSize, ScreenSizeGetter, fit_image, screen_size_function
+    TTYIO, ScreenSize, ScreenSizeGetter, screen_size_function
 )
 
 from ..tui.images import (
@@ -110,7 +109,7 @@ Do not print out anything to STDOUT during operation.
 default=0
 Z-index of the image. When negative, text will be displayed on top of the image. Use
 a double minus for values under the threshold for drawing images under cell background
-colors. For example, --1 evaluates as -1,073,741,825.
+colors. For example, :code:`--1` evaluates as -1,073,741,825.
 
 
 --loop -l
@@ -145,7 +144,7 @@ def get_screen_size() -> ScreenSize:
 
 @run_once
 def options_spec() -> str:
-    return OPTIONS.format(appname='{}-icat'.format(appname))
+    return OPTIONS.format(appname=f'{appname}-icat')
 
 
 def write_gr_cmd(cmd: GraphicsCommand, payload: Optional[bytes] = None) -> None:
@@ -168,20 +167,14 @@ def set_cursor(cmd: GraphicsCommand, width: int, height: int, align: str) -> Non
     ss = get_screen_size()
     cw = int(ss.width / ss.cols)
     num_of_cells_needed = int(ceil(width / cw))
-    if num_of_cells_needed > ss.cols:
-        w, h = fit_image(width, height, ss.width, height)
-        ch = int(ss.height / ss.rows)
-        num_of_rows_needed = int(ceil(height / ch))
-        cmd.c, cmd.r = ss.cols, num_of_rows_needed
-    else:
-        cmd.X = calculate_in_cell_x_offset(width, cw, align)
-        extra_cells = 0
-        if align == 'center':
-            extra_cells = (ss.cols - num_of_cells_needed) // 2
-        elif align == 'right':
-            extra_cells = (ss.cols - num_of_cells_needed)
-        if extra_cells:
-            sys.stdout.buffer.write(b' ' * extra_cells)
+    cmd.X = calculate_in_cell_x_offset(width, cw, align)
+    extra_cells = 0
+    if align == 'center':
+        extra_cells = (ss.cols - num_of_cells_needed) // 2
+    elif align == 'right':
+        extra_cells = (ss.cols - num_of_cells_needed)
+    if extra_cells:
+        sys.stdout.buffer.write(b' ' * extra_cells)
 
 
 def set_cursor_for_place(place: 'Place', cmd: GraphicsCommand, width: int, height: int, align: str) -> None:
@@ -195,7 +188,7 @@ def set_cursor_for_place(place: 'Place', cmd: GraphicsCommand, width: int, heigh
         extra_cells = (place.width - num_of_cells_needed) // 2
     elif align == 'right':
         extra_cells = place.width - num_of_cells_needed
-    sys.stdout.buffer.write('\033[{};{}H'.format(place.top + 1, x + extra_cells).encode('ascii'))
+    sys.stdout.buffer.write(f'\033[{place.top + 1};{x + extra_cells}H'.encode('ascii'))
 
 
 def write_chunked(cmd: GraphicsCommand, data: bytes) -> None:
@@ -369,7 +362,7 @@ def scan(d: str) -> Generator[Tuple[str, str], None, None]:
 def detect_support(wait_for: float = 10, silent: bool = False) -> bool:
     global can_transfer_with_files
     if not silent:
-        print('Checking for graphics ({}s max. wait)...'.format(wait_for), end='\r')
+        print(f'Checking for graphics ({wait_for}s max. wait)...', end='\r')
     sys.stdout.flush()
     try:
         received = b''
@@ -390,7 +383,8 @@ def detect_support(wait_for: float = 10, silent: bool = False) -> bool:
             return 1 not in responses or 2 not in responses
 
         with NamedTemporaryFile() as f:
-            f.write(b'abcd'), f.flush()
+            f.write(b'abcd')
+            f.flush()
             gc = GraphicsCommand()
             gc.a = 'q'
             gc.s = gc.v = gc.i = 1
@@ -448,7 +442,7 @@ def process_single_item(
     item: Union[bytes, str],
     args: IcatCLIOptions,
     parsed_opts: ParsedOpts,
-    url_pat: Optional[Pattern] = None,
+    url_pat: Optional['Pattern[str]'] = None,
     maybe_dir: bool = True
 ) -> None:
     is_tempfile = False
@@ -456,7 +450,8 @@ def process_single_item(
     try:
         if isinstance(item, bytes):
             tf = NamedTemporaryFile(prefix='stdin-image-data-', delete=False)
-            tf.write(item), tf.close()
+            tf.write(item)
+            tf.close()
             item = tf.name
             is_tempfile = True
         if url_pat is not None and url_pat.match(item) is not None:
@@ -466,7 +461,7 @@ def process_single_item(
                     with socket_timeout(30):
                         urlretrieve(item, filename=tf.name)
                 except Exception as e:
-                    raise SystemExit('Failed to download image at URL: {} with error: {}'.format(item, e))
+                    raise SystemExit(f'Failed to download image at URL: {item} with error: {e}')
                 item = tf.name
             is_tempfile = True
             file_removed = process(item, args, parsed_opts, is_tempfile)
@@ -493,25 +488,25 @@ def process_single_item(
 
 def main(args: List[str] = sys.argv) -> None:
     global can_transfer_with_files
-    cli_opts, items_ = parse_args(args[1:], options_spec, usage, help_text, '{} +kitten icat'.format(appname), result_class=IcatCLIOptions)
+    cli_opts, items_ = parse_args(args[1:], options_spec, usage, help_text, f'{appname} +kitten icat', result_class=IcatCLIOptions)
     items: List[Union[str, bytes]] = list(items_)
 
     if cli_opts.print_window_size:
         screen_size_function.cache_clear()
         with open(os.ctermid()) as tty:
             ss = screen_size_function(tty)()
-        print('{}x{}'.format(ss.width, ss.height), end='')
+        print(f'{ss.width}x{ss.height}', end='')
         raise SystemExit(0)
 
     if not sys.stdout.isatty():
         sys.stdout = open(os.ctermid(), 'w')
     stdin_data = None
-    if cli_opts.stdin == 'yes' or (not sys.stdin.isatty() and cli_opts.stdin == 'detect'):
+    if cli_opts.stdin == 'yes' or (cli_opts.stdin == 'detect' and sys.stdin is not None and not sys.stdin.isatty()):
         stdin_data = sys.stdin.buffer.read()
         if stdin_data:
             items.insert(0, stdin_data)
         sys.stdin.close()
-        sys.stdin = open(os.ctermid(), 'r')
+        sys.stdin = open(os.ctermid())
 
     screen_size = get_screen_size_function()
     signal.signal(signal.SIGWINCH, lambda signum, frame: setattr(screen_size, 'changed', True))
@@ -526,12 +521,12 @@ def main(args: List[str] = sys.argv) -> None:
         try:
             parsed_opts.place = parse_place(cli_opts.place)
         except Exception:
-            raise SystemExit('Not a valid place specification: {}'.format(cli_opts.place))
+            raise SystemExit(f'Not a valid place specification: {cli_opts.place}')
 
     try:
         parsed_opts.z_index = parse_z_index(cli_opts.z_index)
     except Exception:
-        raise SystemExit('Not a valid z-index specification: {}'.format(cli_opts.z_index))
+        raise SystemExit(f'Not a valid z-index specification: {cli_opts.z_index}')
 
     if cli_opts.detect_support:
         if not detect_support(wait_for=cli_opts.detection_timeout, silent=True):
