@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -140,27 +139,40 @@ def get_ssh_cli() -> Tuple[Set[str], Set[str]]:
     return boolean_ssh_args, other_ssh_args
 
 
-def get_connection_data(args: List[str]) -> Optional[SSHConnectionData]:
+def get_connection_data(args: List[str], cwd: str = '') -> Optional[SSHConnectionData]:
     boolean_ssh_args, other_ssh_args = get_ssh_cli()
-    found_ssh = ''
     port: Optional[int] = None
-    expecting_port = False
+    expecting_port = expecting_identity = False
     expecting_option_val = False
+    expecting_hostname = False
+    host_name = identity_file = found_ssh = ''
 
     for i, arg in enumerate(args):
         if not found_ssh:
             if os.path.basename(arg).lower() in ('ssh', 'ssh.exe'):
                 found_ssh = arg
             continue
+        if expecting_hostname:
+            host_name = arg
+            continue
         if arg.startswith('-') and not expecting_option_val:
             if arg in boolean_ssh_args:
                 continue
+            if arg == '--':
+                expecting_hostname = True
             if arg.startswith('-p'):
                 if arg[2:].isdigit():
                     with suppress(Exception):
                         port = int(arg[2:])
+                    continue
                 elif arg == '-p':
                     expecting_port = True
+            elif arg.startswith('-i'):
+                if arg == '-i':
+                    expecting_identity = True
+                else:
+                    identity_file = arg[2:]
+                    continue
             expecting_option_val = True
             continue
 
@@ -169,10 +181,22 @@ def get_connection_data(args: List[str]) -> Optional[SSHConnectionData]:
                 with suppress(Exception):
                     port = int(arg)
                 expecting_port = False
+            elif expecting_identity:
+                identity_file = arg
             expecting_option_val = False
             continue
 
-        return SSHConnectionData(found_ssh, arg, port)
+        if not host_name:
+            host_name = arg
+    if not host_name:
+        return None
+    if identity_file:
+        if not os.path.isabs(identity_file):
+            identity_file = os.path.expanduser(identity_file)
+        if not os.path.isabs(identity_file):
+            identity_file = os.path.normpath(os.path.join(cwd or os.getcwd(), identity_file))
+
+    return SSHConnectionData(found_ssh, host_name, port, identity_file)
 
 
 class InvalidSSHArgs(ValueError):
@@ -195,16 +219,16 @@ def parse_ssh_args(args: List[str]) -> Tuple[List[str], List[str], bool]:
     expecting_option_val = False
     passthrough = False
     stop_option_processing = False
-    for arg in args:
+    for argument in args:
         if len(server_args) > 1 or stop_option_processing:
-            server_args.append(arg)
+            server_args.append(argument)
             continue
-        if arg.startswith('-') and not expecting_option_val:
-            if arg == '--':
+        if argument.startswith('-') and not expecting_option_val:
+            if argument == '--':
                 stop_option_processing = True
                 continue
             # could be a multi-character option
-            all_args = arg[1:]
+            all_args = argument[1:]
             for i, arg in enumerate(all_args):
                 arg = '-' + arg
                 if arg in passthrough_args:
@@ -220,13 +244,13 @@ def parse_ssh_args(args: List[str]) -> Tuple[List[str], List[str], bool]:
                     else:
                         expecting_option_val = True
                     break
-                raise InvalidSSHArgs('unknown option -- {}'.format(arg[1:]))
+                raise InvalidSSHArgs(f'unknown option -- {arg[1:]}')
             continue
         if expecting_option_val:
-            ssh_args.append(arg)
+            ssh_args.append(argument)
             expecting_option_val = False
             continue
-        server_args.append(arg)
+        server_args.append(argument)
     if not server_args:
         raise InvalidSSHArgs()
     return ssh_args, server_args, passthrough
