@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 import os
 import string
@@ -110,7 +109,7 @@ def serialize_favorites(favorites: Iterable[int]) -> str:
 
 '''.splitlines()
     for cp in favorites:
-        ans.append('{:x} # {} {}'.format(cp, chr(cp), name(cp)))
+        ans.append(f'{cp:x} # {chr(cp)} {name(cp)}')
     return '\n'.join(ans)
 
 
@@ -148,20 +147,24 @@ class Table:
         self.last_rows = self.last_cols = -1
         self.codepoints: List[int] = []
         self.current_idx = 0
+        self.scroll_rows = 0
         self.text = ''
         self.num_cols = 0
+        self.num_rows = 0
         self.mode = HEX
 
     @property
     def current_codepoint(self) -> Optional[int]:
         if self.codepoints:
             return self.codepoints[self.current_idx]
+        return None
 
     def set_codepoints(self, codepoints: List[int], mode: str = HEX, current_idx: int = 0) -> None:
         self.codepoints = codepoints
         self.mode = mode
         self.layout_dirty = True
         self.current_idx = current_idx if current_idx < len(codepoints) else 0
+        self.scroll_rows = 0
 
     def codepoint_at_hint(self, hint: str) -> int:
         return self.codepoints[decode_hint(hint)]
@@ -213,6 +216,7 @@ class Table:
         if num < 1:
             self.text = ''
             self.num_cols = 0
+            self.num_rows = 0
             return self.text
         idx_size = len(encode_hint(num - 1))
 
@@ -228,16 +232,20 @@ class Table:
         num_cols = self.num_cols = max(cols // col_width, 1)
         buf: List[str] = []
         a = buf.append
-        rows_left = rows
+        rows_left = self.num_rows = rows
+        skip_scroll = self.scroll_rows * num_cols
 
         for i, (idx, c, desc) in enumerate(parts):
-            if i > 0 and i % num_cols == 0:
+            if skip_scroll > 0:
+                skip_scroll -= 1
+                continue
+            buf.extend(cell(i, idx, c, desc))
+            a('  ')
+            if i > 0 and (i+1) % num_cols == 0:
                 rows_left -= 1
                 if rows_left == 0:
                     break
                 a('\r\n')
-            buf.extend(cell(i, idx, c, desc))
-            a('  ')
         self.text = ''.join(buf)
         return self.text
 
@@ -252,6 +260,13 @@ class Table:
             self.current_idx += amt
             self.current_idx = max(0, min(self.current_idx, len(self.codepoints) - 1))
             self.layout_dirty = True
+        first_visible = self.scroll_rows * self.num_cols
+        last_visible = first_visible + ((self.num_cols * self.num_rows) - 1)
+        scroll_amount = self.num_rows
+        if self.current_idx < first_visible:
+            self.scroll_rows = max(self.scroll_rows - scroll_amount, 0)
+        if self.current_idx > last_visible:
+            self.scroll_rows += scroll_amount
 
 
 def is_index(w: str) -> bool:
@@ -366,7 +381,7 @@ class UnicodeInput(Handler):
     def draw_title_bar(self) -> None:
         entries = []
         for name, key, mode in all_modes:
-            entry = ' {} ({}) '.format(name, key)
+            entry = f' {name} ({key}) '
             if mode is self.mode:
                 entry = styled(entry, reverse=False, bold=True)
             entries.append(entry)
@@ -376,6 +391,7 @@ class UnicodeInput(Handler):
             text += ' ' * extra
         self.print(styled(text, reverse=True))
 
+    @Handler.atomic_update
     def draw_screen(self) -> None:
         self.write(clear_screen())
         self.draw_title_bar()
@@ -434,7 +450,7 @@ class UnicodeInput(Handler):
                     self.line_edit.current_input = hex(val - 1)[2:]
                     self.refresh()
                     return
-        if self.mode is NAME and key_event.type is not EventType.RELEASE and not key_event.has_mods:
+        if self.mode is NAME:
             if key_event.matches('shift+tab'):
                 self.table.move_current(cols=-1)
                 self.refresh()
@@ -469,19 +485,19 @@ class UnicodeInput(Handler):
         if key_event.matches('esc'):
             self.quit_loop(1)
             return
-        if key_event.matches('f1'):
+        if key_event.matches_without_mods('f1') or key_event.matches('ctrl+1'):
             self.switch_mode(HEX)
             return
-        if key_event.matches('f2'):
+        if key_event.matches_without_mods('f2') or key_event.matches('ctrl+2'):
             self.switch_mode(NAME)
             return
-        if key_event.matches('f3'):
+        if key_event.matches_without_mods('f3') or key_event.matches('ctrl+3'):
             self.switch_mode(EMOTICONS)
             return
-        if key_event.matches('f4'):
+        if key_event.matches_without_mods('f4') or key_event.matches('ctrl+4'):
             self.switch_mode(FAVORITES)
             return
-        if key_event.matches('f12') and self.mode is FAVORITES:
+        if key_event.matches_without_mods('f12') and self.mode is FAVORITES:
             self.edit_favorites()
             return
         if key_event.matches('ctrl+shift+tab'):
