@@ -78,10 +78,9 @@ static const Child EMPTY_CHILD = {0};
 
 static Child children[MAX_CHILDREN] = {{0}};
 static Child scratch[MAX_CHILDREN] = {{0}};
-static Child add_queue[MAX_CHILDREN] = {{0}}, remove_queue[MAX_CHILDREN] = {{0}};
-static unsigned long remove_notify[MAX_CHILDREN] = {0};
+static Child add_queue[MAX_CHILDREN] = {{0}}, remove_queue[MAX_CHILDREN] = {{0}}, remove_notify[MAX_CHILDREN] = {{0}};
 static size_t add_queue_count = 0, remove_queue_count = 0;
-static struct pollfd fds[MAX_CHILDREN + EXTRA_FDS] = {{0}};
+static struct pollfd children_fds[MAX_CHILDREN + EXTRA_FDS] = {{0}};
 static pthread_mutex_t children_lock, talk_lock;
 static bool kill_signal_received = false, reload_config_signal_received = false;
 static ChildMonitor *the_monitor = NULL;
@@ -145,8 +144,8 @@ new(PyTypeObject *type, PyObject *args, PyObject UNUSED *kwds) {
         parse_func = parse_worker_dump;
     } else parse_func = parse_worker;
     self->count = 0;
-    fds[0].fd = self->io_loop_data.wakeup_read_fd; fds[1].fd = self->io_loop_data.signal_read_fd;
-    fds[0].events = POLLIN; fds[1].events = POLLIN;
+    children_fds[0].fd = self->io_loop_data.wakeup_read_fd; children_fds[1].fd = self->io_loop_data.signal_read_fd;
+    children_fds[0].events = POLLIN; children_fds[1].events = POLLIN;
     the_monitor = self;
 
     return (PyObject*) self;
@@ -182,7 +181,7 @@ wakeup_io_loop(ChildMonitor *self, bool in_signal_handler) {
 
 static void* io_loop(void *data);
 static void* talk_loop(void *data);
-static void send_response(id_type peer_id, const char *msg, size_t msg_sz);
+static void send_response_to_peer(id_type peer_id, const char *msg, size_t msg_sz);
 static void wakeup_talk_loop(bool);
 static bool talk_thread_started = false;
 
@@ -229,56 +228,92 @@ add_child(ChildMonitor *self, PyObject *args) {
     Py_RETURN_NONE;
 }
 
+#define schedule_write_to_child_generic(id, num, va_start, get_next_arg, va_end) \
+    ChildMonitor *self = the_monitor; \
+    bool found = false; \
+    const char *data; \
+    size_t szval, sz = 0; \
+    va_start(ap, num); \
+    for (unsigned int i = 0; i < num; i++) { \
+        get_next_arg(ap); \
+        sz += szval; \
+    } \
+    va_end(ap); \
+    children_mutex(lock); \
+    for (size_t i = 0; i < self->count; i++) { \
+        if (children[i].id == id) { \
+            Screen *screen = children[i].screen; \
+            screen_mutex(lock, write); \
+            size_t space_left = screen->write_buf_sz - screen->write_buf_used; \
+            if (space_left < sz) { \
+                if (screen->write_buf_used + sz > 100 * 1024 * 1024) { \
+                    log_error("Too much data being sent to child with id: %lu, ignoring it", id); \
+                    screen_mutex(unlock, write); \
+                    break; \
+                } \
+                screen->write_buf_sz = screen->write_buf_used + sz; \
+                screen->write_buf = PyMem_RawRealloc(screen->write_buf, screen->write_buf_sz); \
+                if (screen->write_buf == NULL) { fatal("Out of memory."); } \
+            } \
+            found = true; \
+            va_start(ap, num); \
+            for (unsigned int i = 0; i < num; i++) { \
+                get_next_arg(ap); \
+                memcpy(screen->write_buf + screen->write_buf_used, data, szval); \
+                screen->write_buf_used += szval; \
+            } \
+            va_end(ap); \
+            if (screen->write_buf_sz > BUFSIZ && screen->write_buf_used < BUFSIZ) { \
+                screen->write_buf_sz = BUFSIZ; \
+                screen->write_buf = PyMem_RawRealloc(screen->write_buf, screen->write_buf_sz); \
+                if (screen->write_buf == NULL) { fatal("Out of memory."); } \
+            } \
+            if (screen->write_buf_used) wakeup_io_loop(self, false); \
+            screen_mutex(unlock, write); \
+            break; \
+        } \
+    } \
+    children_mutex(unlock); \
+    return found;
+
 bool
 schedule_write_to_child(unsigned long id, unsigned int num, ...) {
-    ChildMonitor *self = the_monitor;
-    bool found = false;
-    const char *data;
-    size_t sz = 0;
     va_list ap;
-    va_start(ap, num);
-    for (unsigned int i = 0; i < num; i++) {
-        va_arg(ap, const char*);
-        sz += va_arg(ap, size_t);
-    }
-    va_end(ap);
-    children_mutex(lock);
-    for (size_t i = 0; i < self->count; i++) {
-        if (children[i].id == id) {
-            found = true;
-            Screen *screen = children[i].screen;
-            screen_mutex(lock, write);
-            size_t space_left = screen->write_buf_sz - screen->write_buf_used;
-            if (space_left < sz) {
-                if (screen->write_buf_used + sz > 100 * 1024 * 1024) {
-                    log_error("Too much data being sent to child with id: %lu, ignoring it", id);
-                    screen_mutex(unlock, write);
-                    break;
-                }
-                screen->write_buf_sz = screen->write_buf_used + sz;
-                screen->write_buf = PyMem_RawRealloc(screen->write_buf, screen->write_buf_sz);
-                if (screen->write_buf == NULL) { fatal("Out of memory."); }
-            }
-            va_start(ap, num);
-            for (unsigned int i = 0; i < num; i++) {
-                data = va_arg(ap, const char*);
-                size_t dsz = va_arg(ap, size_t);
-                memcpy(screen->write_buf + screen->write_buf_used, data, dsz);
-                screen->write_buf_used += dsz;
-            }
-            va_end(ap);
-            if (screen->write_buf_sz > BUFSIZ && screen->write_buf_used < BUFSIZ) {
-                screen->write_buf_sz = BUFSIZ;
-                screen->write_buf = PyMem_RawRealloc(screen->write_buf, screen->write_buf_sz);
-                if (screen->write_buf == NULL) { fatal("Out of memory."); }
-            }
-            if (screen->write_buf_used) wakeup_io_loop(self, false);
-            screen_mutex(unlock, write);
-            break;
-        }
-    }
-    children_mutex(unlock);
-    return found;
+#define get_next_arg(ap) data = va_arg(ap, const char*); szval = va_arg(ap, size_t);
+    schedule_write_to_child_generic(id, num, va_start, get_next_arg, va_end);
+#undef get_next_arg
+}
+
+bool
+schedule_write_to_child_python(unsigned long id, const char *prefix, PyObject *ap, const char *suffix) {
+    if (!PyTuple_Check(ap)) return false;
+    bool has_prefix = prefix && prefix[0], has_suffix = suffix && suffix[0];
+    const size_t extra = (has_prefix ? 1 : 0) + (has_suffix ? 1 : 0);
+    size_t num = PyTuple_GET_SIZE(ap) + extra;
+    Py_ssize_t pidx;
+#define py_start(ap, num) pidx = 0;
+#define py_end(ap) pidx = 0;
+#define get_next_arg(ap) { \
+    size_t pidxf = pidx++; \
+    if (pidxf == 0 && has_prefix) { data = prefix; szval = strlen(prefix); } \
+    else { \
+        if (has_prefix) pidxf--; \
+        if (has_suffix && pidxf >= (size_t)PyTuple_GET_SIZE(ap)) { data = suffix; szval = strlen(suffix); } \
+        else { \
+            PyObject *t = PyTuple_GET_ITEM(ap, pidxf); \
+            if (PyBytes_Check(t)) { data = PyBytes_AS_STRING(t); szval = PyBytes_GET_SIZE(t); } \
+            else { \
+                Py_ssize_t usz; \
+                data = PyUnicode_AsUTF8AndSize(t, &usz); szval = usz; \
+                if (!data) fatal("Failed to convert object to bytes in schedule_write_to_child_python"); \
+            } \
+        } \
+    } \
+}
+    schedule_write_to_child_generic(id, num, py_start, get_next_arg, py_end);
+#undef py_start
+#undef py_end
+#undef get_next_arg
 }
 
 static PyObject *
@@ -308,12 +343,12 @@ shutdown_monitor(ChildMonitor *self, PyObject *a UNUSED) {
 }
 
 static bool
-do_parse(ChildMonitor *self, Screen *screen, monotonic_t now) {
+do_parse(ChildMonitor *self, Screen *screen, monotonic_t now, bool flush) {
     bool input_read = false;
     screen_mutex(lock, read);
     if (screen->read_buf_sz || screen->pending_mode.used) {
         monotonic_t time_since_new_input = now - screen->new_input_at;
-        if (time_since_new_input >= OPT(input_delay)) {
+        if (flush || time_since_new_input >= OPT(input_delay)) {
             bool read_buf_full = screen->read_buf_sz >= READ_BUF_SZ;
             input_read = true;
             parse_func(screen, self->dump_callback, now);
@@ -338,7 +373,8 @@ parse_input(ChildMonitor *self) {
     children_mutex(lock);
     while (remove_queue_count) {
         remove_queue_count--;
-        remove_notify[remove_count] = remove_queue[remove_queue_count].id;
+        remove_notify[remove_count] = remove_queue[remove_queue_count];
+        INCREF_CHILD(remove_notify[remove_count]);
         remove_count++;
         FREE_CHILD(remove_queue[remove_queue_count]);
     }
@@ -382,13 +418,15 @@ parse_input(ChildMonitor *self) {
             Message *msg = msgs + i;
             PyObject *resp = NULL;
             if (msg->data) {
-                resp = PyObject_CallMethod(global_state.boss, "peer_message_received", "y#", msg->data, (int)msg->sz);
+                resp = PyObject_CallMethod(global_state.boss, "peer_message_received", "y#K", msg->data, (int)msg->sz, msg->peer_id);
                 free(msg->data);
                 if (!resp) PyErr_Print();
             }
-            if (resp && PyBytes_Check(resp)) send_response(msg->peer_id, PyBytes_AS_STRING(resp), PyBytes_GET_SIZE(resp));
-            else send_response(msg->peer_id, NULL, 0);
-            Py_CLEAR(resp);
+            if (resp) {
+                if (PyBytes_Check(resp)) send_response_to_peer(msg->peer_id, PyBytes_AS_STRING(resp), PyBytes_GET_SIZE(resp));
+                else if (resp == Py_None) send_response_to_peer(msg->peer_id, NULL, 0);
+                Py_CLEAR(resp);
+            } else send_response_to_peer(msg->peer_id, NULL, 0);
         }
         free(msgs); msgs = NULL;
     }
@@ -397,14 +435,16 @@ parse_input(ChildMonitor *self) {
         // must be done while no locks are held, since the locks are non-recursive and
         // the python function could call into other functions in this module
         remove_count--;
-        PyObject *t = PyObject_CallFunction(self->death_notify, "k", remove_notify[remove_count]);
+        if (remove_notify[remove_count].screen) do_parse(self, remove_notify[remove_count].screen, now, true);
+        PyObject *t = PyObject_CallFunction(self->death_notify, "k", remove_notify[remove_count].id);
         if (t == NULL) PyErr_Print();
         else Py_DECREF(t);
+        FREE_CHILD(remove_notify[remove_count]);
     }
 
     for (size_t i = 0; i < count; i++) {
         if (!scratch[i].needs_removal) {
-            if (do_parse(self, scratch[i].screen, now)) input_read = true;
+            if (do_parse(self, scratch[i].screen, now, false)) input_read = true;
         }
         DECREF_CHILD(scratch[i]);
     }
@@ -499,7 +539,7 @@ pyset_iutf8(ChildMonitor *self, PyObject *args) {
     for (size_t i = 0; i < self->count; i++) {
         if (children[i].id == window_id) {
             found = Py_True;
-            if (!set_iutf8(fds[EXTRA_FDS + i].fd, on & 1)) PyErr_SetFromErrno(PyExc_OSError);
+            if (!set_iutf8(children_fds[EXTRA_FDS + i].fd, on & 1)) PyErr_SetFromErrno(PyExc_OSError);
             break;
         }
     }
@@ -515,13 +555,18 @@ pyset_iutf8(ChildMonitor *self, PyObject *args) {
 
 extern void cocoa_update_menu_bar_title(PyObject*);
 
-static void
+static bool
+cursor_needs_render(Window *w) {
+    return w->cursor_visible_at_last_render != w->render_data.screen->cursor_render_info.is_visible || w->last_cursor_x != w->render_data.screen->cursor_render_info.x || w->last_cursor_y != w->render_data.screen->cursor_render_info.y || w->last_cursor_shape != w->render_data.screen->cursor_render_info.shape;
+}
+
+static bool
 collect_cursor_info(CursorRenderInfo *ans, Window *w, monotonic_t now, OSWindow *os_window) {
     ScreenRenderData *rd = &w->render_data;
     Cursor *cursor = rd->screen->cursor;
     ans->x = cursor->x; ans->y = cursor->y;
     ans->is_visible = false;
-    if (rd->screen->scrolled_by || !screen_is_cursor_visible(rd->screen)) return;
+    if (rd->screen->scrolled_by || !screen_is_cursor_visible(rd->screen)) return cursor_needs_render(w);
     monotonic_t time_since_start_blink = now - os_window->cursor_blink_zero_time;
     bool cursor_blinking = OPT(cursor_blink_interval) > 0 && !cursor->non_blinking && os_window->is_focused && (OPT(cursor_stop_blinking_after) == 0 || time_since_start_blink <= OPT(cursor_stop_blinking_after));
     bool do_draw_cursor = true;
@@ -534,12 +579,11 @@ collect_cursor_info(CursorRenderInfo *ans, Window *w, monotonic_t now, OSWindow
         monotonic_t delay = bucket - time_since_start_blink;
         set_maximum_wait(delay);
     }
-    if (!do_draw_cursor) { ans->is_visible = false; return; }
+    if (!do_draw_cursor) { ans->is_visible = false; return cursor_needs_render(w); }
     ans->is_visible = true;
-    ColorProfile *cp = rd->screen->color_profile;
     ans->shape = cursor->shape ? cursor->shape : OPT(cursor_shape);
-    ans->color = colorprofile_to_color(cp, cp->overridden.cursor_color, cp->configured.cursor_color);
     ans->is_focused = os_window->is_focused;
+    return cursor_needs_render(w);
 }
 
 static void
@@ -579,7 +623,7 @@ prepare_to_render_os_window(OSWindow *os_window, monotonic_t now, unsigned int *
 #define WD w->render_data
         if (w->visible && WD.screen) {
             *num_visible_windows += 1;
-            color_type window_bg = colorprofile_to_color(WD.screen->color_profile, WD.screen->color_profile->overridden.default_bg, WD.screen->color_profile->configured.default_bg);
+            color_type window_bg = colorprofile_to_color(WD.screen->color_profile, WD.screen->color_profile->overridden.default_bg, WD.screen->color_profile->configured.default_bg).rgb;
             if (*num_visible_windows == 1) first_window_bg = window_bg;
             if (first_window_bg != window_bg) *all_windows_have_same_bg = false;
             if (w->last_drag_scroll_at > 0) {
@@ -594,11 +638,17 @@ prepare_to_render_os_window(OSWindow *os_window, monotonic_t now, unsigned int *
             bool is_active_window = i == tab->active_window;
             if (is_active_window) {
                 *active_window_id = w->id;
-                collect_cursor_info(&WD.screen->cursor_render_info, w, now, os_window);
-                if (w->cursor_visible_at_last_render != WD.screen->cursor_render_info.is_visible || w->last_cursor_x != WD.screen->cursor_render_info.x || w->last_cursor_y != WD.screen->cursor_render_info.y || w->last_cursor_shape != WD.screen->cursor_render_info.shape) needs_render = true;
+                if (collect_cursor_info(&WD.screen->cursor_render_info, w, now, os_window)) needs_render = true;
                 set_os_window_title_from_window(w, os_window);
                 *active_window_bg = window_bg;
-            } else WD.screen->cursor_render_info.is_visible = false;
+            } else {
+                if (WD.screen->render_unfocused_cursor) {
+                    if (collect_cursor_info(&WD.screen->cursor_render_info, w, now, os_window)) needs_render = true;
+                    WD.screen->cursor_render_info.is_focused = false;
+                } else {
+                    WD.screen->cursor_render_info.is_visible = false;
+                }
+            }
             if (scan_for_animated_images) {
                 monotonic_t min_gap;
                 if (scan_active_animations(WD.screen->grman, now, &min_gap, true)) needs_render = true;
@@ -630,12 +680,12 @@ render_os_window(OSWindow *os_window, unsigned int active_window_id, color_type
         draw_borders(br->vao_idx, br->num_border_rects, br->rect_buf, br->is_dirty, os_window->viewport_width, os_window->viewport_height, active_window_bg, num_visible_windows, all_windows_have_same_bg, os_window);
         br->is_dirty = false;
     }
-    if (TD.screen && os_window->num_tabs >= OPT(tab_bar_min_tabs)) draw_cells(TD.vao_idx, 0, TD.xstart, TD.ystart, TD.dx * x_ratio, TD.dy * y_ratio, TD.screen, os_window, true, false);
+    if (TD.screen && os_window->num_tabs >= OPT(tab_bar_min_tabs)) draw_cells(TD.vao_idx, 0, &TD, x_ratio, y_ratio, os_window, true, false, NULL);
     for (unsigned int i = 0; i < tab->num_windows; i++) {
         Window *w = tab->windows + i;
         if (w->visible && WD.screen) {
             bool is_active_window = i == tab->active_window;
-            draw_cells(WD.vao_idx, WD.gvao_idx, WD.xstart, WD.ystart, WD.dx * x_ratio, WD.dy * y_ratio, WD.screen, os_window, is_active_window, true);
+            draw_cells(WD.vao_idx, WD.gvao_idx, &WD, x_ratio, y_ratio, os_window, is_active_window, true, w);
             if (WD.screen->start_visual_bell_at != 0) {
                 set_maximum_wait(OPT(repaint_delay));
             }
@@ -891,8 +941,12 @@ process_pending_resizes(monotonic_t now) {
 
 static void
 close_os_window(ChildMonitor *self, OSWindow *os_window) {
+    int w = os_window->window_width, h = os_window->window_height;
+    if (os_window->before_fullscreen.is_set && is_os_window_fullscreen(os_window)) {
+        w = os_window->before_fullscreen.w; h = os_window->before_fullscreen.h;
+    }
     destroy_os_window(os_window);
-    call_boss(on_os_window_closed, "Kii", os_window->id, os_window->window_width, os_window->window_height);
+    call_boss(on_os_window_closed, "Kii", os_window->id, w, h);
     for (size_t t=0; t < os_window->num_tabs; t++) {
         Tab *tab = os_window->tabs + t;
         for (size_t w = 0; w < tab->num_windows; w++) mark_child_for_close(self, tab->windows[w].id);
@@ -1080,8 +1134,8 @@ add_children(ChildMonitor *self) {
         add_queue_count--;
         children[self->count] = add_queue[add_queue_count];
         add_queue[add_queue_count] = EMPTY_CHILD;
-        fds[EXTRA_FDS + self->count].fd = children[self->count].fd;
-        fds[EXTRA_FDS + self->count].events = POLLIN;
+        children_fds[EXTRA_FDS + self->count].fd = children[self->count].fd;
+        children_fds[EXTRA_FDS + self->count].events = POLLIN;
         self->count++;
     }
 }
@@ -1117,11 +1171,11 @@ remove_children(ChildMonitor *self) {
                 remove_queue[remove_queue_count] = children[i];
                 remove_queue_count++;
                 children[i] = EMPTY_CHILD;
-                fds[EXTRA_FDS + i].fd = -1;
+                children_fds[EXTRA_FDS + i].fd = -1;
                 size_t num_to_right = self->count - 1 - i;
                 if (num_to_right > 0) {
                     memmove(children + i, children + i + 1, num_to_right * sizeof(Child));
-                    memmove(fds + EXTRA_FDS + i, fds + EXTRA_FDS + i + 1, num_to_right * sizeof(struct pollfd));
+                    memmove(children_fds + EXTRA_FDS + i, children_fds + EXTRA_FDS + i + 1, num_to_right * sizeof(struct pollfd));
                 }
             }
         }
@@ -1272,28 +1326,28 @@ io_loop(void *data) {
         add_children(self);
         children_mutex(unlock);
         data_received = false;
-        for (i = 0; i < self->count + EXTRA_FDS; i++) fds[i].revents = 0;
+        for (i = 0; i < self->count + EXTRA_FDS; i++) children_fds[i].revents = 0;
         for (i = 0; i < self->count; i++) {
             screen = children[i].screen;
             /* printf("i:%lu id:%lu fd: %d read_buf_sz: %lu write_buf_used: %lu\n", i, children[i].id, children[i].fd, screen->read_buf_sz, screen->write_buf_used); */
             screen_mutex(lock, read); screen_mutex(lock, write);
-            fds[EXTRA_FDS + i].events = (screen->read_buf_sz < READ_BUF_SZ ? POLLIN : 0) | (screen->write_buf_used ? POLLOUT  : 0);
+            children_fds[EXTRA_FDS + i].events = (screen->read_buf_sz < READ_BUF_SZ ? POLLIN : 0) | (screen->write_buf_used ? POLLOUT  : 0);
             screen_mutex(unlock, read); screen_mutex(unlock, write);
         }
         if (has_pending_wakeups) {
             now = monotonic();
             monotonic_t time_delta = OPT(input_delay) - (now - last_main_loop_wakeup_at);
-            if (time_delta >= 0) ret = poll(fds, self->count + EXTRA_FDS, monotonic_t_to_ms(time_delta));
+            if (time_delta >= 0) ret = poll(children_fds, self->count + EXTRA_FDS, monotonic_t_to_ms(time_delta));
             else ret = 0;
         } else {
-            ret = poll(fds, self->count + EXTRA_FDS, -1);
+            ret = poll(children_fds, self->count + EXTRA_FDS, -1);
         }
         if (ret > 0) {
-            if (fds[0].revents && POLLIN) drain_fd(fds[0].fd); // wakeup
-            if (fds[1].revents && POLLIN) {
+            if (children_fds[0].revents && POLLIN) drain_fd(children_fds[0].fd); // wakeup
+            if (children_fds[1].revents && POLLIN) {
                 SignalSet ss = {0};
                 data_received = true;
-                read_signals(fds[1].fd, handle_signal, &ss);
+                read_signals(children_fds[1].fd, handle_signal, &ss);
                 if (ss.kill_signal || ss.reload_config) {
                     children_mutex(lock);
                     if (ss.kill_signal) kill_signal_received = true;
@@ -1303,9 +1357,9 @@ io_loop(void *data) {
                 if (ss.child_died) reap_children(self, OPT(close_on_child_death));
             }
             for (i = 0; i < self->count; i++) {
-                if (fds[EXTRA_FDS + i].revents & (POLLIN | POLLHUP)) {
+                if (children_fds[EXTRA_FDS + i].revents & (POLLIN | POLLHUP)) {
                     data_received = true;
-                    has_more = read_bytes(fds[EXTRA_FDS + i].fd, children[i].screen);
+                    has_more = read_bytes(children_fds[EXTRA_FDS + i].fd, children[i].screen);
                     if (!has_more) {
                         // child is dead
                         children_mutex(lock);
@@ -1313,10 +1367,10 @@ io_loop(void *data) {
                         children_mutex(unlock);
                     }
                 }
-                if (fds[EXTRA_FDS + i].revents & POLLOUT) {
+                if (children_fds[EXTRA_FDS + i].revents & POLLOUT) {
                     write_to_child(children[i].fd, children[i].screen);
                 }
-                if (fds[EXTRA_FDS + i].revents & POLLNVAL) {
+                if (children_fds[EXTRA_FDS + i].revents & POLLNVAL) {
                     // fd was closed
                     children_mutex(lock);
                     children[i].needs_removal = true;
@@ -1326,7 +1380,7 @@ io_loop(void *data) {
             }
 #ifdef DEBUG_POLL_EVENTS
             for (i = 0; i < self->count + EXTRA_FDS; i++) {
-#define P(w) if (fds[i].revents & w) printf("i:%lu %s\n", i, #w);
+#define P(w) if (children_fds[i].revents & w) printf("i:%lu %s\n", i, #w);
                 P(POLLIN); P(POLLPRI); P(POLLOUT); P(POLLERR); P(POLLHUP); P(POLLNVAL);
 #undef P
             }
@@ -1587,7 +1641,7 @@ end:
 }
 
 static void
-send_response(id_type peer_id, const char *msg, size_t msg_sz) {
+send_response_to_peer(id_type peer_id, const char *msg, size_t msg_sz) {
     bool wakeup = false;
     talk_mutex(lock);
     for (size_t i = 0; i < talk_data.num_peers; i++) {
@@ -1662,11 +1716,22 @@ cocoa_set_menubar_title(PyObject *self UNUSED, PyObject *args UNUSED) {
     Py_RETURN_NONE;
 }
 
+static PyObject*
+
+send_data_to_peer(PyObject *self UNUSED, PyObject *args) {
+    char * msg; Py_ssize_t sz;
+    unsigned long long peer_id;
+    if (!PyArg_ParseTuple(args, "Ks#", &peer_id, &msg, &sz)) return NULL;
+    send_response_to_peer(peer_id, msg, sz);
+    Py_RETURN_NONE;
+}
+
 static PyMethodDef module_methods[] = {
     METHODB(safe_pipe, METH_VARARGS),
     {"add_timer", (PyCFunction)add_python_timer, METH_VARARGS, ""},
     {"remove_timer", (PyCFunction)remove_python_timer, METH_VARARGS, ""},
     METHODB(monitor_pid, METH_VARARGS),
+    METHODB(send_data_to_peer, METH_VARARGS),
     METHODB(cocoa_set_menubar_title, METH_VARARGS),
     {NULL}  /* Sentinel */
 };
