@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 
@@ -14,6 +13,7 @@ LaunchCLIOptions = AskCLIOptions = ClipboardCLIOptions = DiffCLIOptions = CLIOpt
 HintsCLIOptions = IcatCLIOptions = PanelCLIOptions = ResizeCLIOptions = CLIOptions
 ErrorCLIOptions = UnicodeCLIOptions = RCOptions = RemoteFileCLIOptions = CLIOptions
 QueryTerminalCLIOptions = BroadcastCLIOptions = ShowKeyCLIOptions = CLIOptions
+ThemesCLIOptions = TransferCLIOptions = CLIOptions
 
 
 def generate_stub() -> None:
@@ -31,7 +31,7 @@ def generate_stub() -> None:
     do(options_spec(), 'LaunchCLIOptions')
 
     from .remote_control import global_options_spec
-    do(global_options_spec(), 'RCOptions', extra_fields=['no_command_response: typing.Optional[bool]'])
+    do(global_options_spec(), 'RCOptions')
 
     from kittens.ask.main import option_text
     do(option_text(), 'AskCLIOptions')
@@ -72,6 +72,12 @@ def generate_stub() -> None:
     from kittens.unicode_input.main import OPTIONS
     do(OPTIONS(), 'UnicodeCLIOptions')
 
+    from kittens.themes.main import OPTIONS
+    do(OPTIONS(), 'ThemesCLIOptions')
+
+    from kittens.transfer.main import option_text as OPTIONS
+    do(OPTIONS(), 'TransferCLIOptions')
+
     from kitty.rc.base import all_command_names, command_for_name
     for cmd_name in all_command_names():
         cmd = command_for_name(cmd_name)
