@@ -18,7 +18,6 @@
 // Needed for _NSGetProgname
 #include <crt_externs.h>
 #include <objc/runtime.h>
-#include <xlocale.h>
 
 #if (MAC_OS_X_VERSION_MAX_ALLOWED < 101200)
 #define NSWindowStyleMaskResizable NSResizableWindowMask
@@ -26,6 +25,10 @@
 #define NSEventModifierFlagCommand NSCommandKeyMask
 #define NSEventModifierFlagControl NSControlKeyMask
 #endif
+#if (MAC_OS_X_VERSION_MAX_ALLOWED < 110000)
+#define UNNotificationPresentationOptionList (1 << 3)
+#define UNNotificationPresentationOptionBanner (1 << 4)
+#endif
 
 typedef int CGSConnectionID;
 typedef int CGSWindowID;
@@ -223,6 +226,15 @@ cocoa_send_notification(PyObject *self UNUSED, PyObject *args) {
 @end
 
 @implementation NotificationDelegate
+    - (void)userNotificationCenter:(UNUserNotificationCenter *)center
+            willPresentNotification:(UNNotification *)notification
+            withCompletionHandler:(void (^)(UNNotificationPresentationOptions))completionHandler {
+        UNNotificationPresentationOptions options = UNNotificationPresentationOptionSound;
+        if (@available(macOS 11.0, *)) options |= UNNotificationPresentationOptionList | UNNotificationPresentationOptionBanner;
+        else options |= (1 << 2); // UNNotificationPresentationOptionAlert avoid deprecated warning
+        completionHandler(options);
+    }
+
     - (void)userNotificationCenter:(UNUserNotificationCenter *)center
             didReceiveNotificationResponse:(UNNotificationResponse *)response
             withCompletionHandler:(void (^)(void))completionHandler {
@@ -248,6 +260,7 @@ schedule_notification(const char *identifier, const char *title, const char *bod
     if (title) content.title = @(title);
     if (body) content.body = @(body);
     if (subtitle) content.subtitle = @(subtitle);
+    content.sound = [UNNotificationSound defaultSound];
     // Deliver the notification
     static unsigned long counter = 1;
     UNNotificationRequest* request = [
@@ -307,7 +320,9 @@ cocoa_send_notification(PyObject *self UNUSED, PyObject *args) {
     if (!center.delegate) center.delegate = [[NotificationDelegate alloc] init];
     queue_notification(identifier, title, body, subtitle);
 
-    [center requestAuthorizationWithOptions:(UNAuthorizationOptionAlert)
+    // The badge permission needs to be requested as well, even though it is not used,
+    // otherwise macOS refuses to show the preference checkbox for enable/disable notification sound.
+    [center requestAuthorizationWithOptions:(UNAuthorizationOptionAlert | UNAuthorizationOptionSound | UNAuthorizationOptionBadge)
         completionHandler:^(BOOL granted, NSError * _Nullable error) {
             if (error != nil) {
                 log_error("Failed to request permission for showing notification: %s", [[error localizedDescription] UTF8String]);
@@ -381,7 +396,16 @@ cocoa_create_global_menu(void) {
                 keyEquivalent:@""];
     [appMenu addItem:[NSMenuItem separatorItem]];
     MENU_ITEM(appMenu, @"Preferences…", edit_config_file);
-    MENU_ITEM(appMenu, @"Reload preferences", reload_config);
+    MENU_ITEM(appMenu, @"Reload Preferences", reload_config);
+    [appMenu addItem:[NSMenuItem separatorItem]];
+
+    NSMenu* servicesMenu = [[NSMenu alloc] init];
+    [NSApp setServicesMenu:servicesMenu];
+    [[appMenu addItemWithTitle:@"Services"
+                        action:NULL
+                 keyEquivalent:@""] setSubmenu:servicesMenu];
+    [servicesMenu release];
+    [appMenu addItem:[NSMenuItem separatorItem]];
 
     [appMenu addItemWithTitle:[NSString stringWithFormat:@"Hide %@", app_name]
                        action:@selector(hide:)
@@ -395,15 +419,6 @@ cocoa_create_global_menu(void) {
                 keyEquivalent:@""];
     [appMenu addItem:[NSMenuItem separatorItem]];
 
-    NSMenu* servicesMenu = [[NSMenu alloc] init];
-    [NSApp setServicesMenu:servicesMenu];
-    [[appMenu addItemWithTitle:@"Services"
-                        action:NULL
-                 keyEquivalent:@""] setSubmenu:servicesMenu];
-    [servicesMenu release];
-
-    [appMenu addItem:[NSMenuItem separatorItem]];
-
     [[appMenu addItemWithTitle:@"Secure Keyboard Entry"
                         action:@selector(toggleSecureInput:)
                  keyEquivalent:@"s"]
@@ -471,7 +486,7 @@ cocoa_create_global_menu(void) {
                 keyEquivalent:@""];
     NSMenu* helpMenu = [[NSMenu alloc] initWithTitle:@"Help"];
     [helpMenuItem setSubmenu:helpMenu];
-    [[helpMenu addItemWithTitle:[NSString stringWithFormat:@"Visit %@ website", app_name]
+    [[helpMenu addItemWithTitle:[NSString stringWithFormat:@"Visit %@ Website", app_name]
                          action:@selector(open_kitty_website_url:)
                   keyEquivalent:@"?"]
                       setTarget:global_menu_target];
@@ -574,12 +589,7 @@ cocoa_get_lang(PyObject UNUSED *self) {
         locale = [[NSLocale currentLocale] localeIdentifier];
     }
     if (!locale) { Py_RETURN_NONE; }
-    // Make sure the locale value is valid, that is it can be used
-    // to construct an actual locale
     const char* locale_utf8 = [locale UTF8String];
-    locale_t test_locale = newlocale(LC_ALL_MASK, locale_utf8, NULL);
-    if (!test_locale) { Py_RETURN_NONE; }
-    freelocale(test_locale);
     return Py_BuildValue("s", locale_utf8);
 
     } // autoreleasepool
@@ -609,6 +619,16 @@ cocoa_set_activation_policy(bool hide_from_tasks) {
     [NSApp setActivationPolicy:(hide_from_tasks ? NSApplicationActivationPolicyAccessory : NSApplicationActivationPolicyRegular)];
 }
 
+void
+cocoa_set_titlebar_appearance(void *w, unsigned int theme)
+{
+    if (!theme) return;
+    @autoreleasepool {
+        NSWindow *window = (NSWindow*)w;
+        [window setAppearance:[NSAppearance appearanceNamed:((theme == 2) ? NSAppearanceNameVibrantDark : NSAppearanceNameVibrantLight)]];
+    } // autoreleasepool
+}
+
 void
 cocoa_set_titlebar_color(void *w, color_type titlebar_color)
 {
@@ -639,12 +659,16 @@ cocoa_set_titlebar_color(void *w, color_type titlebar_color)
     } // autoreleasepool
 }
 
+static NSSound *beep_sound = nil;
+
 static void
 cleanup() {
     @autoreleasepool {
 
     if (dockMenu) [dockMenu release];
     dockMenu = nil;
+    if (beep_sound) [beep_sound release];
+    beep_sound = nil;
 
 #ifndef KITTY_USE_DEPRECATED_MACOS_NOTIFICATION_API
     drain_pending_notifications(NO);
@@ -668,8 +692,15 @@ cocoa_hide_window_title(void *w)
 }
 
 void
-cocoa_system_beep(void) {
-    NSBeep();
+cocoa_system_beep(const char *path) {
+    if (!path) { NSBeep(); return; }
+    static const char *beep_path = NULL;
+    if (beep_path != path) {
+        if (beep_sound) [beep_sound release];
+        beep_sound = [[NSSound alloc] initWithContentsOfFile:@(path) byReference:YES];
+    }
+    if (beep_sound) [beep_sound play];
+    else NSBeep();
 }
 
 static PyMethodDef module_methods[] = {
