@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -18,6 +17,7 @@ from .cli import (
     OptionDict, OptionSpecSeq, options_for_completion, parse_option_spec,
     prettify
 )
+from .constants import shell_integration_dir
 from .fast_data_types import truncate_point_for_length, wcswidth
 from .rc.base import all_command_names, command_for_name
 from .shell import options_for_cmd
@@ -43,8 +43,9 @@ taking the results from kitty's completion system and converting
 them into something your shell will understand.
 '''
 
-parsers: Dict[str, Callable] = {}
-serializers: Dict[str, Callable] = {}
+parsers: Dict[str, 'ParserFunc'] = {}
+serializers: Dict[str, 'SerializerFunc'] = {}
+shell_state: Dict[str, str] = {}
 
 
 class MatchGroup:
@@ -63,6 +64,9 @@ class MatchGroup:
     def __iter__(self) -> Iterator[str]:
         return iter(self.mdict)
 
+    def __bool__(self) -> bool:
+        return bool(self.mdict)
+
     def transformed_words(self) -> Iterator[str]:
         for w in self:
             yield self.word_transforms.get(w, w)
@@ -77,6 +81,12 @@ class MatchGroup:
     def values(self) -> Iterator[str]:
         return iter(self.mdict.values())
 
+    def add_prefix(self, prefix: str) -> None:
+        nmap = {k: prefix + k for k in self.mdict}
+        for k, nk in nmap.items():
+            self.word_transforms[nk] = self.word_transforms.pop(k, k)
+        self.mdict = {prefix + k: v for k, v in self.mdict.items()}
+
 
 def debug(*a: Any, **kw: Any) -> None:
     from kittens.tui.loop import debug_write
@@ -117,6 +127,10 @@ class Completions:
         self.match_groups[name] = m = MatchGroup(x, trailing_space, is_files, word_transforms)
         return m
 
+    def add_prefix(self, prefix: str) -> None:
+        for mg in self.match_groups.values():
+            mg.add_prefix(prefix)
+
 
 @run_once
 def remote_control_command_names() -> Tuple[str, ...]:
@@ -126,6 +140,11 @@ def remote_control_command_names() -> Tuple[str, ...]:
 # Shell specific code {{{
 
 
+def load_fish2_completion() -> str:
+    with open(os.path.join(shell_integration_dir, 'fish', 'vendor_completions.d', 'kitty.fish')) as f:
+        return f.read()
+
+
 completion_scripts = {
     'zsh': '''#compdef kitty
 
@@ -138,7 +157,7 @@ _kitty() {
     fi
 }
 compdef _kitty kitty
-''',
+'''.__str__,
     'bash': '''
 _kitty_completions() {
     local src
@@ -152,7 +171,7 @@ _kitty_completions() {
 }
 
 complete -o nospace -F _kitty_completions kitty
-''',
+'''.__str__,
     'fish': '''
 function __kitty_completions
     # Send all words up to the one before the cursor
@@ -160,7 +179,8 @@ function __kitty_completions
 end
 
 complete -f -c kitty -a "(__kitty_completions)"
-''',
+'''.__str__,
+    'fish2': load_fish2_completion,
 }
 
 ParseResult = Tuple[List[str], bool]
@@ -182,6 +202,18 @@ def output_serializer(func: SerializerFunc) -> SerializerFunc:
 
 @input_parser
 def zsh_input_parser(data: str) -> ParseResult:
+    matcher = shell_state.get('_matcher', '')
+    q = matcher.lower().split(':', maxsplit=1)[0]
+    if q in ('l', 'r', 'b', 'e'):
+        # this is zsh anchor based matching
+        # https://zsh.sourceforge.io/Doc/Release/Completion-Widgets.html#Completion-Matching-Control
+        # can be specified with matcher-list and some systems do it by default,
+        # for example, Debian, which adds the following to zshrc
+        # zstyle ':completion:*' matcher-list '' 'm:{a-z}={A-Z}' 'm:{a-zA-Z}={A-Za-z}' 'r:|[._-]=* r:|=* l:|=*'
+        # For some reason that I dont have the
+        # time/interest to figure out, returning completion candidates for
+        # these matcher types break completion, so just abort in this case.
+        raise SystemExit(1)
     new_word = data.endswith('\n\n')
     words = data.rstrip().splitlines()
     return words, new_word
@@ -199,12 +231,21 @@ def fish_input_parser(data: str) -> ParseResult:
     return data.rstrip().splitlines(), True
 
 
+@input_parser
+def fish2_input_parser(data: str) -> ParseResult:
+    return bash_input_parser(data)
+
+
 @output_serializer
 def zsh_output_serializer(ans: Completions) -> str:
     lines = []
 
-    screen = screen_size_function(sys.stderr.fileno())()
-    width = screen.cols
+    try:
+        screen = screen_size_function(sys.stderr.fileno())()
+    except OSError:
+        width = 80
+    else:
+        width = screen.cols
 
     def fmt_desc(word: str, desc: str, max_word_len: int) -> Iterator[str]:
         if not desc:
@@ -234,10 +275,13 @@ def zsh_output_serializer(ans: Completions) -> str:
             cmd += ['-S', '""']
         if matches.is_files:
             cmd.append('-f')
-            common_prefix = os.path.commonprefix(tuple(matches))
-            if common_prefix:
-                cmd.extend(('-p', shlex.quote(common_prefix)))
-                matches = MatchGroup({k[len(common_prefix):]: v for k, v in matches.items()})
+            allm = tuple(matches)
+            if len(allm) > 1:
+                common_prefix = os.path.commonprefix(allm)
+                if common_prefix and '/' in common_prefix:
+                    common_prefix = os.path.dirname(common_prefix) + '/'
+                    cmd.extend(('-p', shlex.quote(common_prefix)))
+                    matches = MatchGroup({k[len(common_prefix):]: v for k, v in matches.items()})
         has_descriptions = any(matches.values())
         if has_descriptions or matches.word_transforms:
             lines.append('compdescriptions=(')
@@ -275,7 +319,7 @@ def bash_output_serializer(ans: Completions) -> str:
         for word in matches:
             if matches.trailing_space:
                 word += ' '
-            lines.append('COMPREPLY+=({})'.format(shlex.quote(word)))
+            lines.append(f'COMPREPLY+=({shlex.quote(word)})')
     # debug('\n'.join(lines))
     return '\n'.join(lines)
 
@@ -283,11 +327,25 @@ def bash_output_serializer(ans: Completions) -> str:
 @output_serializer
 def fish_output_serializer(ans: Completions) -> str:
     lines = []
-    for matches in ans.match_groups.values():
+    for description, matches in ans.match_groups.items():
         for word in matches:
-            lines.append(shlex.quote(word))
+            lines.append(word.replace('\n', ' '))
     # debug('\n'.join(lines))
     return '\n'.join(lines)
+
+
+@output_serializer
+def fish2_output_serializer(ans: Completions) -> str:
+    lines = []
+    for description, matches in ans.match_groups.items():
+        for word, desc in matches.items():
+            q = word
+            if desc:
+                q = f'{q}\t{desc}'
+            lines.append(q.replace('\n', ' '))
+    # debug('\n'.join(lines))
+    return '\n'.join(lines)
+
 # }}}
 
 
@@ -303,6 +361,8 @@ def completions_for_first_word(ans: Completions, prefix: str, entry_points: Iter
 
 
 def kitty_cli_opts(ans: Completions, prefix: Optional[str] = None) -> None:
+    if not prefix:
+        return
     matches = {}
     for opt in options_for_completion():
         if isinstance(opt, str):
@@ -337,16 +397,15 @@ def complete_kitty_cli_arg(ans: Completions, opt: Optional[OptionDict], prefix:
     elif dest == 'watcher':
         complete_files_and_dirs(ans, prefix, files_group_name='Watcher files')
     elif dest == 'directory':
-        complete_files_and_dirs(ans, prefix, files_group_name='Directories', predicate=os.path.isdir)
-    elif dest == 'start_as':
-        k = 'Start as'
-        ans.add_match_group(k, {x: x for x in 'normal,fullscreen,maximized,minimized'.split(',') if x.startswith(prefix)}, trailing_space=False)
+        complete_dirs(ans, prefix)
     elif dest == 'listen_on':
         if ':' not in prefix:
             k = 'Address type'
             ans.add_match_group(k, {x: x for x in ('unix:', 'tcp:') if x.startswith(prefix)}, trailing_space=False)
         elif prefix.startswith('unix:') and not prefix.startswith('@'):
             complete_files_and_dirs(ans, prefix[len('unix:'):], files_group_name='UNIX sockets', add_prefix='unix:')
+    else:
+        complete_basic_option_args(ans, opt, prefix)
 
 
 CompleteArgsFunc = Callable[[Completions, Optional[OptionDict], str, Delegate], None]
@@ -364,12 +423,28 @@ def complete_alias_map(
     last_word = words[-1] if words else ''
     for i, w in enumerate(words):
         if expecting_arg:
+            prev_word = '' if i == 0 else words[i-1]
+            if w == '=' and i > 0 and prev_word.startswith('--') and prev_word != '--':
+                if w is not last_word:
+                    continue
+                long_opt = option_map.get(prev_word)
+                if long_opt is not None and complete_args is not None:
+                    complete_args(ans, long_opt, '', Delegate())
+                    return
             if w is last_word and not new_word:
                 if opt is not None and complete_args is not None:
                     complete_args(ans, opt, w, Delegate())
                 return
             expecting_arg = False
             continue
+        if w is last_word and not new_word and w.startswith('--') and w != '--':
+            parts = w.split('=', 1)
+            if len(parts) == 2:
+                long_opt = option_map.get(parts[0])
+                if long_opt is not None and complete_args is not None:
+                    complete_args(ans, long_opt, parts[1], Delegate())
+                    ans.add_prefix(parts[0] + '=')
+                return
         opt = option_map.get(w)
         if w is last_word and not new_word:
             if w.startswith('-'):
@@ -381,6 +456,8 @@ def complete_alias_map(
         if opt is None:
             if complete_args is not None:
                 complete_args(ans, None, '' if new_word else last_word, Delegate(words, i, new_word))
+            if w.startswith('--') and '=' in w:
+                continue
             return  # some non-option word encountered
         expecting_arg = not opt.get('type', '').startswith('bool-')
     if expecting_arg:
@@ -436,7 +513,7 @@ def path_completion(prefix: str = '') -> Tuple[List[str], List[str]]:
     src = os.path.expandvars(os.path.expanduser(base))
     src_prefix = os.path.abspath(os.path.expandvars(os.path.expanduser(prefix))) if prefix else ''
     try:
-        items: Iterable[os.DirEntry] = os.scandir(src)
+        items: Iterable['os.DirEntry[str]'] = os.scandir(src)
     except FileNotFoundError:
         items = ()
     for x in items:
@@ -474,6 +551,17 @@ def complete_files_and_dirs(
         ans.add_match_group(files_group_name, files, is_files=True)
 
 
+def complete_basic_option_args(ans: Completions, opt: OptionDict, prefix: str) -> None:
+    if opt['choices']:
+        ans.add_match_group(f'Choices for {opt["dest"]}', tuple(k for k in opt['choices'] if k.startswith(prefix)))
+
+
+def complete_dirs(ans: Completions, prefix: str = '') -> None:
+    dirs, files_ = path_completion(prefix or '')
+    if dirs:
+        ans.add_match_group('Directories', dirs, trailing_space=False, is_files=True)
+
+
 def complete_icat_args(ans: Completions, opt: Optional[OptionDict], prefix: str, unknown_args: Delegate) -> None:
     from .guess_mime_type import guess_type
 
@@ -485,6 +573,18 @@ def complete_icat_args(ans: Completions, opt: Optional[OptionDict], prefix: str,
 
     if opt is None:
         complete_files_and_dirs(ans, prefix, 'Images', icat_file_predicate)
+    else:
+        complete_basic_option_args(ans, opt, prefix)
+
+
+def complete_themes_args(ans: Completions, opt: Optional[OptionDict], prefix: str, unknown_args: Delegate) -> None:
+    if opt is None:
+        from kittens.themes.collection import load_themes
+        themes = load_themes(cache_age=-1, ignore_no_cache=True)
+        names = tuple(t.name for t in themes if t.name.startswith(prefix))
+        ans.add_match_group('Themes', names)
+    else:
+        complete_basic_option_args(ans, opt, prefix)
 
 
 def remote_files_completer(name: str, matchers: Tuple[str, ...]) -> CompleteArgsFunc:
@@ -500,6 +600,8 @@ def remote_files_completer(name: str, matchers: Tuple[str, ...]) -> CompleteArgs
 
         if opt is None:
             complete_files_and_dirs(ans, prefix, name, predicate)
+        else:
+            complete_basic_option_args(ans, opt, prefix)
     return complete_files_map
 
 
@@ -509,6 +611,8 @@ def remote_args_completer(title: str, words: Iterable[str]) -> CompleteArgsFunc:
     def complete_names_for_arg(ans: Completions, opt: Optional[OptionDict], prefix: str, unknown_args: Delegate) -> None:
         if opt is None:
             ans.add_match_group(title, {c: '' for c in items if c.startswith(prefix)})
+        else:
+            complete_basic_option_args(ans, opt, prefix)
 
     return complete_names_for_arg
 
@@ -522,6 +626,8 @@ def complete_diff_args(ans: Completions, opt: Optional[OptionDict], prefix: str,
         complete_files_and_dirs(ans, prefix, 'Files')
     elif opt['dest'] == 'config':
         complete_files_and_dirs(ans, prefix, 'Config Files', config_file_predicate)
+    else:
+        complete_basic_option_args(ans, opt, prefix)
 
 
 def complete_kitten(ans: Completions, kitten: str, words: Sequence[str], new_word: bool) -> None:
@@ -541,13 +647,15 @@ def complete_kitten(ans: Completions, kitten: str, words: Sequence[str], new_wor
     options = cd['options']()
     seq = parse_option_spec(options)[0]
     option_map = {}
-    for opt in seq:
-        if not isinstance(opt, str):
-            for alias in opt['aliases']:
-                option_map[alias] = opt
+    if not new_word:
+        for opt in seq:
+            if not isinstance(opt, str):
+                for alias in opt['aliases']:
+                    option_map[alias] = opt
     complete_alias_map(ans, words, new_word, option_map, {
         'icat': complete_icat_args,
         'diff': complete_diff_args,
+        'themes': complete_themes_args,
     }.get(kitten))
 
 
@@ -557,6 +665,9 @@ def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterabl
         return ans
     words = words[1:]
     if not words or (len(words) == 1 and not new_word):
+        if words and words[0].startswith('--') and '=' in words[0]:
+            complete_cli(ans, words, new_word, options_for_completion(), complete_kitty_cli_arg)
+            return ans
         prefix = words[0] if words else ''
         completions_for_first_word(ans, prefix, entry_points, namespaced_entry_points)
         kitty_cli_opts(ans, prefix)
@@ -606,7 +717,7 @@ def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterabl
 
 
 def setup(cstyle: str) -> None:
-    print(completion_scripts[cstyle])
+    print(completion_scripts[cstyle]())
 
 
 def main(args: Sequence[str], entry_points: Iterable[str], namespaced_entry_points: Iterable[str]) -> None:
@@ -616,11 +727,16 @@ def main(args: Sequence[str], entry_points: Iterable[str], namespaced_entry_poin
     if cstyle == 'setup':
         return setup(args[1])
     data = sys.stdin.read()
+    shell_state.clear()
+    for x in args[1:]:
+        parts = x.split('=', maxsplit=1)
+        if len(parts) == 2:
+            shell_state[parts[0]] = parts[1]
     try:
         parser = parsers[cstyle]
         serializer = serializers[cstyle]
     except KeyError:
-        raise SystemExit('Unknown completion style: {}'.format(cstyle))
+        raise SystemExit(f'Unknown completion style: {cstyle}')
     words, new_word = parser(data)
     ans = find_completions(words, new_word, entry_points, namespaced_entry_points)
     print(serializer(ans), end='')
