@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>
 
 
@@ -13,9 +12,10 @@ from typing import (
 )
 
 from kitty.conf.types import Definition, MultiOption, Option, unset
+from kitty.types import _T
 
 
-def chunks(lst: List, n: int) -> Iterator[List]:
+def chunks(lst: List[_T], n: int) -> Iterator[List[_T]]:
     for i in range(0, len(lst), n):
         yield lst[i:i + n]
 
@@ -52,7 +52,7 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
         ans = ans.replace('NoneType', 'None')
         return ans
 
-    def option_type_data(option: Union[Option, MultiOption]) -> Tuple[Callable, str]:
+    def option_type_data(option: Union[Option, MultiOption]) -> Tuple[Callable[[Any], Any], str]:
         func = option.parser_func
         if func.__module__ == 'builtins':
             return func, func.__name__
@@ -105,12 +105,12 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
             t(f'        ans[{option.name!r}] = {func.__name__}(val)')
             tc_imports.add((func.__module__, func.__name__))
             cnum = int(option.name[5:])
-            color_table[cnum] = '0x{:06x}'.format(func(option.defval_as_string).__int__())
+            color_table[cnum] = f'0x{func(option.defval_as_string).__int__():06x}'
             continue
         else:
             func, typ = option_type_data(option)
             try:
-                params = inspect.signature(func).parameters
+                params = dict(inspect.signature(func).parameters)
             except Exception:
                 params = {}
             if 'dict_with_parse_results' in params:
@@ -165,7 +165,8 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
             resolve_import(imp)
         for fname, ftype in action.fields.items():
             ftype = resolve_import(ftype)
-            a(f'    {fname}: {ftype} = ' '{}')
+            fval = f'{ftype}()' if ftype == 'AliasMap' else '{}'
+            a(f'    {fname}: {ftype} = {fval}')
         parser_function_declaration(aname)
         t(f'        for k in {func.__name__}(val):')
         t(f'            ans[{aname!r}].append(k)')
@@ -173,7 +174,7 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
 
     if defn.has_color_table:
         imports.add(('array', 'array'))
-        a('    color_table: array = array("L", (')
+        a('    color_table: "array[int]" = array("L", (')
         for grp in chunks(color_table, 8):
             a('        ' + ', '.join(grp) + ',')
         a('    ))')
@@ -185,8 +186,11 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
     if defn.has_color_table:
         a('        self.color_table = array(self.color_table.typecode, self.color_table)')
     a('        if options_dict is not None:')
+    a('            null = object()')
     a('            for key in option_names:')
-    a('                setattr(self, key, options_dict[key])')
+    a('                val = options_dict.get(key, null)')
+    a('                if val is not null:')
+    a('                    setattr(self, key, val)')
 
     a('')
     a('    @property')
@@ -268,7 +272,7 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
 
     for aname, func in action_parsers.items():
         a(f'defaults.{aname} = [')
-        only: Dict[str, List[Tuple[str, Callable]]] = {}
+        only: Dict[str, List[Tuple[str, Callable[..., Any]]]] = {}
         for sc in defn.iter_all_maps(aname):
             if not sc.add_to_default:
                 continue
@@ -278,16 +282,16 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
             else:
                 for val in func(text):
                     a(f'    # {sc.name}')
-                    a(f'    {val!r},')
+                    a(f'    {val!r},  # noqa')
         a(']')
         if only:
             imports.add(('kitty.constants', 'is_macos'))
             for cond, items in only.items():
                 cond = 'is_macos' if cond == 'macos' else 'not is_macos'
                 a(f'if {cond}:')
-                for (text, func) in items:
-                    for val in func(text):
-                        a(f'    defaults.{aname}.append({val!r})')
+                for (text, parser_func) in items:
+                    for val in parser_func(text):
+                        a(f'    defaults.{aname}.append({val!r})  # noqa')
 
     t('')
     t('')
@@ -328,10 +332,10 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
     t('        return True')
     t('    return False')
 
-    preamble = ['# generated by gen-config.py DO NOT edit', '# vim:fileencoding=utf-8', '']
+    preamble = ['# generated by gen-config.py DO NOT edit', '']
     a = preamble.append
 
-    def output_imports(imports: Set, add_module_imports: bool = True) -> None:
+    def output_imports(imports: Set[Tuple[str, str]], add_module_imports: bool = True) -> None:
         a('import typing')
         seen_mods = {'typing'}
         mmap: Dict[str, List[str]] = {}
@@ -365,7 +369,7 @@ def generate_class(defn: Definition, loc: str) -> Tuple[str, str]:
     a(' ' + pprint.pformat(tuple(sorted(option_names, key=natural_keys)))[1:] + '  # }}''}')
     class_def = '\n'.join(preamble + ['', ''] + class_lines)
 
-    preamble = ['# generated by gen-config.py DO NOT edit', '# vim:fileencoding=utf-8', '']
+    preamble = ['# generated by gen-config.py DO NOT edit', '']
     a = preamble.append
     output_imports(tc_imports, False)
 
