@@ -1,10 +1,11 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>
 
 import builtins
 import re
+import textwrap
 import typing
+from functools import lru_cache
 from importlib import import_module
 from typing import (
     Any, Callable, Dict, Iterable, Iterator, List, Match, Optional, Set, Tuple,
@@ -12,6 +13,8 @@ from typing import (
 )
 
 import kitty.conf.utils as generic_parsers
+from kitty.constants import website_url
+from kitty.types import run_once
 
 if typing.TYPE_CHECKING:
     Only = typing.Literal['macos', 'linux', '']
@@ -25,30 +28,44 @@ class Unset:
 
 
 unset = Unset()
+ParserFuncType = Callable[[str], Any]
 
 
 def expand_opt_references(conf_name: str, text: str) -> str:
     conf_name += '.'
 
-    def expand(m: Match) -> str:
+    def expand(m: 'Match[str]') -> str:
         ref = m.group(1)
         if '<' not in ref and '.' not in ref:
             full_ref = conf_name + ref
-            return ':opt:`{} <{}>`'.format(ref, full_ref)
+            return f':opt:`{ref} <{full_ref}>`'
         return str(m.group())
 
     return re.sub(r':opt:`(.+?)`', expand, text)
 
 
+@run_once
+def ref_map() -> Dict[str, str]:
+    from kitty.actions import get_all_actions
+    ref_map = {
+        'layouts': f'{website_url("overview")}#layouts',
+        'watchers': f'{website_url("launch")}#watchers',
+        'sessions': f'{website_url("overview")}#startup-sessions',
+        'functional': f'{website_url("keyboard-protocol")}#functional-key-definitions',
+        'shell_integration': website_url("shell-integration"),
+    }
+    for actions in get_all_actions().values():
+        for ac in actions:
+            ref_map[f'action-{ac.name}'] = f'{website_url("actions")}#' + ac.name.replace('_', '-')
+    return ref_map
+
+
 def remove_markup(text: str) -> str:
 
-    def sub(m: Match) -> str:
+    def sub(m: 'Match[str]') -> str:
         if m.group(1) == 'ref':
-            return {
-                'layouts': 'https://sw.kovidgoyal.net/kitty/index.html#layouts',
-                'sessions': 'https://sw.kovidgoyal.net/kitty/index.html#sessions',
-                'functional': 'https://sw.kovidgoyal.net/kitty/keyboard-protocol.html#functional-key-definitions',
-            }[m.group(2)]
+            q = m.group(2).split('<')[-1].rstrip('>')
+            return ref_map()[q]
         return str(m.group(2))
 
     return re.sub(r':([a-zA-Z0-9]+):`(.+?)`', sub, text, flags=re.DOTALL)
@@ -72,30 +89,31 @@ def iter_blocks(lines: Iterable[str]) -> Iterator[Tuple[List[str], int]]:
         yield current_block, indent_size
 
 
-def wrapped_block(lines: Iterable[str]) -> Iterator[str]:
-    wrapper = getattr(wrapped_block, 'wrapper', None)
-    if wrapper is None:
-        import textwrap
-        wrapper = textwrap.TextWrapper(
-            initial_indent='#: ', subsequent_indent='#: ', width=70, break_long_words=False
+@lru_cache(maxsize=8)
+def block_wrapper(comment_symbol: str) -> textwrap.TextWrapper:
+    return textwrap.TextWrapper(
+            initial_indent=comment_symbol, subsequent_indent=comment_symbol, width=70, break_long_words=False
         )
-        setattr(wrapped_block, 'wrapper', wrapper)
+
+
+def wrapped_block(lines: Iterable[str], comment_symbol: str = '#: ') -> Iterator[str]:
+    wrapper = block_wrapper(comment_symbol)
     for block, indent_size in iter_blocks(lines):
         if indent_size > 0:
             for line in block:
                 if not line:
                     yield line
                 else:
-                    yield '#: ' + line
+                    yield comment_symbol + line
         else:
             for line in wrapper.wrap('\n'.join(block)):
                 yield line
 
 
-def render_block(text: str) -> str:
+def render_block(text: str, comment_symbol: str = '#: ') -> str:
     text = remove_markup(text)
     lines = text.splitlines()
-    return '\n'.join(wrapped_block(lines))
+    return '\n'.join(wrapped_block(lines, comment_symbol))
 
 
 class CoalescedIteratorData:
@@ -148,7 +166,7 @@ class CoalescedIteratorData:
 class Option:
 
     def __init__(
-        self, name: str, defval: str, macos_default: Union[Unset, str], parser_func: Callable,
+        self, name: str, defval: str, macos_default: Union[Unset, str], parser_func: ParserFuncType,
         long_text: str, documented: bool, group: 'Group', choices: Tuple[str, ...], ctype: str
     ):
         self.name = name
@@ -176,11 +194,11 @@ class Option:
             return ans
         if option_group:
             sz = max(len(self.name), max(len(o.name) for o in option_group))
-            a(f'{self.name.ljust(sz)} {self.defval_as_string}')
+            a(f'{self.name.ljust(sz)} {self.defval_as_string}'.rstrip())
             for o in option_group:
-                a(f'{o.name.ljust(sz)} {o.defval_as_string}')
+                a(f'{o.name.ljust(sz)} {o.defval_as_string}'.rstrip())
         else:
-            a(f'{self.name} {self.defval_as_string}')
+            a(f'{self.name} {self.defval_as_string}'.rstrip())
         if self.long_text:
             a('')
             a(render_block(self.long_text))
@@ -220,7 +238,7 @@ class MultiVal:
 
 class MultiOption:
 
-    def __init__(self, name: str, parser_func: Callable, long_text: str, group: 'Group', ctype: str):
+    def __init__(self, name: str, parser_func: ParserFuncType, long_text: str, group: 'Group', ctype: str):
         self.name = name
         self.ctype = ctype
         self.parser_func = parser_func
@@ -237,11 +255,15 @@ class MultiOption:
     def as_conf(self, commented: bool = False, level: int = 0) -> List[str]:
         ans: List[str] = []
         a = ans.append
+        documented = False
         for k in self.items:
             if k.documented:
-                prefix = '' if k.add_to_default else '# '
-                a(f'{prefix}{self.name} {k.defval_as_str}')
-        if self.long_text:
+                documented = True
+                a(f'{self.name} {k.defval_as_str if k.add_to_default else ""}'.rstrip())
+                if not k.add_to_default and k.defval_as_str:
+                    a('')
+                    a(f'#: E.g. {self.name} {k.defval_as_str}'.rstrip())
+        if self.long_text and documented:
             a('')
             a(render_block(self.long_text))
             a('')
@@ -255,7 +277,7 @@ class MultiOption:
         a('')
         for k in self.items:
             if k.documented:
-                a(f'    {self.name:s} {k.defval_as_str}')
+                a(f'    {self.name:s} {k.defval_as_str}'.rstrip())
         a('')
         if self.long_text:
             a(expand_opt_references(conf_name, self.long_text))
@@ -280,14 +302,20 @@ class Mapping:
     def key_text(self) -> str:
         return ''
 
-    def as_conf(self, commented: bool = False, level: int = 0) -> List[str]:
+    def as_conf(self, commented: bool = False, level: int = 0, action_group: List['Mapping'] = []) -> List[str]:
         ans: List[str] = []
-        if self.documented:
-            a = ans.append
-            if self.add_to_default:
-                a(self.setting_name + ' ' + self.parseable_text)
-            if self.long_text:
-                a(''), a(render_block(self.long_text.strip())), a('')
+        if not self.documented:
+            return ans
+        a = ans.append
+        if self.short_text:
+            a(render_block(self.short_text.strip())), a('')
+        for sc in [self] + action_group:
+            if sc.documented:
+                prefix = '' if sc.add_to_default else '#::  E.g. '
+                a(f'{prefix}{sc.setting_name} {sc.parseable_text}')
+        if self.long_text:
+            a(''), a(render_block(self.long_text.strip(), '#::  '))
+        a('')
         return ans
 
     def as_rst(
@@ -428,7 +456,7 @@ class Group:
             a(f'.. _conf-{conf_name}-{self.name}:')
             a('')
             a(self.title)
-            heading_level = '+' if level > 1 else '^'
+            heading_level = '+' if level > 1 else '-'
             a(heading_level * (len(self.title) + 20))
             a('')
             if self.start_text:
@@ -463,11 +491,13 @@ class Group:
                 a(render_block(self.start_text))
                 a('')
         else:
-            ans.extend(('# vim:fileencoding=utf-8:ft=conf:foldmethod=marker', ''))
+            ans.extend(('# vim:fileencoding=utf-8:foldmethod=marker', ''))
 
         for item in self.iter_with_coalesced_options():
             if isinstance(item, Option):
                 lines = item.as_conf(option_group=self.coalesced_iterator_data.option_group_for_option(item))
+            elif isinstance(item, Mapping):
+                lines = item.as_conf(commented, level + 1, action_group=self.coalesced_iterator_data.action_group_for_action(item))
             else:
                 lines = item.as_conf(commented, level + 1)
             ans.extend(lines)
@@ -509,7 +539,7 @@ class Group:
         return ans
 
 
-def resolve_import(name: str, module: Any = None) -> Callable:
+def resolve_import(name: str, module: Any = None) -> ParserFuncType:
     ans = None
     if name.count('.') > 1:
         m = import_module(name.rpartition('.')[0])
@@ -522,7 +552,7 @@ def resolve_import(name: str, module: Any = None) -> Callable:
                 ans = getattr(module, name)
     if not callable(ans):
         raise TypeError(f'{name} is not a function')
-    return cast(Callable, ans)
+    return cast(ParserFuncType, ans)
 
 
 class Action:
@@ -554,7 +584,7 @@ class Definition:
         self.shortcut_map: Dict[str, List[ShortcutMapping]] = {}
         self.mouse_map: Dict[str, List[MouseMapping]] = {}
         self.actions = {a.name: a.resolve_imports(self.module_for_parsers) for a in actions}
-        self.deprecations: Dict[Callable, Tuple[str, ...]] = {}
+        self.deprecations: Dict[ParserFuncType, Tuple[str, ...]] = {}
 
     def iter_all_non_groups(self) -> Iterator[NonGroups]:
         yield from self.root_group.iter_all_non_groups()
@@ -566,22 +596,22 @@ class Definition:
 
     def iter_all_maps(self, which: str = 'map') -> Iterator[Union[ShortcutMapping, MouseMapping]]:
         for x in self.iter_all_non_groups():
-            if isinstance(x, ShortcutMapping) and which == 'map':
+            if isinstance(x, ShortcutMapping) and which in ('map', '*'):
                 yield x
-            elif isinstance(x, MouseMapping) and which == 'mouse_map':
+            elif isinstance(x, MouseMapping) and which in ('mouse_map', '*'):
                 yield x
 
-    def parser_func(self, name: str) -> Callable:
+    def parser_func(self, name: str) -> ParserFuncType:
         ans = getattr(builtins, name, None)
         if callable(ans):
-            return cast(Callable, ans)
+            return cast(ParserFuncType, ans)
         ans = getattr(generic_parsers, name, None)
         if callable(ans):
-            return cast(Callable, ans)
+            return cast(ParserFuncType, ans)
         ans = getattr(self.module_for_parsers, name)
         if not callable(ans):
             raise TypeError(f'{name} is not a function')
-        return cast(Callable, ans)
+        return cast(ParserFuncType, ans)
 
     def add_group(self, name: str, title: str = '', start_text: str = '') -> None:
         self.current_group = Group(name, title or name, self.coalesced_iterator_data, start_text.strip(), self.current_group)
