@@ -1,16 +1,17 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
 import re
 import shlex
+from contextlib import contextmanager
 from typing import (
-    Any, Callable, Dict, Generator, Iterable, List, NamedTuple, Optional,
-    Sequence, Set, Tuple, TypeVar, Union
+    Any, Callable, Dict, Generator, Generic, Iterable, Iterator, List,
+    NamedTuple, Optional, Sequence, Set, Tuple, TypeVar, Union
 )
 
-from ..rgb import Color, to_color as as_color
+from ..fast_data_types import Color
+from ..rgb import to_color as as_color
 from ..types import ConvertibleToNumbers, ParsedShortcut
 from ..typing import Protocol
 from ..utils import expandvars, log_error
@@ -30,6 +31,7 @@ class BadLine(NamedTuple):
     number: int
     line: str
     exception: Exception
+    file: str
 
 
 def positive_int(x: ConvertibleToNumbers) -> int:
@@ -96,8 +98,8 @@ def to_cmdline(x: str) -> List[str]:
 
 
 def python_string(text: str) -> str:
-    import ast
-    ans: str = ast.literal_eval("'''" + text.replace("'''", "'\\''") + "'''")
+    from ast import literal_eval
+    ans: str = literal_eval("'''" + text.replace("'''", "'\\''") + "'''")
     return ans
 
 
@@ -118,6 +120,40 @@ def choices(*choices: str) -> Choice:
     return Choice(choices)
 
 
+class CurrentlyParsing:
+    __slots__ = 'line', 'number', 'file'
+
+    def __init__(self, line: str = '', number: int = -1, file: str = ''):
+        self.line = line
+        self.number = number
+        self.file = file
+
+    def __copy__(self) -> 'CurrentlyParsing':
+        return CurrentlyParsing(self.line, self.number, self.file)
+
+    @contextmanager
+    def set_line(self, line: str, number: int) -> Iterator['CurrentlyParsing']:
+        orig = self.line, self.number
+        self.line = line
+        self.number = number
+        try:
+            yield self
+        finally:
+            self.line, self.number = orig
+
+    @contextmanager
+    def set_file(self, file: str) -> Iterator['CurrentlyParsing']:
+        orig = self.file
+        self.file = file
+        try:
+            yield self
+        finally:
+            self.file = orig
+
+
+currently_parsing = CurrentlyParsing()
+
+
 def parse_line(
     line: str,
     parse_conf_item: ItemParser,
@@ -130,7 +166,7 @@ def parse_line(
         return
     m = key_pat.match(line)
     if m is None:
-        log_error('Ignoring invalid config line: {}'.format(line))
+        log_error(f'Ignoring invalid config line: {line}')
         return
     key, val = m.groups()
     if key == 'include':
@@ -139,7 +175,8 @@ def parse_line(
             val = os.path.join(base_path_for_includes, val)
         try:
             with open(val, encoding='utf-8', errors='replace') as include:
-                _parse(include, parse_conf_item, ans, accumulate_bad_lines)
+                with currently_parsing.set_file(val):
+                    _parse(include, parse_conf_item, ans, accumulate_bad_lines)
         except FileNotFoundError:
             log_error(
                 'Could not find included config file: {}, ignoring'.
@@ -152,7 +189,7 @@ def parse_line(
             )
         return
     if not parse_conf_item(key, val, ans):
-        log_error('Ignoring unknown config key: {}'.format(key))
+        log_error(f'Ignoring unknown config key: {key}')
 
 
 def _parse(
@@ -169,13 +206,12 @@ def _parse(
         base_path_for_includes = config_dir
     for i, line in enumerate(lines):
         try:
-            parse_line(
-                line, parse_conf_item, ans, base_path_for_includes, accumulate_bad_lines
-            )
+            with currently_parsing.set_line(line, i + 1):
+                parse_line(line, parse_conf_item, ans, base_path_for_includes, accumulate_bad_lines)
         except Exception as e:
             if accumulate_bad_lines is None:
                 raise
-            accumulate_bad_lines.append(BadLine(i + 1, line.rstrip(), e))
+            accumulate_bad_lines.append(BadLine(i + 1, line.rstrip(), e, currently_parsing.file))
 
 
 def parse_config_base(
@@ -189,7 +225,7 @@ def parse_config_base(
     )
 
 
-def merge_dicts(defaults: Dict, newvals: Dict) -> Dict:
+def merge_dicts(defaults: Dict[str, Any], newvals: Dict[str, Any]) -> Dict[str, Any]:
     ans = defaults.copy()
     ans.update(newvals)
     return ans
@@ -199,8 +235,7 @@ def resolve_config(SYSTEM_CONF: str, defconf: str, config_files_on_cmd_line: Seq
     if config_files_on_cmd_line:
         if 'NONE' not in config_files_on_cmd_line:
             yield SYSTEM_CONF
-            for cf in config_files_on_cmd_line:
-                yield cf
+            yield from config_files_on_cmd_line
     else:
         yield SYSTEM_CONF
         yield defconf
@@ -209,7 +244,7 @@ def resolve_config(SYSTEM_CONF: str, defconf: str, config_files_on_cmd_line: Seq
 def load_config(
     defaults: OptionsProtocol,
     parse_config: Callable[[Iterable[str]], Dict[str, Any]],
-    merge_configs: Callable[[Dict, Dict], Dict],
+    merge_configs: Callable[[Dict[str, Any], Dict[str, Any]], Dict[str, Any]],
     *paths: str,
     overrides: Optional[Iterable[str]] = None
 ) -> Tuple[Dict[str, Any], Tuple[str, ...]]:
@@ -220,33 +255,38 @@ def load_config(
             continue
         try:
             with open(path, encoding='utf-8', errors='replace') as f:
-                vals = parse_config(f)
+                with currently_parsing.set_file(path):
+                    vals = parse_config(f)
         except (FileNotFoundError, PermissionError):
             continue
         found_paths.append(path)
         ans = merge_configs(ans, vals)
     if overrides is not None:
-        vals = parse_config(overrides)
+        with currently_parsing.set_file('<override>'):
+            vals = parse_config(overrides)
         ans = merge_configs(ans, vals)
     return ans, tuple(found_paths)
 
 
-def key_func() -> Tuple[Callable[..., Callable], Dict[str, Callable]]:
-    ans: Dict[str, Callable] = {}
+ReturnType = TypeVar('ReturnType')
+KeyFunc = Callable[[str, str], ReturnType]
 
-    def func_with_args(*names: str) -> Callable:
 
-        def w(f: Callable) -> Callable:
+class KeyFuncWrapper(Generic[ReturnType]):
+    def __init__(self) -> None:
+        self.args_funcs: Dict[str, KeyFunc[ReturnType]] = {}
+
+    def __call__(self, *names: str) -> Callable[[KeyFunc[ReturnType]], KeyFunc[ReturnType]]:
+
+        def w(f: KeyFunc[ReturnType]) -> KeyFunc[ReturnType]:
             for name in names:
-                if ans.setdefault(name, f) is not f:
-                    raise ValueError(
-                        'the args_func {} is being redefined'.format(name)
-                    )
+                if self.args_funcs.setdefault(name, f) is not f:
+                    raise ValueError(f'the args_func {name} is being redefined')
             return f
-
         return w
 
-    return func_with_args, ans
+    def get(self, name: str) -> Optional[KeyFunc[ReturnType]]:
+        return self.args_funcs.get(name)
 
 
 class KeyAction(NamedTuple):
@@ -258,8 +298,14 @@ class KeyAction(NamedTuple):
             return f'KeyAction({self.func!r}, {self.args!r})'
         return f'KeyAction({self.func!r})'
 
+    def pretty(self) -> str:
+        ans = self.func
+        for x in self.args:
+            ans += f' {x}'
+        return ans
 
-def parse_kittens_func_args(action: str, args_funcs: Dict[str, Callable]) -> KeyAction:
+
+def parse_kittens_func_args(action: str, args_funcs: Dict[str, KeyFunc[Tuple[str, Any]]]) -> KeyAction:
     parts = action.strip().split(' ', 1)
     func = parts[0]
     if len(parts) == 1:
@@ -277,7 +323,7 @@ def parse_kittens_func_args(action: str, args_funcs: Dict[str, Callable]) -> Key
     try:
         func, args = parser(func, rest)
     except Exception:
-        raise ValueError('Unknown key action: {}'.format(action))
+        raise ValueError(f'Unknown key action: {action}')
 
     if not isinstance(args, (list, tuple)):
         args = (args, )
@@ -290,7 +336,7 @@ KittensKeyMap = Dict[ParsedShortcut, KeyAction]
 
 
 def parse_kittens_key(
-    val: str, funcs_with_args: Dict[str, Callable]
+    val: str, funcs_with_args: Dict[str, KeyFunc[Tuple[str, Any]]]
 ) -> Optional[KittensKeyDefinition]:
     from ..key_encoding import parse_shortcut
     sc, action = val.partition(' ')[::2]
@@ -310,7 +356,8 @@ def save_type_stub(text: str, fpath: str) -> None:
     fpath += 'i'
     preamble = '# Update this file by running: ./test.py mypy\n\n'
     try:
-        existing = open(fpath).read()
+        with open(fpath) as fs:
+            existing = fs.read()
     except FileNotFoundError:
         existing = ''
     current = preamble + text
