@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 #
@@ -11,8 +10,8 @@ import math
 from functools import partial as p, wraps
 from itertools import repeat
 from typing import (
-    Any, Callable, Dict, Generator, Iterable, List, MutableSequence, Optional,
-    Sequence, Tuple, cast
+    Any, Callable, Dict, Iterable, Iterator, List, MutableSequence, Optional,
+    Sequence, Tuple
 )
 
 scale = (0.001, 1., 1.5, 2.)
@@ -60,17 +59,17 @@ def half_vline(buf: BufType, width: int, height: int, level: int = 1, which: str
 
 
 def get_holes(sz: int, hole_sz: int, num: int) -> List[Tuple[int, ...]]:
-    if num == 1:
-        pts = [sz // 2]
-    elif num == 2:
-        ssz = (sz - 2 * hole_sz) // 3
-        pts = [ssz + hole_sz // 2, 2 * ssz + hole_sz // 2 + hole_sz]
-    elif num == 3:
-        ssz = (sz - 3 * hole_sz) // 4
-        pts = [ssz + hole_sz // 2, 2 * ssz + hole_sz // 2 + hole_sz, 3 * ssz + 2 * hole_sz + hole_sz // 2]
+    all_holes_use = (num + 1) * hole_sz
+    individual_block_size = (sz - all_holes_use) // (num + 1)
+    half_hole_sz = hole_sz // 2
+    pos = - half_hole_sz
     holes = []
-    for c in pts:
-        holes.append(tuple(range(c - hole_sz // 2, c - hole_sz // 2 + hole_sz)))
+    while pos < sz:
+        left = max(0, pos)
+        right = min(sz, pos + hole_sz)
+        if right > left:
+            holes.append(tuple(range(left, right)))
+        pos = right + individual_block_size
     return holes
 
 
@@ -169,11 +168,11 @@ class SSByteArray(bytearray):
     supersample_factor = 1
 
 
-def supersampled(supersample_factor: int = 4) -> Callable:
+def supersampled(supersample_factor: int = 4) -> Callable[[Callable[..., None]], Callable[..., None]]:
     # Anti-alias the drawing performed by the wrapped function by
     # using supersampling
 
-    def create_wrapper(f: Callable) -> Callable:
+    def create_wrapper(f: Callable[..., None]) -> Callable[..., None]:
         @wraps(f)
         def supersampled_wrapper(buf: BufType, width: int, height: int, *args: Any, **kw: Any) -> None:
             w, h = supersample_factor * width, supersample_factor * height
@@ -315,6 +314,7 @@ def mid_lines(buf: SSByteArray, width: int, height: int, level: int = 1, pts: It
             return width - 1, mid_y
         if p == 'b':
             return mid_x, height - 1
+        raise KeyError(f'Unknown p: {p}')
 
     for x in pts:
         p1, p2 = map(pt_to_coords, x)
@@ -354,7 +354,7 @@ def find_bezier_for_D(width: int, height: int) -> int:
         cx += 1
 
 
-def get_bezier_limits(bezier_x: ParameterizedFunc, bezier_y: ParameterizedFunc) -> Generator[Tuple[float, float], None, int]:
+def get_bezier_limits(bezier_x: ParameterizedFunc, bezier_y: ParameterizedFunc) -> Iterator[Tuple[float, float]]:
     start_x = int(bezier_x(0))
     max_x = int(bezier_x(0.5))
     last_t, t_limit = 0., 0.5
@@ -372,7 +372,7 @@ def get_bezier_limits(bezier_x: ParameterizedFunc, bezier_y: ParameterizedFunc)
             if q > x:
                 increment /= 2
                 if increment < 1e-6:
-                    raise ValueError('Failed to find t for x={}'.format(x))
+                    raise ValueError(f'Failed to find t for x={x}')
             else:
                 start_t += increment
                 increment = t_limit - start_t
@@ -624,7 +624,6 @@ def shade(buf: BufType, width: int, height: int, light: bool = False, invert: bo
     dest = bytearray(width * height) if invert else buf
 
     for r in range(number_of_rows):
-        y = r * square_sz
         is_odd = r % 2 != 0
         if is_odd:
             continue
@@ -781,7 +780,7 @@ def braille(buf: BufType, width: int, height: int, which: int = 0) -> None:
             braille_dot(buf, width, height, col, row)
 
 
-box_chars: Dict[str, List[Callable]] = {
+box_chars: Dict[str, List[Callable[[BufType, int, int], Any]]] = {
     '─': [hline],
     '━': [p(hline, level=3)],
     '│': [vline],
@@ -889,8 +888,8 @@ box_chars: Dict[str, List[Callable]] = {
 
     '🬼': [p(smooth_mosaic, a=(0, 0.75), b=(0.5, 1))],
     '🬽': [p(smooth_mosaic, a=(0, 0.75), b=(1, 1))],
-    '🬾': [p(smooth_mosaic, a=(0, 0.5), b=(0.5, 1))],
-    '🬿': [p(smooth_mosaic, a=(0, 0.5), b=(1, 1))],
+    '🬾': [p(smooth_mosaic, a=(0, 0.25), b=(0.5, 1))],
+    '🬿': [p(smooth_mosaic, a=(0, 0.25), b=(1, 1))],
     '🭀': [p(smooth_mosaic, a=(0, 0), b=(0.5, 1))],
 
     '🭁': [p(smooth_mosaic, a=(0, 0.25), b=(0.5, 0))],
@@ -1008,7 +1007,7 @@ for starts, func, pattern in (
 
 for chars, func_ in (('╒╕╘╛', dvcorner), ('╓╖╙╜', dhcorner), ('╔╗╚╝', dcorner), ('╟╢╤╧', dpip)):
     for ch in chars:
-        box_chars[ch] = [p(cast(Callable, func_), which=ch)]
+        box_chars[ch] = [p(func_, which=ch)]
 
 for i in range(256):
     box_chars[chr(0x2800 + i)] = [p(braille, which=i)]
@@ -1044,8 +1043,9 @@ def render_missing_glyph(buf: BufType, width: int, height: int) -> None:
 
 def test_char(ch: str, sz: int = 48) -> None:
     # kitty +runpy "from kitty.fonts.box_drawing import test_char; test_char('XXX')"
-    from .render import display_bitmap, setup_for_testing
     from kitty.fast_data_types import concat_cells, set_send_sprite_to_gpu
+
+    from .render import display_bitmap, setup_for_testing
     with setup_for_testing('monospace', sz) as (_, width, height):
         buf = bytearray(width * height)
         try:
@@ -1063,9 +1063,10 @@ def test_char(ch: str, sz: int = 48) -> None:
 
 
 def test_drawing(sz: int = 48, family: str = 'monospace', start: int = 0x2500, num_rows: int = 10, num_cols: int = 16) -> None:
-    from .render import display_bitmap, setup_for_testing
     from kitty.fast_data_types import concat_cells, set_send_sprite_to_gpu
 
+    from .render import display_bitmap, setup_for_testing
+
     with setup_for_testing(family, sz) as (_, width, height):
         space = bytearray(width * height)
 
@@ -1096,7 +1097,7 @@ def test_drawing(sz: int = 48, family: str = 'monospace', start: int = 0x2500, n
             rgb_data = b''.join(rows)
             width *= 32
             height *= len(rows)
-            assert len(rgb_data) == width * height * 4, '{} != {}'.format(len(rgb_data), width * height * 4)
+            assert len(rgb_data) == width * height * 4, f'{len(rgb_data)} != {width * height * 4}'
             display_bitmap(rgb_data, width, height)
         finally:
             set_send_sprite_to_gpu(None)
