@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import ctypes
@@ -7,7 +6,8 @@ import sys
 from functools import partial
 from math import ceil, cos, floor, pi
 from typing import (
-    Any, Callable, Dict, Generator, List, Optional, Tuple, Union, cast
+    TYPE_CHECKING, Any, Callable, Dict, Generator, List, Optional, Tuple,
+    Union, cast
 )
 
 from kitty.constants import is_macos
@@ -24,9 +24,15 @@ from kitty.typing import CoreTextFont, FontConfigPattern
 from kitty.utils import log_error
 
 if is_macos:
-    from .core_text import get_font_files as get_font_files_coretext, font_for_family as font_for_family_macos, find_font_features
+    from .core_text import (
+        find_font_features, font_for_family as font_for_family_macos,
+        get_font_files as get_font_files_coretext
+    )
 else:
-    from .fontconfig import get_font_files as get_font_files_fontconfig, font_for_family as font_for_family_fontconfig, find_font_features
+    from .fontconfig import (
+        find_font_features, font_for_family as font_for_family_fontconfig,
+        get_font_files as get_font_files_fontconfig
+    )
 
 FontObject = Union[CoreTextFont, FontConfigPattern]
 current_faces: List[Tuple[FontObject, bool, bool]] = []
@@ -201,10 +207,14 @@ def set_font_family(opts: Optional[Options] = None, override_font_size: Optional
     )
 
 
-UnderlineCallback = Callable[[ctypes.Array, int, int, int, int], None]
+if TYPE_CHECKING:
+    CBufType = ctypes.Array[ctypes.c_ubyte]
+else:
+    CBufType = None
+UnderlineCallback = Callable[[CBufType, int, int, int, int], None]
 
 
-def add_line(buf: ctypes.Array, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+def add_line(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
     y = position - thickness // 2
     while thickness > 0 and -1 < y < cell_height:
         thickness -= 1
@@ -212,7 +222,7 @@ def add_line(buf: ctypes.Array, cell_width: int, position: int, thickness: int,
         y += 1
 
 
-def add_dline(buf: ctypes.Array, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+def add_dline(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
     a = min(position - thickness, cell_height - 1)
     b = min(position, cell_height - 1)
     top, bottom = min(a, b), max(a, b)
@@ -232,7 +242,7 @@ def add_dline(buf: ctypes.Array, cell_width: int, position: int, thickness: int,
         ctypes.memset(ctypes.addressof(buf) + (cell_width * y), 255, cell_width)
 
 
-def add_curl(buf: ctypes.Array, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
+def add_curl(buf: CBufType, cell_width: int, position: int, thickness: int, cell_height: int) -> None:
     max_x, max_y = cell_width - 1, cell_height - 1
     xfactor = 2.0 * pi / max_x
     thickness = max(1, thickness)
@@ -281,7 +291,7 @@ def render_special(
     strikethrough_thickness: int = 0,
     dpi_x: float = 96.,
     dpi_y: float = 96.,
-) -> ctypes.Array:
+) -> CBufType:
     underline_position = min(underline_position, cell_height - underline_thickness)
     CharTexture = ctypes.c_ubyte * (cell_width * cell_height)
 
@@ -318,7 +328,7 @@ def render_cursor(
     cell_height: int = 0,
     dpi_x: float = 0,
     dpi_y: float = 0
-) -> ctypes.Array:
+) -> CBufType:
     CharTexture = ctypes.c_ubyte * (cell_width * cell_height)
     ans = CharTexture()
 
@@ -362,7 +372,7 @@ def prerender_function(
     cursor_underline_thickness: float,
     dpi_x: float,
     dpi_y: float
-) -> Tuple[Union[int, ctypes.Array], ...]:
+) -> Tuple[Tuple[int, ...], Tuple[CBufType, ...]]:
     # Pre-render the special underline, strikethrough and missing and cursor cells
     f = partial(
         render_special, cell_width=cell_width, cell_height=cell_height, baseline=baseline,
@@ -375,10 +385,10 @@ def prerender_function(
         cursor_underline_thickness=cursor_underline_thickness, cell_width=cell_width,
         cell_height=cell_height, dpi_x=dpi_x, dpi_y=dpi_y)
     cells = f(1), f(2), f(3), f(0, True), f(missing=True), c(1), c(2), c(3)
-    return tuple(map(ctypes.addressof, cells)) + (cells,)
+    return tuple(map(ctypes.addressof, cells)), cells
 
 
-def render_box_drawing(codepoint: int, cell_width: int, cell_height: int, dpi: float) -> Tuple[int, ctypes.Array]:
+def render_box_drawing(codepoint: int, cell_width: int, cell_height: int, dpi: float) -> Tuple[int, CBufType]:
     CharTexture = ctypes.c_ubyte * (cell_width * cell_height)
     buf = CharTexture()
     render_box_char(
@@ -445,6 +455,7 @@ def shape_string(
 
 def display_bitmap(rgb_data: bytes, width: int, height: int) -> None:
     from tempfile import NamedTemporaryFile
+
     from kittens.icat.main import detect_support, show
     if not hasattr(display_bitmap, 'detected') and not detect_support():
         raise SystemExit('Your terminal does not support the graphics protocol')
