@@ -19,8 +19,9 @@ extern void cocoa_focus_window(void *w);
 extern long cocoa_window_number(void *w);
 extern void cocoa_create_global_menu(void);
 extern void cocoa_hide_window_title(void *w);
-extern void cocoa_system_beep(void);
+extern void cocoa_system_beep(const char*);
 extern void cocoa_set_activation_policy(bool);
+extern void cocoa_set_titlebar_appearance(void *w, unsigned int theme);
 extern void cocoa_set_titlebar_color(void *w, color_type color);
 extern bool cocoa_alt_option_key_pressed(unsigned long);
 extern size_t cocoa_get_workspace_ids(void *w, size_t *workspace_ids, size_t array_sz);
@@ -84,8 +85,8 @@ update_os_window_viewport(OSWindow *window, bool notify_boss) {
     window->viewport_updated_at_least_once = true;
     window->viewport_width = fw; window->viewport_height = fh;
     double xr = window->viewport_x_ratio, yr = window->viewport_y_ratio;
-    window->viewport_x_ratio = w > 0 ? (double)window->viewport_width / (double)w : xr;
-    window->viewport_y_ratio = h > 0 ? (double)window->viewport_height / (double)h : yr;
+    window->viewport_x_ratio = (double)window->viewport_width / (double)w;
+    window->viewport_y_ratio = (double)window->viewport_height / (double)h;
     bool dpi_changed = (xr != 0.0 && xr != window->viewport_x_ratio) || (yr != 0.0 && yr != window->viewport_y_ratio) || (xdpi != window->logical_dpi_x) || (ydpi != window->logical_dpi_y);
 
     window->viewport_size_dirty = true;
@@ -154,19 +155,34 @@ show_mouse_cursor(GLFWwindow *w) {
 }
 
 void
-blank_os_window(OSWindow *w) {
+blank_os_window(OSWindow *osw) {
     color_type color = OPT(background);
-    if (w->num_tabs > 0) {
-        Tab *t = w->tabs + w->active_tab;
+    if (osw->num_tabs > 0) {
+        Tab *t = osw->tabs + osw->active_tab;
         if (t->num_windows == 1) {
             Window *w = t->windows + t->active_window;
             Screen *s = w->render_data.screen;
             if (s) {
-                color = colorprofile_to_color(s->color_profile, s->color_profile->overridden.default_bg, s->color_profile->configured.default_bg);
+                color = colorprofile_to_color(s->color_profile, s->color_profile->overridden.default_bg, s->color_profile->configured.default_bg).rgb;
             }
         }
     }
-    blank_canvas(w->is_semi_transparent ? w->background_opacity : 1.0f, color);
+    blank_canvas(osw->is_semi_transparent ? osw->background_opacity : 1.0f, color);
+}
+
+static void
+window_pos_callback(GLFWwindow* window, int x UNUSED, int y UNUSED) {
+    if (!set_callback_window(window)) return;
+#ifdef __APPLE__
+    // Apple needs IME position to be accurate before the next key event
+    OSWindow *osw = global_state.callback_os_window;
+    if (osw->is_focused && is_window_ready_for_callbacks()) {
+        Tab *tab = osw->tabs + osw->active_tab;
+        Window *w = tab->windows + tab->active_window;
+        if (w->render_data.screen) update_ime_position(w, w->render_data.screen);
+    }
+#endif
+    global_state.callback_os_window = NULL;
 }
 
 static void
@@ -367,6 +383,11 @@ window_focus_callback(GLFWwindow *w, int focused) {
         WINDOW_CALLBACK(on_focus, "O", focused ? Py_True : Py_False);
         GLFWIMEUpdateEvent ev = { .type = GLFW_IME_UPDATE_FOCUS, .focused = focused };
         glfwUpdateIMEState(global_state.callback_os_window->handle, &ev);
+        if (focused) {
+            Tab *tab = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+            Window *window = tab->windows + tab->active_window;
+            if (window->render_data.screen) update_ime_position(window, window->render_data.screen);
+        }
     }
     request_tick_callback();
     global_state.callback_os_window = NULL;
@@ -411,13 +432,29 @@ apple_file_open_callback(const char* filepath) {
     set_cocoa_pending_action(OPEN_FILE, filepath);
     return true;
 }
+
+extern bool cocoa_render_line_of_text(const char *text, const color_type fg, const color_type bg, uint8_t *rgba_output, const size_t width, const size_t height);
+
+bool
+draw_window_title(OSWindow *window UNUSED, const char *text, color_type fg, color_type bg, uint8_t *output_buf, size_t width, size_t height) {
+    return cocoa_render_line_of_text(text, fg, bg, output_buf, width, height);
+}
+
+extern uint8_t* render_single_ascii_char_as_mask(const char ch, size_t *result_width, size_t *result_height);
+
+uint8_t*
+draw_single_ascii_char(const char ch, size_t *result_width, size_t *result_height) {
+    uint8_t *ans = render_single_ascii_char_as_mask(ch, result_width, result_height);
+    if (PyErr_Occurred()) PyErr_Print();
+    return ans;
+}
+
 #else
 
 static FreeTypeRenderCtx csd_title_render_ctx = NULL;
 
 static bool
-draw_text_callback(GLFWwindow *window, const char *text, uint32_t fg, uint32_t bg, uint8_t *output_buf, size_t width, size_t height, float x_offset, float y_offset, size_t right_margin) {
-    if (!set_callback_window(window)) return false;
+ensure_csd_title_render_ctx(void) {
     if (!csd_title_render_ctx) {
         csd_title_render_ctx = create_freetype_render_context(NULL, true, false);
         if (!csd_title_render_ctx) {
@@ -425,6 +462,13 @@ draw_text_callback(GLFWwindow *window, const char *text, uint32_t fg, uint32_t b
             return false;
         }
     }
+    return true;
+}
+
+static bool
+draw_text_callback(GLFWwindow *window, const char *text, uint32_t fg, uint32_t bg, uint8_t *output_buf, size_t width, size_t height, float x_offset, float y_offset, size_t right_margin) {
+    if (!set_callback_window(window)) return false;
+    if (!ensure_csd_title_render_ctx()) return false;
     double xdpi, ydpi;
     get_window_dpi(window, &xdpi, &ydpi);
     unsigned px_sz = (unsigned)(global_state.callback_os_window->font_sz_in_pts * ydpi / 72.);
@@ -435,6 +479,26 @@ draw_text_callback(GLFWwindow *window, const char *text, uint32_t fg, uint32_t b
     if (!ok && PyErr_Occurred()) PyErr_Print();
     return ok;
 }
+
+bool
+draw_window_title(OSWindow *window, const char *text, color_type fg, color_type bg, uint8_t *output_buf, size_t width, size_t height) {
+    if (!ensure_csd_title_render_ctx()) return false;
+    unsigned px_sz = (unsigned)(window->fonts_data->font_sz_in_pts * window->fonts_data->logical_dpi_y / 72.);
+    px_sz = MIN(px_sz, 3 * height / 4);
+#define RGB2BGR(x) (x & 0xFF000000) | ((x & 0xFF0000) >> 16) | (x & 0x00FF00) | ((x & 0x0000FF) << 16)
+    bool ok = render_single_line(csd_title_render_ctx, text, px_sz, RGB2BGR(fg), RGB2BGR(bg), output_buf, width, height, 0, 0, 0);
+#undef RGB2BGR
+    if (!ok && PyErr_Occurred()) PyErr_Print();
+    return ok;
+}
+
+uint8_t*
+draw_single_ascii_char(const char ch, size_t *result_width, size_t *result_height) {
+    if (!ensure_csd_title_render_ctx()) return NULL;
+    uint8_t *ans = render_single_ascii_char_as_mask(csd_title_render_ctx, ch, result_width, result_height);
+    if (PyErr_Occurred()) PyErr_Print();
+    return ans;
+}
 #endif
 // }}}
 
@@ -478,6 +542,7 @@ make_os_window_context_current(OSWindow *w) {
     GLFWwindow *current_context = glfwGetCurrentContext();
     if (w->handle != current_context) {
         glfwMakeContextCurrent(w->handle);
+        global_state.current_opengl_context_id = w->id;
     }
 }
 
@@ -529,18 +594,21 @@ set_os_window_dpi(OSWindow *w) {
 }
 
 static bool
-do_toggle_fullscreen(OSWindow *w) {
+do_toggle_fullscreen(OSWindow *w, unsigned int flags, bool restore_sizes) {
     int width, height, x, y;
     glfwGetWindowSize(w->handle, &width, &height);
     glfwGetWindowPos(w->handle, &x, &y);
-    if (glfwToggleFullscreen(w->handle, 0)) {
+    bool was_maximized = glfwGetWindowAttrib(w->handle, GLFW_MAXIMIZED);
+    if (glfwToggleFullscreen(w->handle, flags)) {
         w->before_fullscreen.is_set = true;
         w->before_fullscreen.w = width; w->before_fullscreen.h = height; w->before_fullscreen.x = x; w->before_fullscreen.y = y;
+        w->before_fullscreen.was_maximized = was_maximized;
         return true;
     }
-    if (w->before_fullscreen.is_set) {
+    if (w->before_fullscreen.is_set && restore_sizes) {
         glfwSetWindowSize(w->handle, w->before_fullscreen.w, w->before_fullscreen.h);
         glfwSetWindowPos(w->handle, w->before_fullscreen.x, w->before_fullscreen.y);
+        if (w->before_fullscreen.was_maximized) glfwMaximizeWindow(w->handle);
     }
     return false;
 }
@@ -549,13 +617,23 @@ static bool
 toggle_fullscreen_for_os_window(OSWindow *w) {
     if (w && w->handle) {
 #ifdef __APPLE__
-        if (!OPT(macos_traditional_fullscreen)) return glfwToggleFullscreen(w->handle, 1);
+        if (!OPT(macos_traditional_fullscreen)) return do_toggle_fullscreen(w, 1, false);
 #endif
-        return do_toggle_fullscreen(w);
+        return do_toggle_fullscreen(w, 0, true);
     }
     return false;
 }
 
+bool
+is_os_window_fullscreen(OSWindow *w) {
+    unsigned int flags = 0;
+#ifdef __APPLE__
+    if (!OPT(macos_traditional_fullscreen)) flags = 1;
+#endif
+    if (w && w->handle) return glfwIsFullscreen(w->handle, flags);
+    return false;
+}
+
 static bool
 toggle_maximized_for_os_window(OSWindow *w) {
     bool maximized = false;
@@ -600,11 +678,12 @@ intercept_cocoa_fullscreen(GLFWwindow *w) {
 #endif
 
 void
-set_titlebar_color(OSWindow *w, color_type color, bool use_system_color) {
+set_titlebar_color(OSWindow *w, color_type color, bool use_system_color, unsigned int system_color UNUSED) {
     if (w->handle && (!w->last_titlebar_color || (w->last_titlebar_color & 0xffffff) != (color & 0xffffff))) {
         w->last_titlebar_color = (1 << 24) | (color & 0xffffff);
 #ifdef __APPLE__
         if (!use_system_color) cocoa_set_titlebar_color(glfwGetCocoaWindow(w->handle), color);
+        else cocoa_set_titlebar_appearance(glfwGetCocoaWindow(w->handle), system_color);
 #else
         if (global_state.is_wayland && glfwWaylandSetTitlebarColor) glfwWaylandSetTitlebarColor(w->handle, color, use_system_color);
 #endif
@@ -762,7 +841,7 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
     if (logo.pixels && logo.width && logo.height) glfwSetWindowIcon(glfw_window, 1, &logo);
     glfwSetCursor(glfw_window, standard_cursor);
     update_os_window_viewport(w, false);
-    // missing pos callback
+    glfwSetWindowPosCallback(glfw_window, window_pos_callback);
     // missing size callback
     glfwSetWindowCloseCallback(glfw_window, window_close_callback);
     glfwSetWindowRefreshCallback(glfw_window, refresh_callback);
@@ -794,6 +873,7 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
     w->cursor_blink_zero_time = now;
     w->last_mouse_activity_at = now;
     w->is_semi_transparent = is_semi_transparent;
+    global_state.current_opengl_context_id = w->id;
     if (want_semi_transparent && !w->is_semi_transparent) {
         static bool warned = false;
         if (!warned) {
@@ -1115,9 +1195,10 @@ ring_audio_bell(void) {
     if (last_bell_at >= 0 && now - last_bell_at <= ms_to_monotonic_t(100ll)) return;
     last_bell_at = now;
 #ifdef __APPLE__
-    cocoa_system_beep();
+    cocoa_system_beep(OPT(bell_path));
 #else
-    play_canberra_sound("bell", "kitty bell");
+    if (OPT(bell_path)) play_canberra_sound(OPT(bell_path), "kitty bell", true, "event");
+    else play_canberra_sound("bell", "kitty bell", false, "event");
 #endif
 }
 
@@ -1196,9 +1277,36 @@ request_window_attention(id_type kitty_window_id, bool audio_bell) {
     }
 }
 
+static void
+strip_csi_(const char *title, char *buf, size_t bufsz) {
+    enum { NORMAL, IN_ESC, IN_CSI} state = NORMAL;
+    char *dest = buf, *last = &buf[bufsz-1];
+    *dest = 0; *last = 0;
+
+    for (; *title && dest < last; title++) {
+        const char ch = *title;
+        switch (state) {
+            case NORMAL: {
+                if (ch == 0x1b) { state = IN_ESC; }
+                else *(dest++) = ch;
+            } break;
+            case IN_ESC: {
+                if (ch == '[') { state = IN_CSI; }
+                else { state = NORMAL; }
+            } break;
+            case IN_CSI: {
+                if (!(('0' <= ch && ch <= '9') || ch == ';' || ch == ':')) state = NORMAL;
+            } break;
+        }
+    }
+    *dest = 0;
+}
+
 void
 set_os_window_title(OSWindow *w, const char *title) {
-    glfwSetWindowTitle(w->handle, title);
+    static char buf[2048];
+    strip_csi_(title, buf, arraysz(buf));
+    glfwSetWindowTitle(w->handle, buf);
 }
 
 void
@@ -1398,8 +1506,14 @@ dbus_send_notification(PyObject *self UNUSED, PyObject *args) {
     unsigned long long notification_id = glfwDBusUserNotify(app_name, icon, summary, body, action_name, timeout, dbus_notification_created_callback, NULL);
     return PyLong_FromUnsignedLongLong(notification_id);
 }
+
 #endif
 
+static PyObject*
+get_click_interval(PyObject *self UNUSED, PyObject *args UNUSED) {
+    return PyFloat_FromDouble(monotonic_t_to_s_double(OPT(click_interval)));
+}
+
 id_type
 add_main_loop_timer(monotonic_t interval, bool repeats, timer_callback_fun callback, void *callback_data, timer_callback_fun free_callback) {
     return glfwAddTimer(interval, repeats, callback, callback_data, free_callback);
@@ -1429,6 +1543,17 @@ stop_main_loop(void) {
     glfwStopMainLoop();
 }
 
+static PyObject*
+strip_csi(PyObject *self UNUSED, PyObject *src) {
+    if (!PyUnicode_Check(src)) { PyErr_SetString(PyExc_TypeError, "Unicode string expected"); return NULL; }
+    Py_ssize_t sz;
+    const char *title = PyUnicode_AsUTF8AndSize(src, &sz);
+    if (!title) return NULL;
+    FREE_AFTER_FUNCTION char *buf = malloc(sz + 1);
+    if (!buf) { return PyErr_NoMemory(); }
+    strip_csi_(title, buf, sz + 1);
+    return PyUnicode_FromString(buf);
+}
 
 // Boilerplate {{{
 
@@ -1446,8 +1571,10 @@ static PyMethodDef module_methods[] = {
     METHODB(glfw_window_hint, METH_VARARGS),
     METHODB(get_primary_selection, METH_NOARGS),
     METHODB(x11_display, METH_NOARGS),
+    METHODB(get_click_interval, METH_NOARGS),
     METHODB(x11_window_id, METH_O),
     METHODB(set_primary_selection, METH_VARARGS),
+    METHODB(strip_csi, METH_O),
 #ifndef __APPLE__
     METHODB(dbus_send_notification, METH_VARARGS),
 #endif
