@@ -79,9 +79,9 @@ active_window(void) {
     return NULL;
 }
 
-static void
-update_ime_position(OSWindow *os_window, Window* w, Screen *screen) {
-    unsigned int cell_width = os_window->fonts_data->cell_width, cell_height = os_window->fonts_data->cell_height;
+void
+update_ime_position(Window* w, Screen *screen) {
+    unsigned int cell_width = global_state.callback_os_window->fonts_data->cell_width, cell_height = global_state.callback_os_window->fonts_data->cell_height;
     unsigned int left = w->geometry.left, top = w->geometry.top;
     left += screen->cursor->x * cell_width;
     top += screen->cursor->y * cell_height;
@@ -130,7 +130,7 @@ on_key_input(GLFWkeyevent *ev) {
 
     switch(ev->ime_state) {
         case GLFW_IME_PREEDIT_CHANGED:
-            update_ime_position(global_state.callback_os_window, w, screen);
+            update_ime_position(w, screen);
             screen_draw_overlay_text(screen, text);
             debug("updated pre-edit text: '%s'\n", text);
             return;
@@ -145,7 +145,7 @@ on_key_input(GLFWkeyevent *ev) {
             // for macOS, update ime position on every key input
             // because the position is required before next input
 #if defined(__APPLE__)
-            update_ime_position(global_state.callback_os_window, w, screen);
+            update_ime_position(w, screen);
 #endif
             break;
         default:
@@ -204,10 +204,10 @@ on_key_input(GLFWkeyevent *ev) {
     int size = encode_glfw_key_event(ev, screen->modes.mDECCKM, screen_current_key_encoding_flags(screen), encoded_key);
     if (size == SEND_TEXT_TO_CHILD) {
         schedule_write_to_child(w->id, 1, text, strlen(text));
-        debug("sent text to child\n");
+        debug("sent key as text to child\n");
     } else if (size > 0) {
         schedule_write_to_child(w->id, 1, encoded_key, size);
-        debug("sent key to child\n");
+        debug("sent encoded key to child\n");
     } else {
         debug("ignoring as keyboard mode does not support encoding this event\n");
     }
