@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import locale
@@ -28,8 +27,10 @@ from .fast_data_types import (
 from .fonts.box_drawing import set_scale
 from .fonts.render import set_font_family
 from .options.types import Options
+from .options.utils import DELETE_ENV_VAR
 from .os_window_size import initial_window_size_func
 from .session import get_os_window_sizing_data
+from .shell_integration import setup_shell_integration
 from .types import SingleKey
 from .utils import (
     detach, expandvars, log_error, single_instance,
@@ -50,18 +51,21 @@ def set_custom_ibeam_cursor() -> None:
     try:
         set_custom_cursor(GLFW_IBEAM_CURSOR, images, 4, 8)
     except Exception as e:
-        log_error('Failed to set custom beam cursor with error: {}'.format(e))
+        log_error(f'Failed to set custom beam cursor with error: {e}')
 
 
 def talk_to_instance(args: CLIOptions) -> None:
     import json
     import socket
+    stdin = ''
+    if args.session == '-':
+        stdin = sys.stdin.read()
     data = {'cmd': 'new_instance', 'args': tuple(sys.argv),
             'startup_id': os.environ.get('DESKTOP_STARTUP_ID'),
-            'cwd': os.getcwd()}
+            'cwd': os.getcwd(), 'stdin': stdin}
     notify_socket = None
     if args.wait_for_single_instance_window_close:
-        address = '\0{}-os-window-close-notify-{}-{}'.format(appname, os.getpid(), os.geteuid())
+        address = f'\0{appname}-os-window-close-notify-{os.getpid()}-{os.geteuid()}'
         notify_socket = socket.socket(family=socket.AF_UNIX)
         try:
             notify_socket.bind(address)
@@ -104,12 +108,17 @@ def init_glfw(opts: Options, debug_keyboard: bool = False, debug_rendering: bool
     return glfw_module
 
 
-def get_macos_shortcut_for(opts: Options, function: str = 'new_os_window', args: Tuple = (), lookup_name: str = '') -> Optional[SingleKey]:
+def get_macos_shortcut_for(
+    func_map: Dict[Tuple[str, ...], List[SingleKey]], defn: str = 'new_os_window', lookup_name: str = ''
+) -> Optional[SingleKey]:
+    # for maximum robustness we should use opts.alias_map to resolve
+    # aliases however this requires parsing everything on startup which could
+    # be potentially slow. Lets just hope the user doesnt alias these
+    # functions.
     ans = None
     candidates = []
-    for k, v in opts.keymap.items():
-        if v.func == function and v.args == args:
-            candidates.append(k)
+    qkey = tuple(defn.split())
+    candidates = func_map[qkey]
     if candidates:
         from .fast_data_types import cocoa_set_global_shortcut
         alt_mods = GLFW_MOD_ALT, GLFW_MOD_ALT | GLFW_MOD_SHIFT
@@ -120,7 +129,7 @@ def get_macos_shortcut_for(opts: Options, function: str = 'new_os_window', args:
                 # presumably because Apple reserves them for IME, see
                 # https://github.com/kovidgoyal/kitty/issues/3515
                 continue
-            if cocoa_set_global_shortcut(lookup_name or function, candidate[0], candidate[2]):
+            if cocoa_set_global_shortcut(lookup_name or qkey[0], candidate[0], candidate[2]):
                 ans = candidate
                 break
     return ans
@@ -135,29 +144,34 @@ def set_x11_window_icon() -> None:
 def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
     global_shortcuts: Dict[str, SingleKey] = {}
     if is_macos:
+        from collections import defaultdict
+        func_map = defaultdict(list)
+        for k, v in opts.keymap.items():
+            parts = tuple(v.split())
+            func_map[parts].append(k)
+
         for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab',
                    'next_tab', 'new_tab', 'new_window', 'close_window'):
-            val = get_macos_shortcut_for(opts, ac)
+            val = get_macos_shortcut_for(func_map, ac)
             if val is not None:
                 global_shortcuts[ac] = val
-        val = get_macos_shortcut_for(opts, 'clear_terminal', args=('reset', True), lookup_name='reset_terminal')
+        val = get_macos_shortcut_for(func_map, 'clear_terminal reset active', lookup_name='reset_terminal')
         if val is not None:
             global_shortcuts['reset_terminal'] = val
-        val = get_macos_shortcut_for(opts, 'load_config_file', args=(), lookup_name='reload_config')
+        val = get_macos_shortcut_for(func_map, 'load_config_file', lookup_name='reload_config')
         if val is not None:
             global_shortcuts['reload_config'] = val
     if is_macos and opts.macos_custom_beam_cursor:
         set_custom_ibeam_cursor()
     if not is_wayland() and not is_macos:  # no window icons on wayland
         set_x11_window_icon()
-    load_shader_programs.use_selection_fg = opts.selection_foreground is not None
     with cached_values_for(run_app.cached_values_name) as cached_values:
         with startup_notification_handler(extra_callback=run_app.first_window_callback) as pre_show_callback:
             window_id = create_os_window(
                     run_app.initial_window_size_func(get_os_window_sizing_data(opts), cached_values),
                     pre_show_callback,
                     args.title or appname, args.name or args.cls or appname,
-                    args.cls or appname, load_all_shaders)
+                    args.cls or appname, load_all_shaders, disallow_override_title=bool(args.title))
         boss = Boss(opts, args, cached_values, global_shortcuts)
         boss.start(window_id)
         if bad_lines:
@@ -185,7 +199,9 @@ class AppRunner:
             set_options(None)
             free_font_data()  # must free font data before glfw/freetype/fontconfig/opengl etc are finalized
             if is_macos:
-                from kitty.fast_data_types import cocoa_set_notification_activated_callback
+                from kitty.fast_data_types import (
+                    cocoa_set_notification_activated_callback
+                )
                 cocoa_set_notification_activated_callback(None)
 
 
@@ -195,10 +211,15 @@ run_app = AppRunner()
 def ensure_macos_locale() -> None:
     # Ensure the LANG env var is set. See
     # https://github.com/kovidgoyal/kitty/issues/90
-    from .fast_data_types import cocoa_get_lang
+    from .fast_data_types import cocoa_get_lang, locale_is_valid
     if 'LANG' not in os.environ:
         lang = cocoa_get_lang()
         if lang is not None:
+            if not locale_is_valid(lang):
+                if lang.startswith('en_'):
+                    lang = 'en_US'
+                else:
+                    log_error(f'Could not set LANG Cocoa returns language as: {lang}')
             os.environ['LANG'] = lang + '.UTF-8'
 
 
@@ -257,6 +278,37 @@ def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
     return listen_on
 
 
+def safe_samefile(a: str, b: str) -> bool:
+    with suppress(OSError):
+        return os.path.samefile(a, b)
+    return os.path.abspath(os.path.realpath(a)) == os.path.abspath(os.path.realpath(b))
+
+
+def prepend_if_not_present(path: str, paths_serialized: str) -> str:
+    # prepend a path only if path/kitty is not already present, even as a symlink
+    pq = os.path.join(path, 'kitty')
+    for candidate in paths_serialized.split(os.pathsep):
+        q = os.path.join(candidate, 'kitty')
+        if safe_samefile(q, pq):
+            return paths_serialized
+    return path + os.pathsep + paths_serialized
+
+
+def ensure_kitty_in_path() -> None:
+    # Ensure the correct kitty is in PATH
+    rpath = sys._xoptions.get('bundle_exe_dir')
+    if not rpath:
+        return
+    if rpath:
+        modify_path = is_macos or getattr(sys, 'frozen', False) or sys._xoptions.get('kitty_from_source') == '1'
+        existing = shutil.which('kitty')
+        if modify_path or not existing:
+            env_path = os.environ.get('PATH', '')
+            correct_kitty = os.path.join(rpath, 'kitty')
+            if not existing or not safe_samefile(existing, correct_kitty):
+                os.environ['PATH'] = prepend_if_not_present(rpath, env_path)
+
+
 def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     from_config_file = False
     if not cli_opts.listen_on and opts.listen_on.startswith('unix:'):
@@ -265,7 +317,17 @@ def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     if cli_opts.listen_on and opts.allow_remote_control != 'n':
         cli_opts.listen_on = expand_listen_on(cli_opts.listen_on, from_config_file)
         os.environ['KITTY_LISTEN_ON'] = cli_opts.listen_on
-    set_default_env(opts.env.copy())
+    env = opts.env.copy()
+    ensure_kitty_in_path()
+    setup_shell_integration(opts, env)
+    kitty_path = shutil.which('kitty')
+    if kitty_path:
+        child_path = env.get('PATH')
+        # if child_path is None it will be inherited from os.environ,
+        # the other values mean the user doesn't want a PATH
+        if child_path not in ('', DELETE_ENV_VAR) and child_path is not None:
+            env['PATH'] = prepend_if_not_present(os.path.dirname(kitty_path), env['PATH'])
+    set_default_env(env)
 
 
 def set_locale() -> None:
@@ -292,15 +354,6 @@ def _main() -> None:
     except Exception:
         log_error('Failed to set locale, ignoring')
 
-    # Ensure the correct kitty is in PATH
-    rpath = sys._xoptions.get('bundle_exe_dir')
-    if rpath:
-        modify_path = is_macos or getattr(sys, 'frozen', False) or sys._xoptions.get('kitty_from_source') == '1'
-        if modify_path or not shutil.which('kitty'):
-            existing_paths = list(filter(None, os.environ.get('PATH', '').split(os.pathsep)))
-            existing_paths.insert(0, rpath)
-            os.environ['PATH'] = os.pathsep.join(existing_paths)
-
     args = sys.argv[1:]
     if is_macos and os.environ.pop('KITTY_LAUNCHED_BY_LAUNCH_SERVICES', None) == '1':
         os.chdir(os.path.expanduser('~'))
@@ -331,6 +384,10 @@ def _main() -> None:
     opts = create_opts(cli_opts, accumulate_bad_lines=bad_lines)
     init_glfw(opts, cli_opts.debug_keyboard, cli_opts.debug_rendering)
     setup_environment(opts, cli_opts)
+    if cli_opts.watcher:
+        from .window import global_watchers
+        global_watchers.set_extra(cli_opts.watcher)
+        log_error('The --watcher command line option has been deprecated in favor of using the watcher option in kitty.conf')
     try:
         with setup_profiling(cli_opts):
             # Avoid needing to launch threads to reap zombies
