@@ -66,7 +66,7 @@ encode_button(unsigned int button) {
 static char mouse_event_buf[64];
 
 static int
-encode_mouse_event_impl(unsigned int x, unsigned int y, int mouse_tracking_protocol, int button, MouseAction action, int mods) {
+encode_mouse_event_impl(const MousePosition *mpos, int mouse_tracking_protocol, int button, MouseAction action, int mods) {
     unsigned int cb = 0;
     if (action == MOVE) {
         cb = 3;
@@ -75,11 +75,16 @@ encode_mouse_event_impl(unsigned int x, unsigned int y, int mouse_tracking_proto
         if (cb == UINT_MAX) return 0;
     }
     if (action == DRAG || action == MOVE) cb |= MOTION_INDICATOR;
-    else if (action == RELEASE && mouse_tracking_protocol != SGR_PROTOCOL) cb = 3;
+    else if (action == RELEASE && mouse_tracking_protocol < SGR_PROTOCOL) cb = 3;
     if (mods & GLFW_MOD_SHIFT) cb |= SHIFT_INDICATOR;
     if (mods & GLFW_MOD_ALT) cb |= ALT_INDICATOR;
     if (mods & GLFW_MOD_CONTROL) cb |= CONTROL_INDICATOR;
+    int x = mpos->cell_x + 1, y = mpos->cell_y + 1;
     switch(mouse_tracking_protocol) {
+        case SGR_PIXEL_PROTOCOL:
+            x = (int)round(mpos->global_x);
+            y = (int)round(mpos->global_y);
+            /* fallthrough */
         case SGR_PROTOCOL:
             return snprintf(mouse_event_buf, sizeof(mouse_event_buf), "<%d;%d;%d%s", cb, x, y, action == RELEASE ? "m" : "M");
             break;
@@ -106,13 +111,22 @@ encode_mouse_event_impl(unsigned int x, unsigned int y, int mouse_tracking_proto
 
 static int
 encode_mouse_event(Window *w, int button, MouseAction action, int mods) {
-    unsigned int x = w->mouse_pos.cell_x + 1, y = w->mouse_pos.cell_y + 1; // 1 based indexing
     Screen *screen = w->render_data.screen;
-    return encode_mouse_event_impl(x, y, screen->modes.mouse_tracking_protocol, button, action, mods);
+    return encode_mouse_event_impl(&w->mouse_pos, screen->modes.mouse_tracking_protocol, button, action, mods);
 }
 
 static int
 encode_mouse_button(Window *w, int button, MouseAction action, int mods) {
+    if (button == GLFW_MOUSE_BUTTON_LEFT) {
+        switch(action) {
+            case PRESS:
+                global_state.tracked_drag_in_window = w->id; break;
+            case RELEASE:
+                global_state.tracked_drag_in_window = 0; break;
+            default:
+                break;
+        }
+    }
     return encode_mouse_event(w, button_map(button), action, mods);
 }
 
@@ -207,11 +221,11 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
     double mouse_x = global_state.callback_os_window->mouse_x;
     double mouse_y = global_state.callback_os_window->mouse_y;
     double left = window_left(w), top = window_top(w), right = window_right(w), bottom = window_bottom(w);
+    w->mouse_pos.global_x = mouse_x - left; w->mouse_pos.global_y = mouse_y - top;
     if (clamp_to_window) {
         mouse_x = MIN(MAX(mouse_x, left), right);
         mouse_y = MIN(MAX(mouse_y, top), bottom);
     }
-    w->mouse_pos.x = mouse_x - left; w->mouse_pos.y = mouse_y - top;
     if (mouse_x < left || mouse_y < top || mouse_x > right || mouse_y > bottom) return false;
     if (mouse_x >= g->right) {
         qx = screen->columns - 1;
@@ -337,7 +351,7 @@ HANDLER(handle_move_event) {
     if (handle_in_kitty) {
         handle_mouse_movement_in_kitty(w, button, mouse_cell_changed | cell_half_changed);
     } else {
-        if (!mouse_cell_changed) return;
+        if (!mouse_cell_changed && screen->modes.mouse_tracking_protocol != SGR_PIXEL_PROTOCOL) return;
         int sz = encode_mouse_button(w, MAX(0, button), button >=0 ? DRAG : MOVE, modifiers);
         if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, CSI, mouse_event_buf); }
     }
@@ -360,7 +374,7 @@ release_is_click(Window *w, int button) {
     ClickQueue *q = &w->click_queues[button];
     double click_allowed_radius = 1.2 * (global_state.callback_os_window ? global_state.callback_os_window->fonts_data->cell_height : 20);
     monotonic_t now = monotonic();
-    return (q->length > 0 && distance(N(1).x, N(1).y, w->mouse_pos.x, w->mouse_pos.y) <= click_allowed_radius && now - N(1).at < OPT(click_interval));
+    return (q->length > 0 && distance(N(1).x, N(1).y, MAX(0, w->mouse_pos.global_x), MAX(0, w->mouse_pos.global_y)) <= click_allowed_radius && now - N(1).at < OPT(click_interval));
 }
 
 static unsigned
@@ -392,7 +406,7 @@ add_press(Window *w, int button, int modifiers) {
     ClickQueue *q = &w->click_queues[button];
     if (q->length == CLICK_QUEUE_SZ) { memmove(q->clicks, q->clicks + 1, sizeof(Click) * (CLICK_QUEUE_SZ - 1)); q->length--; }
     monotonic_t now = monotonic();
-    N(0).at = now; N(0).button = button; N(0).modifiers = modifiers; N(0).x = w->mouse_pos.x; N(0).y = w->mouse_pos.y;
+    N(0).at = now; N(0).button = button; N(0).modifiers = modifiers; N(0).x = MAX(0, w->mouse_pos.global_x); N(0).y = MAX(0, w->mouse_pos.global_y);
     q->length++;
     Screen *screen = w->render_data.screen;
     int count = multi_click_count(w, button);
@@ -403,18 +417,40 @@ add_press(Window *w, int button, int modifiers) {
 }
 #undef N
 
-void
+bool
 mouse_open_url(Window *w) {
     Screen *screen = w->render_data.screen;
     detect_url(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
-    screen_open_url(screen);
+    return screen_open_url(screen);
 }
 
+bool
+mouse_set_last_visited_cmd_output(Window *w) {
+    Screen *screen = w->render_data.screen;
+    return screen_set_last_visited_prompt(screen, w->mouse_pos.cell_y);
+}
+
+bool
+mouse_select_cmd_output(Window *w) {
+    Screen *screen = w->render_data.screen;
+    return screen_select_cmd_output(screen, w->mouse_pos.cell_y);
+}
+
+bool
+move_cursor_to_mouse_if_at_shell_prompt(Window *w) {
+    Screen *screen = w->render_data.screen;
+    int y = screen_cursor_at_a_shell_prompt(screen);
+    if (y < 0 || (unsigned)y > w->mouse_pos.cell_y) return false;
+    return screen_fake_move_cursor_to_position(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
+}
+
+
 typedef struct PendingClick {
     id_type window_id;
     int button, count, modifiers;
     bool grabbed;
     monotonic_t at;
+    MousePosition mouse_pos;
 } PendingClick;
 
 static void
@@ -426,7 +462,10 @@ send_pending_click_to_window(Window *w, void *data) {
     ClickQueue *q = &w->click_queues[pc->button];
     // only send click if no presses have happened since the release that triggered the click
     if (q->length && q->clicks[q->length - 1].at <= pc->at) {
+        MousePosition current_pos = w->mouse_pos;
+        w->mouse_pos = pc->mouse_pos;
         dispatch_mouse_event(w, pc->button, pc->count, pc->modifiers, pc->grabbed);
+        w->mouse_pos = current_pos;
     }
 }
 
@@ -438,6 +477,7 @@ dispatch_possible_click(Window *w, int button, int modifiers) {
         PendingClick *pc = calloc(sizeof(PendingClick), 1);
         if (pc) {
             pc->window_id = w->id;
+            pc->mouse_pos = w->mouse_pos;
             pc->at = monotonic();
             pc->button = button;
             pc->count = count == 2 ? -3 : -2;
@@ -477,6 +517,7 @@ currently_pressed_button(void) {
 
 HANDLER(handle_event) {
     modifiers &= ~GLFW_LOCK_MASK;
+    set_mouse_cursor_for_screen(w->render_data.screen);
     if (button == -1) {
         button = currently_pressed_button();
         handle_move_event(w, button, modifiers, window_idx);
@@ -486,13 +527,10 @@ HANDLER(handle_event) {
 }
 
 static void
-handle_tab_bar_mouse(int button, int UNUSED modifiers) {
-    static monotonic_t last_click_at = 0;
-    if (button != GLFW_MOUSE_BUTTON_LEFT || !global_state.callback_os_window->mouse_button_pressed[button]) return;
-    monotonic_t now = monotonic();
-    bool is_double = now - last_click_at <= OPT(click_interval);
-    last_click_at = is_double ? 0 : now;
-    call_boss(activate_tab_at, "KdO", global_state.callback_os_window->id, global_state.callback_os_window->mouse_x, is_double ? Py_True : Py_False);
+handle_tab_bar_mouse(int button, int modifiers, int action) {
+    if (button > -1) {  // dont report motion events, as they are expensive and useless
+        call_boss(handle_click_on_tab, "Kdiii", global_state.callback_os_window->id, global_state.callback_os_window->mouse_x, button, modifiers, action);
+    }
 }
 
 static bool
@@ -517,8 +555,16 @@ static Window*
 window_for_event(unsigned int *window_idx, bool *in_tab_bar) {
     Region central, tab_bar;
     os_window_regions(global_state.callback_os_window, &central, &tab_bar);
-    *in_tab_bar = !mouse_in_region(&central);
-    if (!*in_tab_bar && global_state.callback_os_window->num_tabs > 0) {
+    const bool in_central = mouse_in_region(&central);
+    *in_tab_bar = false;
+    const OSWindow* w = global_state.callback_os_window;
+    if (!in_central) {
+        if (
+                (tab_bar.top < central.top && w->mouse_y <= central.top) ||
+                (tab_bar.bottom > central.bottom && w->mouse_y >= central.bottom)
+           ) *in_tab_bar = true;
+    }
+    if (in_central && global_state.callback_os_window->num_tabs > 0) {
         Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
         for (unsigned int i = 0; i < t->num_windows; i++) {
             if (contains_mouse(t->windows + i) && t->windows[i].render_data.screen) {
@@ -634,22 +680,31 @@ mouse_selection(Window *w, int code, int button) {
 
 
 void
-mouse_event(int button, int modifiers, int action) {
+mouse_event(const int button, int modifiers, int action) {
     MouseShape old_cursor = mouse_cursor_shape;
     bool in_tab_bar;
     unsigned int window_idx = 0;
     Window *w = NULL;
     debug("%s mouse_button: %d %s", action == GLFW_RELEASE ? "\x1b[32mRelease\x1b[m" : (button < 0 ? "\x1b[36mMove\x1b[m" : "\x1b[31mPress\x1b[m"), button, format_mods(modifiers));
+    if (global_state.redirect_mouse_handling) {
+        w = window_for_event(&window_idx, &in_tab_bar);
+        call_boss(mouse_event, "OK iiii dd",
+                (in_tab_bar ? Py_True : Py_False), (w ? w->id : 0),
+                action, modifiers, button, currently_pressed_button(),
+                global_state.callback_os_window->mouse_x, global_state.callback_os_window->mouse_y
+        );
+        debug("mouse handling redirected\n");
+        return;
+    }
     if (global_state.active_drag_in_window) {
         if (button == -1) {  // drag move
             w = window_for_id(global_state.active_drag_in_window);
             if (w) {
-                button = currently_pressed_button();
-                if (button == global_state.active_drag_button) {
+                if (currently_pressed_button() == global_state.active_drag_button) {
                     clamp_to_window = true;
                     Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
                     for (window_idx = 0; window_idx < t->num_windows && t->windows[window_idx].id != w->id; window_idx++);
-                    handle_move_event(w, button, modifiers, window_idx);
+                    handle_move_event(w, currently_pressed_button(), modifiers, window_idx);
                     clamp_to_window = false;
                     debug("handled as drag move\n");
                     return;
@@ -666,10 +721,40 @@ mouse_event(int button, int modifiers, int action) {
             }
         }
     }
+    if (global_state.tracked_drag_in_window) {
+        if (button == -1) {  // drag move
+            w = window_for_id(global_state.tracked_drag_in_window);
+            if (w) {
+                if (currently_pressed_button() == GLFW_MOUSE_BUTTON_LEFT) {
+                    if (w->render_data.screen->modes.mouse_tracking_mode >= MOTION_MODE && w->render_data.screen->modes.mouse_tracking_protocol == SGR_PIXEL_PROTOCOL) {
+                        clamp_to_window = true;
+                        Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+                        for (window_idx = 0; window_idx < t->num_windows && t->windows[window_idx].id != w->id; window_idx++);
+                        handle_move_event(w, button, modifiers, window_idx);
+                        clamp_to_window = false;
+                        debug("sent to child as drag move\n");
+                        return;
+                    }
+                }
+            }
+        } else if (action == GLFW_RELEASE && button == GLFW_MOUSE_BUTTON_LEFT) {
+            w = window_for_id(global_state.tracked_drag_in_window);
+            if (w && w->render_data.screen->modes.mouse_tracking_mode >= MOTION_MODE && w->render_data.screen->modes.mouse_tracking_protocol == SGR_PIXEL_PROTOCOL) {
+                global_state.tracked_drag_in_window = 0;
+                clamp_to_window = true;
+                Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+                for (window_idx = 0; window_idx < t->num_windows && t->windows[window_idx].id != w->id; window_idx++);
+                handle_button_event(w, button, modifiers, window_idx);
+                clamp_to_window = false;
+                debug("sent to child as drag end\n");
+                return;
+            }
+        }
+    }
     w = window_for_event(&window_idx, &in_tab_bar);
     if (in_tab_bar) {
         mouse_cursor_shape = HAND;
-        handle_tab_bar_mouse(button, modifiers);
+        handle_tab_bar_mouse(button, modifiers, action);
         debug("handled by tab bar\n");
     } else if (w) {
         debug("grabbed: %d\n", w->render_data.screen->modes.mouse_tracking_mode != 0);
@@ -684,9 +769,7 @@ mouse_event(int button, int modifiers, int action) {
             clamp_to_window = false;
         } else debug("no window for event\n");
     } else debug("\n");
-    if (mouse_cursor_shape != old_cursor) {
-        set_mouse_cursor(mouse_cursor_shape);
-    }
+    if (mouse_cursor_shape != old_cursor) set_mouse_cursor(mouse_cursor_shape);
 }
 
 void
@@ -695,15 +778,34 @@ scroll_event(double UNUSED xoffset, double yoffset, int flags, int modifiers) {
     static id_type window_for_momentum_scroll = 0;
     static bool main_screen_for_momentum_scroll = false;
     unsigned int window_idx = 0;
+    // allow scroll events even if window is not currently focused (in
+    // which case on some platforms such as macOS the mouse location is zeroed so
+    // window_for_event() does not work).
+    OSWindow *osw = global_state.callback_os_window;
+    if (!osw->is_focused && osw->handle) {
+        double mouse_x, mouse_y;
+        glfwGetCursorPos((GLFWwindow*)osw->handle, &mouse_x, &mouse_y);
+        osw->mouse_x = mouse_x * osw->viewport_x_ratio;
+        osw->mouse_y = mouse_y * osw->viewport_y_ratio;
+    }
     Window *w = window_for_event(&window_idx, &in_tab_bar);
     if (!w && !in_tab_bar) {
-        // allow scroll events even if window is not currently focused (in
-        // which case on some platforms such as macOS the mouse location is zeroed so
-        // window_for_event() does not work).
-        Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+        // fallback to last active window
+        Tab *t = osw->tabs + osw->active_tab;
         if (t) w = t->windows + t->active_window;
     }
     if (!w) return;
+    // Also update mouse cursor position while kitty OS window is not focused.
+    // Allows scroll events to be delivered to the child with correct pointer co-ordinates even when
+    // the window is not focused on macOS
+    if (!osw->is_focused) {
+        unsigned int x = 0, y = 0;
+        bool in_left_half_of_cell;
+        if (cell_for_pos(w, &x, &y, &in_left_half_of_cell, osw)) {
+            w->mouse_pos.cell_x = x; w->mouse_pos.cell_y = y;
+            w->mouse_pos.in_left_half_of_cell = in_left_half_of_cell;
+        }
+    }
     Screen *screen = w->render_data.screen;
 
     enum MomentumData { NoMomentumData, MomentumPhaseBegan, MomentumPhaseStationary, MomentumPhaseActive, MomentumPhaseEnded, MomentumPhaseCancelled, MomentumPhaseMayBegin };
@@ -785,7 +887,8 @@ send_mouse_event(PyObject *self UNUSED, PyObject *args) {
 
     MouseTrackingMode mode = screen->modes.mouse_tracking_mode;
     if (mode == ANY_MODE || (mode == MOTION_MODE && action != MOVE) || (mode == BUTTON_MODE && (action == PRESS || action == RELEASE))) {
-        int sz = encode_mouse_event_impl(x + 1, y + 1, screen->modes.mouse_tracking_protocol, button, action, mods);
+        MousePosition mpos = {.cell_x = x, .cell_y = y};
+        int sz = encode_mouse_event_impl(&mpos, screen->modes.mouse_tracking_protocol, button, action, mods);
         if (sz > 0) {
             mouse_event_buf[sz] = 0;
             write_escape_code_to_child(screen, CSI, mouse_event_buf);
@@ -800,7 +903,8 @@ test_encode_mouse(PyObject *self UNUSED, PyObject *args) {
     unsigned int x, y;
     int mouse_tracking_protocol, button, action, mods;
     if (!PyArg_ParseTuple(args, "IIiiii", &x, &y, &mouse_tracking_protocol, &button, &action, &mods)) return NULL;
-    int sz = encode_mouse_event_impl(x, y, mouse_tracking_protocol, button, action, mods);
+    MousePosition mpos = {.cell_x = x - 1, .cell_y = y - 1};
+    int sz = encode_mouse_event_impl(&mpos, mouse_tracking_protocol, button, action, mods);
     return PyUnicode_FromStringAndSize(mouse_event_buf, sz);
 }
 
@@ -825,7 +929,7 @@ send_mock_mouse_event_to_window(PyObject *self UNUSED, PyObject *args) {
     if (!w) return NULL;
     if (clear_clicks) clear_click_queue(w, button);
     bool mouse_cell_changed = x != w->mouse_pos.cell_x || y != w->mouse_pos.cell_y || w->mouse_pos.in_left_half_of_cell != in_left_half_of_cell;
-    w->mouse_pos.x = 10 * x; w->mouse_pos.y = 20 * y;
+    w->mouse_pos.global_x = 10 * x; w->mouse_pos.global_y = 20 * y;
     w->mouse_pos.cell_x = x; w->mouse_pos.cell_y = y;
     w->mouse_pos.in_left_half_of_cell = in_left_half_of_cell;
     static int last_button_pressed = GLFW_MOUSE_BUTTON_LEFT;
