@@ -1,47 +1,50 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>
 
 
-import os
 import re
 import sys
+from functools import lru_cache
 from typing import (
-    Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Sequence, Tuple,
-    Union
+    Any, Callable, Container, Dict, FrozenSet, Iterable, Iterator, List,
+    NamedTuple, Optional, Sequence, Tuple, Union
 )
 
 import kitty.fast_data_types as defines
 from kitty.conf.utils import (
-    KeyAction, key_func, positive_float, positive_int, python_string, to_bool,
-    to_cmdline, to_color, uniq, unit_float
+    CurrentlyParsing, KeyAction, KeyFuncWrapper, currently_parsing,
+    positive_float, positive_int, python_string, to_bool, to_cmdline, to_color,
+    uniq, unit_float
+)
+from kitty.constants import is_macos
+from kitty.fast_data_types import (
+    CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE, Color
 )
-from kitty.constants import config_dir, is_macos
-from kitty.fast_data_types import CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE
 from kitty.fonts import FontFeature
 from kitty.key_names import (
     character_key_name_aliases, functional_key_name_aliases,
     get_key_name_lookup
 )
-from kitty.rgb import Color, color_as_int
+from kitty.rgb import color_as_int
 from kitty.types import FloatEdges, MouseEvent, SingleKey
-from kitty.utils import expandvars, log_error
+from kitty.utils import expandvars, log_error, resolve_abs_or_config_path
 
-KeyMap = Dict[SingleKey, KeyAction]
-MouseMap = Dict[MouseEvent, KeyAction]
+KeyMap = Dict[SingleKey, str]
+MouseMap = Dict[MouseEvent, str]
 KeySequence = Tuple[SingleKey, ...]
-SubSequenceMap = Dict[KeySequence, KeyAction]
+SubSequenceMap = Dict[KeySequence, str]
 SequenceMap = Dict[SingleKey, SubSequenceMap]
 MINIMUM_FONT_SIZE = 4
 default_tab_separator = ' ┇'
-mod_map = {'CTRL': 'CONTROL', 'CMD': 'SUPER', '⌘': 'SUPER',
-           '⌥': 'ALT', 'OPTION': 'ALT', 'KITTY_MOD': 'KITTY'}
+mod_map = {'⌃': 'CONTROL', 'CTRL': 'CONTROL', '⇧': 'SHIFT', '⌥': 'ALT', 'OPTION': 'ALT', 'OPT': 'ALT',
+           '⌘': 'SUPER', 'COMMAND': 'SUPER', 'CMD': 'SUPER', 'KITTY_MOD': 'KITTY'}
 character_key_name_aliases_with_ascii_lowercase: Dict[str, str] = character_key_name_aliases.copy()
 for x in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
     character_key_name_aliases_with_ascii_lowercase[x] = x.lower()
 sequence_sep = '>'
-func_with_args, args_funcs = key_func()
 FuncArgsType = Tuple[str, Sequence[Any]]
+func_with_args = KeyFuncWrapper[FuncArgsType]()
+DELETE_ENV_VAR = '_delete_this_env_var_'
 
 
 class InvalidMods(ValueError):
@@ -52,20 +55,12 @@ class InvalidMods(ValueError):
 @func_with_args(
     'pass_selection_to_program', 'new_window', 'new_tab', 'new_os_window',
     'new_window_with_cwd', 'new_tab_with_cwd', 'new_os_window_with_cwd',
-    'launch'
+    'launch', 'mouse_handle_click'
     )
 def shlex_parse(func: str, rest: str) -> FuncArgsType:
     return func, to_cmdline(rest)
 
 
-@func_with_args('combine')
-def combine_parse(func: str, rest: str) -> FuncArgsType:
-    sep, rest = rest.split(maxsplit=1)
-    parts = re.split(r'\s*' + re.escape(sep) + r'\s*', rest)
-    args = tuple(map(parse_key_action, filter(None, parts)))
-    return func, args
-
-
 def parse_send_text_bytes(text: str) -> bytes:
     return python_string(text).encode('utf-8')
 
@@ -91,7 +86,7 @@ def kitten_parse(func: str, rest: str) -> FuncArgsType:
     else:
         args = rest.split(maxsplit=2)[1:]
         func = 'kitten'
-    return func, args
+    return func, [args[0]] + (to_cmdline(args[1]) if len(args) > 1 else [])
 
 
 @func_with_args('goto_tab')
@@ -103,15 +98,21 @@ def goto_tab_parse(func: str, rest: str) -> FuncArgsType:
 @func_with_args('detach_window')
 def detach_window_parse(func: str, rest: str) -> FuncArgsType:
     if rest not in ('new', 'new-tab', 'ask', 'tab-prev', 'tab-left', 'tab-right'):
-        log_error('Ignoring invalid detach_window argument: {}'.format(rest))
+        log_error(f'Ignoring invalid detach_window argument: {rest}')
         rest = 'new'
     return func, (rest,)
 
 
+@func_with_args('close_window_with_confirmation')
+def close_window_with_confirmation(func: str, rest: str) -> FuncArgsType:
+    ignore_shell = rest == 'ignore-shell'
+    return func, (ignore_shell,)
+
+
 @func_with_args('detach_tab')
 def detach_tab_parse(func: str, rest: str) -> FuncArgsType:
     if rest not in ('new', 'ask'):
-        log_error('Ignoring invalid detach_tab argument: {}'.format(rest))
+        log_error(f'Ignoring invalid detach_tab argument: {rest}')
         rest = 'new'
     return func, (rest,)
 
@@ -144,7 +145,7 @@ def signal_child_parse(func: str, rest: str) -> FuncArgsType:
 def parse_change_font_size(func: str, rest: str) -> Tuple[str, Tuple[bool, Optional[str], float]]:
     vals = rest.strip().split(maxsplit=1)
     if len(vals) != 2:
-        log_error('Invalid change_font_size specification: {}, treating it as default'.format(rest))
+        log_error(f'Invalid change_font_size specification: {rest}, treating it as default')
         return func, (True, None, 0)
     c_all = vals[0].lower() == 'all'
     sign: Optional[str] = None
@@ -176,12 +177,22 @@ def paste_from_buffer(func: str, rest: str) -> FuncArgsType:
     return func, [rest]
 
 
+@func_with_args('paste')
+def paste_parse(func: str, rest: str) -> FuncArgsType:
+    text = ''
+    try:
+        text = python_string(rest)
+    except Exception:
+        log_error('Ignoring invalid paste string: ' + rest)
+    return func, [text]
+
+
 @func_with_args('neighboring_window')
 def neighboring_window(func: str, rest: str) -> FuncArgsType:
     rest = rest.lower()
     rest = {'up': 'top', 'down': 'bottom'}.get(rest, rest)
     if rest not in ('left', 'right', 'top', 'bottom'):
-        log_error('Invalid neighbor specification: {}'.format(rest))
+        log_error(f'Invalid neighbor specification: {rest}')
         rest = 'right'
     return func, [rest]
 
@@ -195,14 +206,14 @@ def resize_window(func: str, rest: str) -> FuncArgsType:
     else:
         quality = vals[0].lower()
         if quality not in ('taller', 'shorter', 'wider', 'narrower'):
-            log_error('Invalid quality specification: {}'.format(quality))
+            log_error(f'Invalid quality specification: {quality}')
             quality = 'wider'
         increment = 1
         if len(vals) == 2:
             try:
                 increment = int(vals[1])
             except Exception:
-                log_error('Invalid increment specification: {}'.format(vals[1]))
+                log_error(f'Invalid increment specification: {vals[1]}')
         args = [quality, increment]
     return func, args
 
@@ -216,7 +227,7 @@ def move_window(func: str, rest: str) -> FuncArgsType:
         prest = int(prest)
     except Exception:
         if prest not in ('left', 'right', 'top', 'bottom'):
-            log_error('Invalid move_window specification: {}'.format(rest))
+            log_error(f'Invalid move_window specification: {rest}')
             prest = 0
     return func, [prest]
 
@@ -242,20 +253,20 @@ def set_colors(func: str, rest: str) -> FuncArgsType:
 
 @func_with_args('remote_control')
 def remote_control(func: str, rest: str) -> FuncArgsType:
-    import shlex
-    r = shlex.split(rest)
-    if len(r) < 1:
+    func, args = shlex_parse(func, rest)
+    if len(args) < 1:
         log_error('Too few arguments to remote_control function')
-    return func, r
+    return func, args
 
 
-@func_with_args('nth_window')
-def nth_window(func: str, rest: str) -> FuncArgsType:
+@func_with_args('nth_os_window', 'nth_window', 'scroll_to_prompt', 'visual_window_select_action_trigger')
+def single_integer_arg(func: str, rest: str) -> FuncArgsType:
     try:
         num = int(rest)
     except Exception:
-        log_error('Invalid nth_window number: {}'.format(rest))
-        num = 1
+        if rest:
+            log_error(f'Invalid number for {func}: {rest}')
+        num = -1 if func == 'scroll_to_prompt' else 1
     return func, [num]
 
 
@@ -267,9 +278,9 @@ def disable_ligatures_in(func: str, rest: str) -> FuncArgsType:
     else:
         where, strategy = parts
     if where not in ('active', 'all', 'tab'):
-        raise ValueError('{} is not a valid set of windows to disable ligatures in'.format(where))
+        raise ValueError(f'{where} is not a valid set of windows to disable ligatures in')
     if strategy not in ('never', 'always', 'cursor'):
-        raise ValueError('{} is not a valid disable ligatures strategy'.format(strategy))
+        raise ValueError(f'{strategy} is not a valid disable ligatures strategy')
     return func, [where, strategy]
 
 
@@ -293,7 +304,7 @@ def parse_marker_spec(ftype: str, parts: Sequence[str]) -> Tuple[str, Union[str,
             try:
                 color = max(1, min(int(parts[i]), 3))
             except Exception:
-                raise ValueError('color {} in marker specification is not an integer'.format(parts[i]))
+                raise ValueError(f'color {parts[i]} in marker specification is not an integer')
             sspec = parts[i + 1]
             if 'regex' not in ftype:
                 sspec = re.escape(sspec)
@@ -303,7 +314,7 @@ def parse_marker_spec(ftype: str, parts: Sequence[str]) -> Tuple[str, Union[str,
     elif ftype == 'function':
         spec = ' '.join(parts)
     else:
-        raise ValueError('Unknown marker type: {}'.format(ftype))
+        raise ValueError(f'Unknown marker type: {ftype}')
     return ftype, spec, flags
 
 
@@ -312,7 +323,7 @@ def toggle_marker(func: str, rest: str) -> FuncArgsType:
     import shlex
     parts = rest.split(maxsplit=1)
     if len(parts) != 2:
-        raise ValueError('{} is not a valid marker specification'.format(rest))
+        raise ValueError(f'{rest} is not a valid marker specification')
     ftype, spec = parts
     parts = shlex.split(spec)
     return func, list(parse_marker_spec(ftype, parts))
@@ -330,7 +341,7 @@ def scroll_to_mark(func: str, rest: str) -> FuncArgsType:
         try:
             return func, [True, max(0, min(int(q), 3))]
         except Exception:
-            raise ValueError('{} is not a valid scroll_to_mark destination'.format(rest))
+            raise ValueError(f'{rest} is not a valid scroll_to_mark destination')
     return func, [parts[0] != 'next', max(0, min(int(parts[1]), 3))]
 
 
@@ -369,7 +380,7 @@ def parse_mods(parts: Iterable[str], sc: str) -> Optional[int]:
             mods |= getattr(defines, 'GLFW_MOD_' + map_mod(m.upper()))
         except AttributeError:
             if m.upper() != 'NONE':
-                log_error('Shortcut: {} has unknown modifier, ignoring'.format(sc))
+                log_error(f'Shortcut: {sc} has unknown modifier, ignoring')
             return None
 
     return mods
@@ -491,9 +502,9 @@ def url_style(x: str) -> int:
     return url_style_map.get(x, url_style_map['curly'])
 
 
-url_style_map = dict(
-    ((v, i) for i, v in enumerate('none single double curly'.split()))
-)
+url_style_map = {
+    v: i for i, v in enumerate('none single double curly'.split())
+}
 
 
 def url_prefixes(x: str) -> Tuple[str, ...]:
@@ -516,21 +527,25 @@ def window_size(val: str) -> Tuple[int, str]:
     return positive_int(val.rstrip('c')), unit
 
 
-def to_layout_names(raw: str) -> List[str]:
+def parse_layout_names(parts: Iterable[str]) -> List[str]:
     from kitty.layout.interface import all_layouts
-    parts = [x.strip().lower() for x in raw.split(',')]
-    ans: List[str] = []
+    ans = []
     for p in parts:
+        p = p.lower()
         if p in ('*', 'all'):
             ans.extend(sorted(all_layouts))
             continue
         name = p.partition(':')[0]
         if name not in all_layouts:
-            raise ValueError('The window layout {} is unknown'.format(p))
+            raise ValueError(f'The window layout {p} is unknown')
         ans.append(p)
     return uniq(ans)
 
 
+def to_layout_names(raw: str) -> List[str]:
+    return parse_layout_names(x.strip() for x in raw.split(','))
+
+
 def window_border_width(x: Union[str, int, float]) -> Tuple[float, str]:
     unit = 'pt'
     if isinstance(x, str):
@@ -579,6 +594,18 @@ def resize_draw_strategy(x: str) -> int:
     return cmap.get(x.lower(), 0)
 
 
+def visual_window_select_characters(x: str) -> str:
+    import string
+    valid_characters = string.digits + string.ascii_uppercase
+    ans = x.upper()
+    ans_chars = set(ans)
+    if not ans_chars.issubset(set(valid_characters)):
+        raise ValueError(f'Invalid characters in visual_window_select_characters: {x} Only numbers (0-9) and alphabets (a-z,A-Z) are allowed. Ignoring.')
+    if len(ans_chars) < len(x):
+        raise ValueError(f'Invalid characters in visual_window_select_characters: {x} Contains identical numbers or alphabets, case insensitive. Ignoring.')
+    return ans
+
+
 def tab_separator(x: str) -> str:
     for q in '\'"':
         if x.startswith(q) and x.endswith(q):
@@ -631,14 +658,10 @@ def active_tab_title_template(x: str) -> Optional[str]:
     return None if x == 'none' else x
 
 
-def config_or_absolute_path(x: str) -> Optional[str]:
-    if x.lower() == 'none':
+def config_or_absolute_path(x: str, env: Optional[Dict[str, str]] = None) -> Optional[str]:
+    if not x or x.lower() == 'none':
         return None
-    x = os.path.expanduser(x)
-    x = os.path.expandvars(x)
-    if not os.path.isabs(x):
-        x = os.path.join(config_dir, x)
-    return x
+    return resolve_abs_or_config_path(x, env)
 
 
 def allow_remote_control(x: str) -> str:
@@ -657,13 +680,26 @@ def allow_hyperlinks(x: str) -> int:
     return 1 if to_bool(x) else 0
 
 
-def macos_titlebar_color(x: str) -> int:
+def titlebar_color(x: str) -> int:
     x = x.strip('"')
     if x == 'system':
         return 0
     if x == 'background':
         return 1
-    return (color_as_int(to_color(x)) << 8) | 2
+    try:
+        return (color_as_int(to_color(x)) << 8) | 2
+    except ValueError:
+        log_error(f'Ignoring invalid title bar color: {x}')
+    return 0
+
+
+def macos_titlebar_color(x: str) -> int:
+    x = x.strip('"')
+    if x == 'light':
+        return -1
+    if x == 'dark':
+        return -2
+    return titlebar_color(x)
 
 
 def macos_option_as_alt(x: str) -> int:
@@ -690,12 +726,19 @@ class TabBarMarginHeight(NamedTuple):
 def tab_bar_margin_height(x: str) -> TabBarMarginHeight:
     parts = x.split(maxsplit=1)
     if len(parts) != 2:
-        log_error(f'Invalid tab_bar_margin_height: {tab_bar_margin_height}, ignoring')
+        log_error(f'Invalid tab_bar_margin_height: {x}, ignoring')
         return TabBarMarginHeight()
     ans = map(positive_float, parts)
     return TabBarMarginHeight(next(ans), next(ans))
 
 
+def clear_all_mouse_actions(val: str, dict_with_parse_results: Optional[Dict[str, Any]] = None) -> bool:
+    ans = to_bool(val)
+    if ans and dict_with_parse_results is not None:
+        dict_with_parse_results['mouse_map'] = [None]
+    return ans
+
+
 def clear_all_shortcuts(val: str, dict_with_parse_results: Optional[Dict[str, Any]] = None) -> bool:
     ans = to_bool(val)
     if ans and dict_with_parse_results is not None:
@@ -708,7 +751,7 @@ def font_features(val: str) -> Iterable[Tuple[str, Tuple[FontFeature, ...]]]:
         return
     parts = val.split()
     if len(parts) < 2:
-        log_error("Ignoring invalid font_features {}".format(val))
+        log_error(f"Ignoring invalid font_features {val}")
         return
     if parts[0]:
         features = []
@@ -716,30 +759,55 @@ def font_features(val: str) -> Iterable[Tuple[str, Tuple[FontFeature, ...]]]:
             try:
                 parsed = defines.parse_font_feature(feat)
             except ValueError:
-                log_error('Ignoring invalid font feature: {}'.format(feat))
+                log_error(f'Ignoring invalid font feature: {feat}')
             else:
                 features.append(FontFeature(feat, parsed))
         yield parts[0], tuple(features)
 
 
 def env(val: str, current_val: Dict[str, str]) -> Iterable[Tuple[str, str]]:
-    key, val = val.partition('=')[::2]
-    key, val = key.strip(), val.strip()
-    if key:
-        yield key, expandvars(val, current_val)
+    val = val.strip()
+    if val:
+        if '=' in val:
+            key, v = val.split('=', 1)
+            key, v = key.strip(), v.strip()
+            if key:
+                if v:
+                    v = expandvars(v, current_val)
+                yield key, v
+        else:
+            yield val, DELETE_ENV_VAR
 
 
-def kitten_alias(val: str) -> Iterable[Tuple[str, List[str]]]:
-    parts = val.split(maxsplit=2)
-    if len(parts) >= 2:
-        name = parts.pop(0)
-        yield name, parts
+def store_multiple(val: str, current_val: Container[str]) -> Iterable[Tuple[str, str]]:
+    val = val.strip()
+    if val not in current_val:
+        yield val, val
+
+
+def shell_integration(x: str) -> FrozenSet[str]:
+    s = frozenset({'enabled', 'disabled', 'no-rc', 'no-cursor', 'no-title', 'no-prompt-mark', 'no-complete'})
+    q = frozenset(x.lower().split())
+    if not q.issubset(s):
+        log_error(f'Invalid shell integration options: {q - s}, ignoring')
+        return q & s
+    return q
+
+
+def action_alias(val: str) -> Iterable[Tuple[str, str]]:
+    parts = val.split(maxsplit=1)
+    if len(parts) > 1:
+        alias_name, rest = parts
+        yield alias_name, rest
+
+
+kitten_alias = action_alias
 
 
 def symbol_map(val: str) -> Iterable[Tuple[Tuple[int, int], str]]:
     parts = val.split()
 
-    def abort() -> Dict[Tuple[int, int], str]:
+    def abort() -> None:
         log_error(f'Symbol map: {val} is invalid, ignoring')
 
     if len(parts) < 2:
@@ -752,7 +820,7 @@ def symbol_map(val: str) -> Iterable[Tuple[Tuple[int, int], str]]:
         return int(x[2:], 16)
 
     for x in parts[0].split(','):
-        a_, b_ = x.partition('-')[::2]
+        a_, b_ = x.replace('–', '-').partition('-')[::2]
         b_ = b_ or a_
         try:
             a, b = map(to_chr, (a_, b_))
@@ -763,57 +831,138 @@ def symbol_map(val: str) -> Iterable[Tuple[Tuple[int, int], str]]:
         yield (a, b), family
 
 
-def parse_key_action(action: str, action_type: str = 'map') -> Optional[KeyAction]:
+def parse_key_action(action: str, action_type: str = 'map') -> KeyAction:
     parts = action.strip().split(maxsplit=1)
     func = parts[0]
     if len(parts) == 1:
         return KeyAction(func, ())
     rest = parts[1]
-    parser = args_funcs.get(func)
-    if parser is not None:
-        try:
-            func, args = parser(func, rest)
-        except Exception as err:
-            log_error(f'Ignoring invalid {action_type} action: {action} with err: {err}')
-        else:
-            return KeyAction(func, tuple(args))
+    parser = func_with_args.get(func)
+    if parser is None:
+        raise KeyError(f'Unknown action: {func}')
+    func, args = parser(func, rest)
+    return KeyAction(func, tuple(args))
+
+
+class ActionAlias(NamedTuple):
+    name: str
+    value: str
+    replace_second_arg: bool = False
+
+
+class AliasMap:
+
+    def __init__(self) -> None:
+        self.aliases: Dict[str, List[ActionAlias]] = {}
+
+    def append(self, name: str, aa: ActionAlias) -> None:
+        self.aliases.setdefault(name, []).append(aa)
+
+    def update(self, aa: 'AliasMap') -> None:
+        self.aliases.update(aa.aliases)
+
+    @lru_cache(maxsize=256)
+    def resolve_aliases(self, definition: str, map_type: str = 'map') -> Tuple[KeyAction, ...]:
+        return tuple(resolve_aliases_and_parse_actions(definition, self.aliases, map_type))
+
+
+def build_action_aliases(raw: Dict[str, str], first_arg_replacement: str = '') -> AliasMap:
+    ans = AliasMap()
+    if first_arg_replacement:
+        for alias_name, rest in raw.items():
+            ans.append(first_arg_replacement, ActionAlias(alias_name, rest, True))
     else:
-        log_error(f'Ignoring unknown {action_type} action: {action}')
-    return None
+        for alias_name, rest in raw.items():
+            ans.append(alias_name, ActionAlias(alias_name, rest))
+    return ans
+
+
+def resolve_aliases_and_parse_actions(
+    defn: str, aliases: Dict[str, List[ActionAlias]], map_type: str
+) -> Iterator[KeyAction]:
+    parts = defn.split(maxsplit=1)
+    if len(parts) == 1:
+        possible_alias = defn
+        rest = ''
+    else:
+        possible_alias = parts[0]
+        rest = parts[1]
+    for alias in aliases.get(possible_alias, ()):
+        if alias.replace_second_arg:  # kitten_alias
+            if not rest:
+                continue
+            parts = rest.split(maxsplit=1)
+            if parts[0] != alias.name:
+                continue
+            new_defn = possible_alias + ' ' + alias.value + ((' ' + parts[1]) if len(parts) > 1 else '')
+            new_aliases = aliases.copy()
+            new_aliases[possible_alias] = [a for a in aliases[possible_alias] if a is not alias]
+            yield from resolve_aliases_and_parse_actions(new_defn, new_aliases, map_type)
+            return
+        else:  # action_alias
+            new_defn = alias.value + ((' ' + rest) if rest else '')
+            new_aliases = aliases.copy()
+            new_aliases.pop(possible_alias)
+            yield from resolve_aliases_and_parse_actions(new_defn, new_aliases, map_type)
+            return
+
+    if possible_alias == 'combine':
+        sep, rest = rest.split(maxsplit=1)
+        parts = re.split(r'\s*' + re.escape(sep) + r'\s*', rest)
+        for x in parts:
+            if x:
+                yield from resolve_aliases_and_parse_actions(x, aliases, map_type)
+    else:
+        yield parse_key_action(defn, map_type)
 
 
 class BaseDefinition:
-    action: KeyAction
+    no_op_actions = frozenset(('noop', 'no-op', 'no_op'))
+    map_type: str = 'map'
+    definition_location: CurrentlyParsing
 
-    def resolve_kitten_aliases(self, aliases: Dict[str, List[str]]) -> KeyAction:
-        if not self.action.args or not aliases:
-            return self.action
-        kitten = self.action.args[0]
-        rest = str(self.action.args[1] if len(self.action.args) > 1 else '')
-        changed = False
-        for key, expanded in aliases.items():
-            if key == kitten:
-                changed = True
-                kitten = expanded[0]
-                if len(expanded) > 1:
-                    rest = expanded[1] + ' ' + rest
-        return self.action._replace(args=(kitten, rest.rstrip())) if changed else self.action
+    def __init__(self, definition: str = '') -> None:
+        self.definition = definition
+        self.definition_location = currently_parsing.__copy__()
+
+    @property
+    def is_no_op(self) -> bool:
+        return self.definition in self.no_op_actions
+
+    def pretty_repr(self, *fields: str) -> str:
+        kwds = []
+        defaults = self.__class__()
+        for f in fields:
+            val = getattr(self, f)
+            if val != getattr(defaults, f):
+                kwds.append(f'{f}={val!r}')
+        if self.definition:
+            kwds.append(f'definition={self.definition!r}')
+        return f'{self.__class__.__name__}({", ".join(kwds)})'
 
 
 class MouseMapping(BaseDefinition):
+    map_type: str = 'mouse_map'
 
-    def __init__(self, button: int, mods: int, repeat_count: int, grabbed: bool, action: KeyAction):
+    def __init__(
+        self, button: int = 0, mods: int = 0, repeat_count: int = 1, grabbed: bool = False,
+        definition: str = ''
+    ):
+        super().__init__(definition)
         self.button = button
         self.mods = mods
         self.repeat_count = repeat_count
         self.grabbed = grabbed
-        self.action = action
 
     def __repr__(self) -> str:
-        return f'MouseMapping({self.button}, {self.mods}, {self.repeat_count}, {self.grabbed}, {self.action})'
+        return self.pretty_repr('button', 'mods', 'repeat_count', 'grabbed')
 
-    def resolve_and_copy(self, kitty_mod: int, aliases: Dict[str, List[str]]) -> 'MouseMapping':
-        return MouseMapping(self.button, defines.resolve_key_mods(kitty_mod, self.mods), self.repeat_count, self.grabbed, self.resolve_kitten_aliases(aliases))
+    def resolve_and_copy(self, kitty_mod: int) -> 'MouseMapping':
+        ans = MouseMapping(
+            self.button, defines.resolve_key_mods(kitty_mod, self.mods), self.repeat_count, self.grabbed,
+            self.definition)
+        ans.definition_location = self.definition_location
+        return ans
 
     @property
     def trigger(self) -> MouseEvent:
@@ -822,23 +971,28 @@ class MouseMapping(BaseDefinition):
 
 class KeyDefinition(BaseDefinition):
 
-    def __init__(self, is_sequence: bool, action: KeyAction, mods: int, is_native: bool, key: int, rest: Tuple[SingleKey, ...] = ()):
+    def __init__(
+        self, is_sequence: bool = False, trigger: SingleKey = SingleKey(),
+            rest: Tuple[SingleKey, ...] = (), definition: str = ''
+    ):
+        super().__init__(definition)
         self.is_sequence = is_sequence
-        self.action = action
-        self.trigger = SingleKey(mods, is_native, key)
+        self.trigger = trigger
         self.rest = rest
 
     def __repr__(self) -> str:
-        return f'KeyDefinition({self.is_sequence}, {self.action}, {self.trigger.mods}, {self.trigger.is_native}, {self.trigger.key}, {self.rest})'
+        return self.pretty_repr('is_sequence', 'trigger', 'rest')
 
-    def resolve_and_copy(self, kitty_mod: int, aliases: Dict[str, List[str]]) -> 'KeyDefinition':
+    def resolve_and_copy(self, kitty_mod: int) -> 'KeyDefinition':
         def r(k: SingleKey) -> SingleKey:
             mods = defines.resolve_key_mods(kitty_mod, k.mods)
             return k._replace(mods=mods)
-        return KeyDefinition(
-            self.is_sequence, self.resolve_kitten_aliases(aliases),
-            defines.resolve_key_mods(kitty_mod, self.trigger.mods),
-            self.trigger.is_native, self.trigger.key, tuple(map(r, self.rest)))
+        ans = KeyDefinition(
+            self.is_sequence, r(self.trigger), tuple(map(r, self.rest)),
+            self.definition
+        )
+        ans.definition_location = self.definition_location
+        return ans
 
 
 def parse_map(val: str) -> Iterable[KeyDefinition]:
@@ -860,7 +1014,7 @@ def parse_map(val: str) -> Iterable[KeyDefinition]:
                 return
             if key == 0:
                 if mods is not None:
-                    log_error('Shortcut: {} has unknown key, ignoring'.format(sc))
+                    log_error(f'Shortcut: {sc} has unknown key, ignoring')
                 return
             if trigger is None:
                 trigger = SingleKey(mods, is_native, key)
@@ -874,21 +1028,14 @@ def parse_map(val: str) -> Iterable[KeyDefinition]:
             return
         if key == 0:
             if mods is not None:
-                log_error('Shortcut: {} has unknown key, ignoring'.format(sc))
+                log_error(f'Shortcut: {sc} has unknown key, ignoring')
             return
-    try:
-        paction = parse_key_action(action)
-    except Exception:
-        log_error('Invalid shortcut action: {}. Ignoring.'.format(
-            action))
+    if is_sequence:
+        if trigger is not None:
+            yield KeyDefinition(True, trigger, rest, definition=action)
     else:
-        if paction is not None:
-            if is_sequence:
-                if trigger is not None:
-                    yield KeyDefinition(True, paction, trigger[0], trigger[1], trigger[2], rest)
-            else:
-                assert key is not None
-                yield KeyDefinition(False, paction, mods, is_native, key)
+        assert key is not None
+        yield KeyDefinition(False, SingleKey(mods, is_native, key), definition=action)
 
 
 def parse_mouse_map(val: str) -> Iterable[MouseMapping]:
@@ -921,21 +1068,14 @@ def parse_mouse_map(val: str) -> Iterable[MouseMapping]:
     if specified_modes - {'grabbed', 'ungrabbed'}:
         log_error(f'Mouse modes: {modes} not recognized, ignoring')
         return
-    try:
-        paction = parse_key_action(action, 'mouse_map')
-    except Exception:
-        log_error(f'Invalid mouse action: {action}. Ignoring.')
-        return
-    if paction is None:
-        return
     for mode in sorted(specified_modes):
-        yield MouseMapping(button, mods, count, mode == 'grabbed', paction)
+        yield MouseMapping(button, mods, count, mode == 'grabbed', definition=action)
 
 
 def deprecated_hide_window_decorations_aliases(key: str, val: str, ans: Dict[str, Any]) -> None:
     if not hasattr(deprecated_hide_window_decorations_aliases, key):
         setattr(deprecated_hide_window_decorations_aliases, key, True)
-        log_error('The option {} is deprecated. Use hide_window_decorations instead.'.format(key))
+        log_error(f'The option {key} is deprecated. Use hide_window_decorations instead.')
     if to_bool(val):
         if is_macos and key == 'macos_hide_titlebar' or (not is_macos and key == 'x11_hide_window_decorations'):
             ans['hide_window_decorations'] = True
@@ -944,7 +1084,7 @@ def deprecated_hide_window_decorations_aliases(key: str, val: str, ans: Dict[str
 def deprecated_macos_show_window_title_in_menubar_alias(key: str, val: str, ans: Dict[str, Any]) -> None:
     if not hasattr(deprecated_macos_show_window_title_in_menubar_alias, key):
         setattr(deprecated_macos_show_window_title_in_menubar_alias, 'key', True)
-        log_error('The option {} is deprecated. Use macos_show_window_title_in menubar instead.'.format(key))
+        log_error(f'The option {key} is deprecated. Use macos_show_window_title_in menubar instead.')
     macos_show_window_title_in = ans.get('macos_show_window_title_in', 'all')
     if to_bool(val):
         if macos_show_window_title_in == 'none':
@@ -970,6 +1110,6 @@ def deprecated_send_text(key: str, val: str, ans: Dict[str, Any]) -> None:
         return abort('Incomplete')
     mode, sc = parts[:2]
     text = ' '.join(parts[2:])
-    key_str = '{} send_text {} {}'.format(sc, mode, text)
+    key_str = f'{sc} send_text {mode} {text}'
     for k in parse_map(key_str):
         ans['map'].append(k)
