@@ -17,6 +17,7 @@
 
 extern PyTypeObject Screen_Type;
 #define EXTENDED_OSC_SENTINEL 0x1bu
+#define PENDING_BUF_INCREMENT (16u * 1024u)
 
 // utils {{{
 static const uint64_t pow_10_array[] = {
@@ -419,6 +420,7 @@ dispatch_osc(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
             break;
         case 9:
         case 99:
+        case 777:
             START_DISPATCH
             DISPATCH_OSC_WITH_CODE(desktop_notify)
             END_DISPATCH
@@ -441,6 +443,14 @@ dispatch_osc(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
             DISPATCH_OSC_WITH_CODE(clipboard_control);
             if (code == -52) continue_osc_52(screen);
             END_DISPATCH
+        case 133:
+            START_DISPATCH
+            DISPATCH_OSC(shell_prompt_marking);
+            END_DISPATCH
+        case FILE_TRANSFER_CODE:
+            START_DISPATCH
+            DISPATCH_OSC(file_transmission);
+            END_DISPATCH
         case 30001:
             REPORT_COMMAND(screen_push_dynamic_colors);
             screen_push_colors(screen, 0);
@@ -877,9 +887,11 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
                 REPORT_COMMAND(screen_save_cursor);
                 screen_save_cursor(screen);
                 break;
-            } else if (start_modifier == '?' && !end_modifier && !num_params) {
-                REPORT_COMMAND(screen_save_modes);
-                screen_save_modes(screen);
+            } else if (start_modifier == '?' && !end_modifier) {
+                if (!num_params) {
+                    REPORT_COMMAND(screen_save_modes);
+                    screen_save_modes(screen);
+                } else { SET_MODE(screen_save_mode); }
                 break;
             }
             REPORT_ERROR("Unknown CSI s sequence with start and end modifiers: '%c' '%c' and %u parameters", start_modifier, end_modifier, num_params);
@@ -905,6 +917,7 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
                     break;
                 case 22:
                 case 23:
+                    if (num_params == 3 && !params[2]) num_params = 2; // ignore extra 0, generated by weechat or ncurses
                     CALL_CSI_HANDLER2(screen_manipulate_title_stack, 22, 0);
                     break;
                 default:
@@ -941,9 +954,11 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
             if (!start_modifier && !end_modifier) {
                 // DECSTBM
                 CALL_CSI_HANDLER2(screen_set_margins, 0, 0);
-            } else if (start_modifier == '?' && !end_modifier && !num_params) {
-                REPORT_COMMAND(screen_restore_modes);
-                screen_restore_modes(screen);
+            } else if (start_modifier == '?' && !end_modifier) {
+                if (!num_params) {
+                    REPORT_COMMAND(screen_restore_modes);
+                    screen_restore_modes(screen);
+                } else { SET_MODE(screen_restore_mode); }
                 break;
             }
             REPORT_ERROR("Unknown CSI r sequence with start and end modifiers: '%c' '%c' and %u parameters", start_modifier, end_modifier, num_params);
@@ -984,7 +999,11 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
             break;
         case 'm':
             if (start_modifier == '>' && !end_modifier) {
-                CALL_CSI_HANDLER2(screen_xtmodkeys, 0, 0);
+                REPORT_ERROR(
+                    "The application is trying to use XTerm's modifyOtherKeys."
+                    " This is superseded by the kitty keyboard protocol: https://sw.kovidgoyal.net/kitty/keyboard-protocol/"
+                    " the application should be updated to use that"
+                );
                 break;
             }
             /* fallthrough */
@@ -1347,6 +1366,13 @@ FLUSH_DRAW;
 
 static void
 write_pending_char(Screen *screen, uint32_t ch) {
+    if (screen->pending_mode.capacity < screen->pending_mode.used + 8) {
+        if (screen->pending_mode.capacity) {
+            screen->pending_mode.capacity += screen->pending_mode.capacity >= READ_BUF_SZ ? PENDING_BUF_INCREMENT : screen->pending_mode.capacity;
+        } else screen->pending_mode.capacity = PENDING_BUF_INCREMENT;
+        screen->pending_mode.buf = realloc(screen->pending_mode.buf, screen->pending_mode.capacity);
+        if (!screen->pending_mode.buf) fatal("Out of memory");
+    }
     screen->pending_mode.used += encode_utf8(ch, (char*)screen->pending_mode.buf + screen->pending_mode.used);
 }
 
@@ -1454,19 +1480,9 @@ end:
     return i;
 }
 
-static void
-create_pending_space(Screen *screen, size_t needed_space) {
-    screen->pending_mode.capacity = MAX(screen->pending_mode.capacity * 2, screen->pending_mode.used + needed_space);
-    if (screen->pending_mode.capacity > READ_BUF_SZ) screen->pending_mode.capacity = screen->pending_mode.used + needed_space;
-    screen->pending_mode.buf = realloc(screen->pending_mode.buf, screen->pending_mode.capacity);
-    if (!screen->pending_mode.buf) fatal("Out of memory");
-}
-
 static void
 dump_partial_escape_code_to_pending(Screen *screen) {
     if (screen->parser_buf_pos) {
-        const size_t needed_space = 4 * screen->parser_buf_pos + 8;
-        if (screen->pending_mode.used + needed_space >= screen->pending_mode.capacity) create_pending_space(screen, needed_space);
         write_pending_char(screen, screen->parser_state);
         for (unsigned i = 0; i < screen->parser_buf_pos; i++) write_pending_char(screen, screen->parser_buf[i]);
     }
@@ -1499,6 +1515,11 @@ do_parse_bytes(Screen *screen, const uint8_t *read_buf, const size_t read_buf_sz
                 _parse_bytes(screen, screen->pending_mode.buf, screen->pending_mode.used, dump_callback);
                 screen->pending_mode.used = 0;
                 screen->pending_mode.activated_at = 0;  // ignore any pending starts in the pending bytes
+                if (screen->pending_mode.capacity > READ_BUF_SZ + PENDING_BUF_INCREMENT) {
+                    screen->pending_mode.capacity = READ_BUF_SZ;
+                    screen->pending_mode.buf = realloc(screen->pending_mode.buf, screen->pending_mode.capacity);
+                    if (!screen->pending_mode.buf) fatal("Out of memory");
+                }
                 if (screen->pending_mode.stop_escape_code_type) {
                     if (screen->pending_mode.stop_escape_code_type == DCS) { REPORT_COMMAND(screen_stop_pending_mode); }
                     else if (screen->pending_mode.stop_escape_code_type == CSI) { REPORT_COMMAND(screen_reset_mode, 2026, 1); }
@@ -1514,16 +1535,12 @@ do_parse_bytes(Screen *screen, const uint8_t *read_buf, const size_t read_buf_sz
                 break;
 
             case QUEUE_PENDING: {
-                const size_t needed_space = read_buf_sz * 2;
                 screen->pending_mode.stop_escape_code_type = 0;
-                if (screen->pending_mode.capacity - screen->pending_mode.used < needed_space) {
-                    if (screen->pending_mode.capacity >= READ_BUF_SZ) {
-                        dump_partial_escape_code_to_pending(screen);
-                        screen->pending_mode.activated_at = 0;
-                        state = START;
-                        break;
-                    }
-                    create_pending_space(screen, needed_space);
+                if (screen->pending_mode.used >= READ_BUF_SZ) {
+                    dump_partial_escape_code_to_pending(screen);
+                    screen->pending_mode.activated_at = 0;
+                    state = START;
+                    break;
                 }
                 if (!screen->pending_mode.used) parser_state_at_start_of_pending = screen->parser_state;
                 read_buf_pos += queue_pending_bytes(screen, read_buf + read_buf_pos, read_buf_sz - read_buf_pos, dump_callback);
