@@ -1,26 +1,51 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 import base64
-import os
 import sys
-from typing import TYPE_CHECKING, Dict, Generator, List, Optional
+from functools import partial
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Union
 
-from kitty.options.utils import parse_send_text_bytes
+from kitty.fast_data_types import (
+    KeyEvent as WindowSystemKeyEvent, add_timer, get_boss, remove_timer
+)
 from kitty.key_encoding import decode_key_event_as_window_system_key
-from kitty.fast_data_types import KeyEvent as WindowSystemKeyEvent
+from kitty.options.utils import parse_send_text_bytes
 
 from .base import (
-    MATCH_TAB_OPTION, MATCH_WINDOW_OPTION, ArgsType, Boss, MatchError,
-    PayloadGetType, PayloadType, RCOptions, RemoteCommand, ResponseType,
-    Window
+    MATCH_TAB_OPTION, MATCH_WINDOW_OPTION, ArgsType, Boss, CmdGenerator,
+    MatchError, PayloadGetType, PayloadType, RCOptions, RemoteCommand,
+    ResponseType, Window
 )
 
 if TYPE_CHECKING:
     from kitty.cli_stub import SendTextRCOptions as CLIOptions
 
 
+class Session:
+    id: str
+    window_ids: Set[int]
+    timer_id: int = 0
+
+    def __init__(self, id: str):
+        self.id = id
+        self.window_ids = set()
+
+
+sessions_map: Dict[str, Session] = {}
+TIMEOUT_FOR_SESSION = 3.0
+
+
+def clear_unfocused_cursors(sid: str, *a: Any) -> None:
+    s = sessions_map.pop(sid, None)
+    if s is not None:
+        boss = get_boss()
+        for wid in s.window_ids:
+            qw = boss.window_id_map.get(wid)
+            if qw is not None:
+                qw.screen.render_unfocused_cursor = 0
+
+
 class SendText(RemoteCommand):
     '''
     data+: The data being sent. Can be either: text: followed by text or base64: followed by standard base64 encoded bytes
@@ -28,6 +53,7 @@ class SendText(RemoteCommand):
     match_tab: A string indicating the tab to send text to
     all: A boolean indicating all windows should be matched.
     exclude_active: A boolean that prevents sending text to the active window
+    session_id: A string that identifies a "broadcast session"
     '''
     short_desc = 'Send arbitrary text to specified windows'
     desc = (
@@ -65,25 +91,24 @@ Do not send text to the active window, even if it is one of the matched windows.
         limit = 1024
         ret = {'match': opts.match, 'data': '', 'match_tab': opts.match_tab, 'all': opts.all, 'exclude_active': opts.exclude_active}
 
-        def pipe() -> Generator[Dict, None, None]:
+        def pipe() -> CmdGenerator:
             if sys.stdin.isatty():
                 ret['exclude_active'] = True
-                import select
-                fd = sys.stdin.fileno()
                 keep_going = True
-                while keep_going:
-                    rd = select.select([fd], [], [])[0]
-                    if not rd:
-                        break
-                    data = os.read(fd, limit)
-                    if not data:
-                        break  # eof
-                    decoded_data = data.decode('utf-8')
-                    if '\x04' in decoded_data:
-                        decoded_data = decoded_data[:decoded_data.index('\x04')]
-                        keep_going = False
-                    ret['data'] = 'text:' + decoded_data
-                    yield ret
+                from kitty.utils import TTYIO
+                with TTYIO(read_with_timeout=False) as tty:
+                    while keep_going:
+                        if not tty.wait_till_read_available():
+                            break
+                        data = tty.read(limit)
+                        if not data:
+                            break
+                        decoded_data = data.decode('utf-8')
+                        if '\x04' in decoded_data:
+                            decoded_data = decoded_data[:decoded_data.index('\x04')]
+                            keep_going = False
+                        ret['data'] = 'text:' + decoded_data
+                        yield ret
             else:
                 while True:
                     data = sys.stdin.buffer.read(limit)
@@ -92,14 +117,14 @@ Do not send text to the active window, even if it is one of the matched windows.
                     ret['data'] = 'base64:' + base64.standard_b64encode(data).decode('ascii')
                     yield ret
 
-        def chunks(text: str) -> Generator[Dict, None, None]:
+        def chunks(text: str) -> CmdGenerator:
             data = parse_send_text_bytes(text).decode('utf-8')
             while data:
                 ret['data'] = 'text:' + data[:limit]
                 yield ret
                 data = data[limit:]
 
-        def file_pipe(path: str) -> Generator[Dict, None, None]:
+        def file_pipe(path: str) -> CmdGenerator:
             with open(path, 'rb') as f:
                 while True:
                     data = f.read(limit)
@@ -118,7 +143,7 @@ Do not send text to the active window, even if it is one of the matched windows.
         text = ' '.join(args)
         sources.append(chunks(text))
 
-        def chain() -> Generator[Dict, None, None]:
+        def chain() -> CmdGenerator:
             for src in sources:
                 yield from src
         return chain()
@@ -139,26 +164,63 @@ Do not send text to the active window, even if it is one of the matched windows.
                     raise MatchError(payload_get('match_tab'), 'tabs')
                 for tab in tabs:
                     windows += tuple(tab)
-        encoding, _, q = payload_get('data').partition(':')
+        pdata: str = payload_get('data')
+        encoding, _, q = pdata.partition(':')
+        session = ''
         if encoding == 'text':
-            data = q.encode('utf-8')
+            data: Union[bytes, WindowSystemKeyEvent] = q.encode('utf-8')
         elif encoding == 'base64':
             data = base64.standard_b64decode(q)
         elif encoding == 'kitty-key':
-            data = base64.standard_b64decode(q)
-            data = decode_key_event_as_window_system_key(data)
+            bdata = base64.standard_b64decode(q)
+            candidate = decode_key_event_as_window_system_key(bdata.decode('ascii'))
+            if candidate is None:
+                raise ValueError(f'Could not decode window system key: {q}')
+            data = candidate
+        elif encoding == 'session':
+            session = q
         else:
             raise TypeError(f'Invalid encoding for send-text data: {encoding}')
         exclude_active = payload_get('exclude_active')
-        for window in windows:
+        actual_windows = (w for w in windows if w is not None and (not exclude_active or w is not boss.active_window))
+        sid = payload_get('session_id', '')
+
+        def create_or_update_session() -> Session:
+            s = sessions_map.setdefault(sid, Session(sid))
+            if s.timer_id:
+                remove_timer(s.timer_id)
+                s.timer_id = 0
+            return s
+
+        if session == 'end':
+            s = create_or_update_session()
+            for w in actual_windows:
+                w.screen.render_unfocused_cursor = 0
+                s.window_ids.discard(w.id)
+            clear_unfocused_cursors(sid)
+        elif session == 'start':
+            s = create_or_update_session()
             if window is not None:
-                if not exclude_active or window is not boss.active_window:
-                    if isinstance(data, WindowSystemKeyEvent):
-                        kdata = window.encoded_key(data)
-                        if kdata:
-                            window.write_to_child(kdata)
-                    else:
-                        window.write_to_child(data)
+                window.actions_on_close.append(partial(clear_unfocused_cursors, sid))
+            s.timer_id = add_timer(partial(clear_unfocused_cursors, sid), TIMEOUT_FOR_SESSION, False)
+            for w in actual_windows:
+                w.screen.render_unfocused_cursor = 1
+                s.window_ids.add(w.id)
+        else:
+            if sid:
+                s = create_or_update_session()
+                s.timer_id = add_timer(partial(clear_unfocused_cursors, sid), TIMEOUT_FOR_SESSION, False)
+            for w in actual_windows:
+                if sid:
+                    w.screen.render_unfocused_cursor = 1
+                    s.window_ids.add(w.id)
+                if isinstance(data, WindowSystemKeyEvent):
+                    kdata = w.encoded_key(data)
+                    if kdata:
+                        w.write_to_child(kdata)
+                else:
+                    w.write_to_child(data)
+        return None
 
 
 send_text = SendText()
