@@ -1,34 +1,10 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import re
 from contextlib import suppress
-from typing import Optional, NamedTuple
-
-
-class Color(NamedTuple):
-    red: int = 0
-    green: int = 0
-    blue: int = 0
-
-    def luminance(self) -> float:
-        return 0.299 * self.red + 0.587 * self.green + 0.114 * self.blue
-
-    def contrast(self, other: 'Color') -> float:
-        a = self.luminance()
-        b = other.luminance()
-        if a < b:
-            a, b = b, a
-        return (a + 0.05) / (b + 0.05)
-
-    def __int__(self) -> int:
-        return self.red << 16 | self.green << 8 | self.blue
-
-    def as_bytearray(self, alpha: Optional[int] = None) -> bytearray:
-        if alpha is None:
-            return bytearray((self.red, self.green, self.blue))
-        return bytearray((self.red, self.green, self.blue, alpha))
+from typing import Optional
+from .fast_data_types import Color
 
 
 def alpha_blend_channel(top_color: int, bottom_color: int, alpha: float) -> int:
@@ -69,15 +45,15 @@ def color_from_int(x: int) -> Color:
 
 
 def color_as_int(x: Color) -> int:
-    return x.red << 16 | x.green << 8 | x.blue
+    return int(x)
 
 
 def color_as_sharp(x: Color) -> str:
-    return '#{:02x}{:02x}{:02x}'.format(*x)
+    return x.as_sharp
 
 
 def color_as_sgr(x: Color) -> str:
-    return ':2:{}:{}:{}'.format(*x)
+    return x.as_sgr
 
 
 def to_color(raw: str, validate: bool = False) -> Optional[Color]:
@@ -93,7 +69,7 @@ def to_color(raw: str, validate: bool = False) -> Optional[Color]:
         elif raw[:4].lower() == 'rgb:':
             val = parse_rgb(raw[4:])
     if val is None and validate:
-        raise ValueError('Invalid color name: {}'.format(raw))
+        raise ValueError(f'Invalid color name: {raw}')
     return val
 
 
@@ -856,8 +832,8 @@ color_names = {
 
 if __name__ == '__main__':
     # Read RGB color table from specified rgb.txt file
-    import sys
     import pprint
+    import sys
     data = {}
     with open(sys.argv[-1]) as f:
         for line in f:
