@@ -137,6 +137,7 @@ send_bgimage_to_gpu(BackgroundImageLayout layout, BackgroundImage *bgimage) {
     RepeatStrategy r;
     switch (layout) {
         case SCALED:
+        case CLAMPED:
             r = REPEAT_CLAMP; break;
         case MIRRORED:
             r = REPEAT_MIRROR; break;
@@ -223,12 +224,40 @@ release_gpu_resources_for_window(Window *w) {
     w->render_data.gvao_idx = -1;
 }
 
+static bool
+set_window_logo(Window *w, const char *path, const ImageAnchorPosition pos, float alpha, bool is_default) {
+    bool ok = false;
+    if (path && path[0]) {
+        window_logo_id_t wl = find_or_create_window_logo(global_state.all_window_logos, path);
+        if (wl) {
+            if (w->window_logo.id) decref_window_logo(global_state.all_window_logos, w->window_logo.id);
+            w->window_logo.id = wl;
+            w->window_logo.position = pos;
+            w->window_logo.alpha = alpha;
+            ok = true;
+        }
+    } else {
+        if (w->window_logo.id) {
+            decref_window_logo(global_state.all_window_logos, w->window_logo.id);
+            w->window_logo.id = 0;
+        }
+        ok = true;
+    }
+    w->window_logo.using_default = is_default;
+    if (ok && w->render_data.screen) w->render_data.screen->is_dirty = true;
+    return ok;
+}
+
 static void
 initialize_window(Window *w, PyObject *title, bool init_gpu_resources) {
     w->id = ++global_state.window_id_counter;
     w->visible = true;
     w->title = title;
     Py_XINCREF(title);
+    if (!set_window_logo(w, OPT(default_window_logo), OPT(window_logo_position), OPT(window_logo_alpha), true)) {
+        log_error("Failed to load default window logo: %s", OPT(default_window_logo));
+        if (PyErr_Occurred()) PyErr_Print();
+    }
     if (init_gpu_resources) create_gpu_resources_for_window(w);
     else {
         w->render_data.vao_idx = -1;
@@ -250,21 +279,16 @@ add_window(id_type os_window_id, id_type tab_id, PyObject *title) {
 
 static void
 update_window_title(id_type os_window_id, id_type tab_id, id_type window_id, PyObject *title) {
-    WITH_TAB(os_window_id, tab_id);
-    for (size_t i = 0; i < tab->num_windows; i++) {
-        if (tab->windows[i].id == window_id) {
-            Py_CLEAR(tab->windows[i].title);
-            tab->windows[i].title = title;
-            Py_INCREF(tab->windows[i].title);
-            break;
-        }
-    }
-    END_WITH_TAB;
+    WITH_WINDOW(os_window_id, tab_id, window_id)
+        Py_CLEAR(window->title);
+        window->title = title;
+        Py_XINCREF(window->title);
+    END_WITH_WINDOW;
 }
 
 void
 set_os_window_title_from_window(Window *w, OSWindow *os_window) {
-    if (os_window->disallow_title_changes) return;
+    if (os_window->disallow_title_changes || os_window->title_is_overriden) return;
     if (w->title && w->title != os_window->window_title) {
         Py_XDECREF(os_window->window_title);
         os_window->window_title = w->title;
@@ -287,7 +311,13 @@ update_os_window_title(OSWindow *os_window) {
 static void
 destroy_window(Window *w) {
     Py_CLEAR(w->render_data.screen); Py_CLEAR(w->title);
+    Py_CLEAR(w->title_bar_data.last_drawn_title_object_id);
+    free(w->title_bar_data.buf); w->title_bar_data.buf = NULL;
     release_gpu_resources_for_window(w);
+    if (w->window_logo.id) {
+        decref_window_logo(global_state.all_window_logos, w->window_logo.id);
+        w->window_logo.id = 0;
+    }
 }
 
 static void
@@ -439,6 +469,12 @@ remove_os_window(id_type os_window_id) {
     return found;
 }
 
+static void
+mark_os_window_dirty(id_type os_window_id) {
+    WITH_OS_WINDOW(os_window_id)
+        os_window->needs_render = true;
+    END_WITH_OS_WINDOW
+}
 
 static void
 set_active_tab(id_type os_window_id, unsigned int idx) {
@@ -474,7 +510,11 @@ add_borders_rect(id_type os_window_id, id_type tab_id, uint32_t left, uint32_t t
         if (!left && !top && !right && !bottom) { br->num_border_rects = 0; return; }
         ensure_space_for(br, rect_buf, BorderRect, br->num_border_rects + 1, capacity, 32, false);
         BorderRect *r = br->rect_buf + br->num_border_rects++;
-        r->left = left; r->right = right; r->top = top; r->bottom = bottom; r->color = color;
+        r->left = gl_pos_x(left, osw->viewport_width);
+        r->top = gl_pos_y(top, osw->viewport_height);
+        r->right = r->left + gl_size(right - left, osw->viewport_width);
+        r->bottom = r->top - gl_size(bottom - top, osw->viewport_height);
+        r->color = color;
     END_WITH_TAB
 }
 
@@ -561,6 +601,30 @@ send_pending_click_to_window_id(id_type timer_id UNUSED, void *data) {
     }
 }
 
+bool
+update_ime_position_for_window(id_type window_id, bool force) {
+    for (size_t o = 0; o < global_state.num_os_windows; o++) {
+        OSWindow *osw = global_state.os_windows + o;
+        for (size_t t = 0; t < osw->num_tabs; t++) {
+            Tab *qtab = osw->tabs + t;
+            for (size_t w = 0; w < qtab->num_windows; w++) {
+                Window *window = qtab->windows + w;
+                if (window->id == window_id) {
+                    if (window->render_data.screen && (force || osw->is_focused)) {
+                        OSWindow *orig = global_state.callback_os_window;
+                        global_state.callback_os_window = osw;
+                        update_ime_position(window, window->render_data.screen);
+                        global_state.callback_os_window = orig;
+                        return true;
+                    }
+                    return false;
+                }
+            }
+        }
+    }
+    return false;
+}
+
 
 // Python API {{{
 #define PYWRAP0(name) static PyObject* py##name(PYNOARG)
@@ -572,6 +636,7 @@ send_pending_click_to_window_id(id_type timer_id UNUSED, void *data) {
 #define TWO_ID(name) PYWRAP1(name) { id_type a, b; PA("KK", &a, &b); name(a, b); Py_RETURN_NONE; }
 #define THREE_ID(name) PYWRAP1(name) { id_type a, b, c; PA("KKK", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
 #define THREE_ID_OBJ(name) PYWRAP1(name) { id_type a, b, c; PyObject *o; PA("KKKO", &a, &b, &c, &o); name(a, b, c, o); Py_RETURN_NONE; }
+#define K(name) PYWRAP1(name) { id_type a; PA("K", &a); name(a); Py_RETURN_NONE; }
 #define KI(name) PYWRAP1(name) { id_type a; unsigned int b; PA("KI", &a, &b); name(a, b); Py_RETURN_NONE; }
 #define KII(name) PYWRAP1(name) { id_type a; unsigned int b, c; PA("KII", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
 #define KKI(name) PYWRAP1(name) { id_type a, b; unsigned int c; PA("KKI", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
@@ -584,6 +649,14 @@ send_pending_click_to_window_id(id_type timer_id UNUSED, void *data) {
     PyObject *key, *value; Py_ssize_t pos = 0; \
     while (PyDict_Next(d, &pos, &key, &value))
 
+PYWRAP1(update_ime_position_for_window) {
+    id_type window_id;
+    int force = 0;
+    PA("K|p", &window_id, &force);
+    if (update_ime_position_for_window(window_id, force)) Py_RETURN_TRUE;
+    Py_RETURN_FALSE;
+}
+
 PYWRAP0(next_window_id) {
     return PyLong_FromUnsignedLongLong(global_state.window_id_counter + 1);
 }
@@ -632,13 +705,23 @@ PYWRAP1(set_options) {
 
 BOOL_SET(in_sequence_mode)
 
+static void
+init_screen_render_data(OSWindow *osw, const WindowGeometry *g, ScreenRenderData *d) {
+    d->dx = gl_size(osw->fonts_data->cell_width, osw->viewport_width);
+    d->dy = gl_size(osw->fonts_data->cell_height, osw->viewport_height);
+    d->xstart = gl_pos_x(g->left, osw->viewport_width);
+    d->ystart = gl_pos_y(g->top, osw->viewport_height);
+}
+
 PYWRAP1(set_tab_bar_render_data) {
     ScreenRenderData d = {0};
+    WindowGeometry g = {0};
     id_type os_window_id;
-    PA("KffffO", &os_window_id, &d.xstart, &d.ystart, &d.dx, &d.dy, &d.screen);
+    PA("KOIIII", &os_window_id, &d.screen, &g.left, &g.top, &g.right, &g.bottom);
     WITH_OS_WINDOW(os_window_id)
         Py_CLEAR(os_window->tab_bar_render_data.screen);
         d.vao_idx = os_window->tab_bar_render_data.vao_idx;
+        init_screen_render_data(os_window, &g, &d);
         os_window->tab_bar_render_data = d;
         Py_INCREF(os_window->tab_bar_render_data.screen);
     END_WITH_OS_WINDOW
@@ -742,9 +825,10 @@ PYWRAP1(set_titlebar_color) {
     id_type os_window_id;
     unsigned int color;
     int use_system_color = 0;
-    PA("KI|p", &os_window_id, &color, &use_system_color);
+    unsigned int system_color = 0;
+    PA("KI|pI", &os_window_id, &color, &use_system_color, &system_color);
     WITH_OS_WINDOW(os_window_id)
-        set_titlebar_color(os_window, color, use_system_color);
+        set_titlebar_color(os_window, color, use_system_color, system_color);
         Py_RETURN_TRUE;
     END_WITH_OS_WINDOW
     Py_RETURN_FALSE;
@@ -794,12 +878,13 @@ PYWRAP1(set_window_render_data) {
     id_type os_window_id, tab_id, window_id;
     ScreenRenderData d = {0};
     WindowGeometry g = {0};
-    PA("KKKffffOIIII", &os_window_id, &tab_id, &window_id, A(xstart), A(ystart), A(dx), A(dy), A(screen), B(left), B(top), B(right), B(bottom));
+    PA("KKKOIIII", &os_window_id, &tab_id, &window_id, A(screen), B(left), B(top), B(right), B(bottom));
 
     WITH_WINDOW(os_window_id, tab_id, window_id);
         Py_CLEAR(window->render_data.screen);
         d.vao_idx = window->render_data.vao_idx;
         d.gvao_idx = window->render_data.gvao_idx;
+        init_screen_render_data(osw, &g, &d);
         window->render_data = d;
         window->geometry = g;
         Py_INCREF(window->render_data.screen);
@@ -831,6 +916,24 @@ PYWRAP1(sync_os_window_title) {
     Py_RETURN_NONE;
 }
 
+PYWRAP1(set_os_window_title) {
+    id_type os_window_id;
+    const char *title;
+    PA("Ks", &os_window_id, &title);
+    WITH_OS_WINDOW(os_window_id)
+        if (strlen(title)) {
+            os_window->title_is_overriden = true;
+            set_os_window_title(os_window, title);
+        } else {
+            os_window->title_is_overriden = false;
+            if (os_window->window_title) set_os_window_title(os_window, PyUnicode_AsUTF8(os_window->window_title));
+            update_os_window_title(os_window);
+        }
+    END_WITH_OS_WINDOW
+    Py_RETURN_NONE;
+}
+
+
 
 PYWRAP1(pt_to_px) {
     double pt;
@@ -923,7 +1026,15 @@ PYWRAP0(apply_options_update) {
         get_platform_dependent_config_values(os_window->handle);
         os_window->background_opacity = OPT(background_opacity);
         os_window->is_damaged = true;
-        break;
+        for (size_t t = 0; t < os_window->num_tabs; t++) {
+            Tab *tab = os_window->tabs + t;
+            for (size_t w = 0; w < tab->num_windows; w++) {
+                Window *window = tab->windows + w;
+                if (window->window_logo.using_default) {
+                    set_window_logo(window, OPT(default_window_logo), OPT(window_logo_position), OPT(window_logo_alpha), true);
+                }
+            }
+        }
     }
     Py_RETURN_NONE;
 }
@@ -935,10 +1046,11 @@ PYWRAP1(patch_global_colors) {
 #define P(name) { \
     PyObject *val = PyDict_GetItemString(spec, #name); \
     if (val) { \
-        OPT(name) = PyLong_AsLong(val); \
+        if (val == Py_None) OPT(name) = 0; \
+        else if (PyLong_Check(val)) OPT(name) = PyLong_AsLong(val); \
     } \
 }
-    P(active_border_color); P(inactive_border_color); P(bell_border_color);
+    P(active_border_color); P(inactive_border_color); P(bell_border_color); P(tab_bar_background); P(tab_bar_margin_color);
     if (configured) {
         P(background); P(url_color);
         P(mark1_background); P(mark1_foreground); P(mark2_background); P(mark2_foreground);
@@ -954,9 +1066,9 @@ pyset_background_image(PyObject *self UNUSED, PyObject *args) {
     PyObject *layout_name = NULL;
     PyObject *os_window_ids;
     int configured = 0;
-    PA("zO!|pU", &path, &PyTuple_Type, &os_window_ids, &configured, &layout_name);
+    PA("zO!|pO", &path, &PyTuple_Type, &os_window_ids, &configured, &layout_name);
     size_t size;
-    BackgroundImageLayout layout = layout_name ? bglayout(layout_name) : OPT(background_image_layout);
+    BackgroundImageLayout layout = PyUnicode_Check(layout_name) ? bglayout(layout_name) : OPT(background_image_layout);
     BackgroundImage *bgimage = NULL;
     if (path) {
         bgimage = calloc(1, sizeof(BackgroundImage));
@@ -1020,13 +1132,35 @@ pycreate_mock_window(PyObject *self UNUSED, PyObject *args) {
     return ans;
 }
 
-static void
+static bool
 click_mouse_url(id_type os_window_id, id_type tab_id, id_type window_id) {
+    bool clicked = false;
     WITH_WINDOW(os_window_id, tab_id, window_id);
-    mouse_open_url(window);
+    clicked = mouse_open_url(window);
     END_WITH_WINDOW;
+    return clicked;
 }
 
+static bool
+click_mouse_cmd_output(id_type os_window_id, id_type tab_id, id_type window_id, int select_cmd_output) {
+    bool handled = false;
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+    handled = mouse_set_last_visited_cmd_output(window);
+    if (select_cmd_output && handled) handled = mouse_select_cmd_output(window);
+    END_WITH_WINDOW;
+    return handled;
+}
+
+static bool
+move_cursor_to_mouse_if_in_prompt(id_type os_window_id, id_type tab_id, id_type window_id) {
+    bool moved = false;
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+    moved = move_cursor_to_mouse_if_at_shell_prompt(window);
+    END_WITH_WINDOW;
+    return moved;
+}
+
+
 static PyObject*
 pymouse_selection(PyObject *self UNUSED, PyObject *args) {
     id_type os_window_id, tab_id, window_id;
@@ -1038,9 +1172,45 @@ pymouse_selection(PyObject *self UNUSED, PyObject *args) {
     Py_RETURN_NONE;
 }
 
+PYWRAP1(set_window_logo) {
+    id_type os_window_id, tab_id, window_id;
+    const char *path; PyObject *position;
+    float alpha = 0.5;
+    PA("KKKsUf", &os_window_id, &tab_id, &window_id, &path, &position, &alpha);
+    bool ok = false;
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+    ok = set_window_logo(window, path, PyObject_IsTrue(position) ? bganchor(position) : OPT(window_logo_position), (0 <= alpha && alpha <= 1) ? alpha : OPT(window_logo_alpha), false);
+    END_WITH_WINDOW;
+    if (ok) Py_RETURN_TRUE;
+    Py_RETURN_FALSE;
+}
+
+
+PYWRAP1(click_mouse_url) {
+    id_type a, b, c; PA("KKK", &a, &b, &c);
+    if (click_mouse_url(a, b, c)) { Py_RETURN_TRUE; }
+    Py_RETURN_FALSE;
+}
+
+PYWRAP1(click_mouse_cmd_output) {
+    id_type a, b, c; int d; PA("KKKp", &a, &b, &c, &d);
+    if (click_mouse_cmd_output(a, b, c, d)) { Py_RETURN_TRUE; }
+    Py_RETURN_FALSE;
+}
+
+PYWRAP1(move_cursor_to_mouse_if_in_prompt) {
+    id_type a, b, c; PA("KKK", &a, &b, &c);
+    if (move_cursor_to_mouse_if_in_prompt(a, b, c)) Py_RETURN_TRUE;
+    Py_RETURN_FALSE;
+}
+
+PYWRAP1(redirect_mouse_handling) {
+    global_state.redirect_mouse_handling = PyObject_IsTrue(args) ? true : false;
+    Py_RETURN_NONE;
+}
+
 THREE_ID_OBJ(update_window_title)
 THREE_ID(remove_window)
-THREE_ID(click_mouse_url)
 THREE_ID(detach_window)
 THREE_ID(attach_window)
 PYWRAP1(resolve_key_mods) { int mods, kitty_mod; PA("ii", &kitty_mod, &mods); return PyLong_FromLong(resolve_mods(kitty_mod, mods)); }
@@ -1049,6 +1219,7 @@ PYWRAP1(add_window) { PyObject *title; id_type a, b; PA("KKO", &a, &b, &title);
 PYWRAP0(current_os_window) { OSWindow *w = current_os_window(); if (!w) Py_RETURN_NONE; return PyLong_FromUnsignedLongLong(w->id); }
 TWO_ID(remove_tab)
 KI(set_active_tab)
+K(mark_os_window_dirty)
 KKK(set_active_window)
 KII(swap_tabs)
 KK5I(add_borders_rect)
@@ -1062,10 +1233,15 @@ static PyMethodDef module_methods[] = {
     MW(set_options, METH_VARARGS),
     MW(get_options, METH_NOARGS),
     MW(click_mouse_url, METH_VARARGS),
+    MW(click_mouse_cmd_output, METH_VARARGS),
+    MW(move_cursor_to_mouse_if_in_prompt, METH_VARARGS),
+    MW(redirect_mouse_handling, METH_O),
     MW(mouse_selection, METH_VARARGS),
+    MW(set_window_logo, METH_VARARGS),
     MW(set_in_sequence_mode, METH_O),
     MW(resolve_key_mods, METH_VARARGS),
     MW(handle_for_window_id, METH_VARARGS),
+    MW(update_ime_position_for_window, METH_VARARGS),
     MW(pt_to_px, METH_VARARGS),
     MW(add_tab, METH_O),
     MW(add_window, METH_VARARGS),
@@ -1075,6 +1251,7 @@ static PyMethodDef module_methods[] = {
     MW(detach_window, METH_VARARGS),
     MW(attach_window, METH_VARARGS),
     MW(set_active_tab, METH_VARARGS),
+    MW(mark_os_window_dirty, METH_VARARGS),
     MW(set_active_window, METH_VARARGS),
     MW(swap_tabs, METH_VARARGS),
     MW(add_borders_rect, METH_VARARGS),
@@ -1094,6 +1271,7 @@ static PyMethodDef module_methods[] = {
     MW(background_opacity_of, METH_O),
     MW(update_window_visibility, METH_VARARGS),
     MW(sync_os_window_title, METH_VARARGS),
+    MW(set_os_window_title, METH_VARARGS),
     MW(global_font_size, METH_VARARGS),
     MW(set_background_image, METH_VARARGS),
     MW(os_window_font_size, METH_VARARGS),
@@ -1116,12 +1294,15 @@ finalize(void) {
     }
     if (detached_windows.windows) free(detached_windows.windows);
     detached_windows.capacity = 0;
-    if (OPT(background_image)) free(OPT(background_image));
+#define F(x) free(OPT(x)); OPT(x) = NULL;
+    F(background_image); F(bell_path); F(default_window_logo);
+#undef F
     // we leak the texture here since it is not guaranteed
     // that freeing the texture will work during shutdown and
     // the GPU driver should take care of it when the OpenGL context is
     // destroyed.
     free_bgimage(&global_state.bgimage, false);
+    free_window_logo_table(&global_state.all_window_logos);
     global_state.bgimage = NULL;
     free_url_prefixes();
     free(OPT(select_by_word_characters)); OPT(select_by_word_characters) = NULL;
@@ -1137,6 +1318,8 @@ init_state(PyObject *module) {
 #define DPI 96.0
 #endif
     global_state.default_dpi.x = DPI; global_state.default_dpi.y = DPI;
+    global_state.all_window_logos = alloc_window_logo_table();
+    if (!global_state.all_window_logos) { PyErr_NoMemory(); return false; }
     if (PyModule_AddFunctions(module, module_methods) != 0) return false;
     if (PyStructSequence_InitType2(&RegionType, &region_desc) != 0) return false;
     Py_INCREF((PyObject *) &RegionType);
