@@ -8,6 +8,7 @@
 #include "data-types.h"
 #include "screen.h"
 #include "monotonic.h"
+#include "window_logo.h"
 
 #define OPT(name) global_state.opts.name
 
@@ -31,7 +32,7 @@ typedef struct {
     unsigned int scrollback_pager_history_size;
     bool scrollback_fill_enlarged_window;
     char_type *select_by_word_characters;
-    color_type url_color, background, foreground, active_border_color, inactive_border_color, bell_border_color;
+    color_type url_color, background, foreground, active_border_color, inactive_border_color, bell_border_color, tab_bar_background, tab_bar_margin_color;
     color_type mark1_foreground, mark1_background, mark2_foreground, mark2_background, mark3_foreground, mark3_background;
     monotonic_t repaint_delay, input_delay;
     bool focus_follows_mouse;
@@ -40,14 +41,16 @@ typedef struct {
     unsigned int macos_option_as_alt;
     float macos_thicken_font;
     WindowTitleIn macos_show_window_title_in;
+    char *bell_path;
     int adjust_line_height_px, adjust_column_width_px, adjust_baseline_px;
     float adjust_line_height_frac, adjust_column_width_frac, adjust_baseline_frac;
     float background_opacity, dim_opacity;
 
-    char* background_image;
+    char *background_image, *default_window_logo;
     BackgroundImageLayout background_image_layout;
+    ImageAnchorPosition window_logo_position;
     bool background_image_linear;
-    float background_tint;
+    float background_tint, window_logo_alpha;
 
     bool dynamic_background_opacity;
     float inactive_text_alpha;
@@ -79,9 +82,17 @@ typedef struct {
     } tab_bar_margin_height;
 } Options;
 
+typedef struct WindowLogoRenderData {
+    window_logo_id_t id;
+    WindowLogo *instance;
+    ImageAnchorPosition position;
+    float alpha;
+    bool using_default;
+} WindowLogoRenderData;
+
 typedef struct {
     ssize_t vao_idx, gvao_idx;
-    float xstart, ystart, dx, dy;
+    float xstart, ystart, dx, dy, xratio, yratio;
     Screen *screen;
 } ScreenRenderData;
 
@@ -101,6 +112,12 @@ typedef struct {
     unsigned int length;
 } ClickQueue;
 
+typedef struct MousePosition {
+    unsigned int cell_x, cell_y;
+    double global_x, global_y;
+    bool in_left_half_of_cell;
+} MousePosition;
+
 typedef struct {
     id_type id;
     bool visible, cursor_visible_at_last_render;
@@ -108,11 +125,8 @@ typedef struct {
     CursorShape last_cursor_shape;
     PyObject *title;
     ScreenRenderData render_data;
-    struct {
-        unsigned int cell_x, cell_y;
-        double x, y;
-        bool in_left_half_of_cell;
-    } mouse_pos;
+    WindowLogoRenderData window_logo;
+    MousePosition mouse_pos;
     struct {
         unsigned int left, top, right, bottom;
     } padding;
@@ -120,10 +134,16 @@ typedef struct {
     ClickQueue click_queues[8];
     monotonic_t last_drag_scroll_at;
     uint32_t last_special_key_pressed;
+    struct {
+        unsigned width, height;
+        uint8_t *buf;
+        PyObject *last_drawn_title_object_id;
+    } title_bar_data;
 } Window;
 
 typedef struct {
-    uint32_t left, top, right, bottom, color;
+    float left, top, right, bottom;
+    uint32_t color;
 } BorderRect;
 
 typedef struct {
@@ -140,11 +160,6 @@ typedef struct {
     BorderRects border_rects;
 } Tab;
 
-typedef struct {
-    int x, y, w, h;
-    bool is_set;
-} OSWindowGeometry;
-
 enum RENDER_STATE { RENDER_FRAME_NOT_REQUESTED, RENDER_FRAME_REQUESTED, RENDER_FRAME_READY };
 typedef enum { NO_CLOSE_REQUESTED, CONFIRMABLE_CLOSE_REQUESTED, CLOSE_BEING_CONFIRMED, IMPERATIVE_CLOSE_REQUESTED } CloseRequest;
 
@@ -161,7 +176,10 @@ typedef struct {
     void *handle;
     id_type id;
     uint32_t offscreen_framebuffer;
-    OSWindowGeometry before_fullscreen;
+    struct {
+        int x, y, w, h;
+        bool is_set, was_maximized;
+    } before_fullscreen;
     int viewport_width, viewport_height, window_width, window_height;
     double viewport_x_ratio, viewport_y_ratio;
     Tab *tabs;
@@ -176,7 +194,7 @@ typedef struct {
     double logical_dpi_x, logical_dpi_y, font_sz_in_pts;
     bool mouse_button_pressed[32];
     PyObject *window_title;
-    bool disallow_title_changes;
+    bool disallow_title_changes, title_is_overriden;
     bool viewport_size_dirty, viewport_updated_at_least_once;
     LiveResizeInfo live_resize;
     bool has_pending_resizes, is_semi_transparent, shown_once, is_damaged;
@@ -198,7 +216,7 @@ typedef struct {
 typedef struct {
     Options opts;
 
-    id_type os_window_id_counter, tab_id_counter, window_id_counter;
+    id_type os_window_id_counter, tab_id_counter, window_id_counter, current_opengl_context_id;
     PyObject *boss;
     BackgroundImage *bgimage;
     OSWindow *os_windows;
@@ -211,9 +229,11 @@ typedef struct {
     bool in_sequence_mode;
     bool check_for_active_animated_images;
     struct { double x, y; } default_dpi;
-    id_type active_drag_in_window;
+    id_type active_drag_in_window, tracked_drag_in_window;
     int active_drag_button;
     CloseRequest quit_request;
+    bool redirect_mouse_handling;
+    WindowLogoTable *all_window_logos;
 } GlobalState;
 
 extern GlobalState global_state;
@@ -253,7 +273,7 @@ ssize_t create_cell_vao(void);
 ssize_t create_graphics_vao(void);
 ssize_t create_border_vao(void);
 bool send_cell_data_to_gpu(ssize_t, ssize_t, float, float, float, float, Screen *, OSWindow *);
-void draw_cells(ssize_t, ssize_t, float, float, float, float, Screen *, OSWindow *, bool, bool);
+void draw_cells(ssize_t, ssize_t, const ScreenRenderData*, float, float, OSWindow *, bool, bool, Window*);
 void draw_centered_alpha_mask(OSWindow *w, size_t screen_width, size_t screen_height, size_t width, size_t height, uint8_t *canvas);
 void update_surface_size(int, int, uint32_t);
 void free_texture(uint32_t*);
@@ -262,7 +282,7 @@ void send_image_to_gpu(uint32_t*, const void*, int32_t, int32_t, bool, bool, boo
 void send_sprite_to_gpu(FONTS_DATA_HANDLE fg, unsigned int, unsigned int, unsigned int, pixel*);
 void blank_canvas(float, color_type);
 void blank_os_window(OSWindow *);
-void set_titlebar_color(OSWindow *w, color_type color, bool use_system_color);
+void set_titlebar_color(OSWindow *w, color_type color, bool use_system_color, unsigned int system_color);
 FONTS_DATA_HANDLE load_fonts_data(double, double, double);
 void send_prerendered_sprites_for_window(OSWindow *w);
 #ifdef __APPLE__
@@ -302,9 +322,17 @@ void set_os_window_title_from_window(Window *w, OSWindow *os_window);
 void update_os_window_title(OSWindow *os_window);
 void fake_scroll(Window *w, int amount, bool upwards);
 Window* window_for_window_id(id_type kitty_window_id);
-void mouse_open_url(Window *w);
+bool mouse_open_url(Window *w);
+bool mouse_set_last_visited_cmd_output(Window *w);
+bool mouse_select_cmd_output(Window *w);
+bool move_cursor_to_mouse_if_at_shell_prompt(Window *w);
 void mouse_selection(Window *w, int code, int button);
 const char* format_mods(unsigned mods);
 void send_pending_click_to_window_id(id_type, void*);
 void send_pending_click_to_window(Window*, void*);
 void get_platform_dependent_config_values(void *glfw_window);
+bool draw_window_title(OSWindow *window, const char *text, color_type fg, color_type bg, uint8_t *output_buf, size_t width, size_t height);
+uint8_t* draw_single_ascii_char(const char ch, size_t *result_width, size_t *result_height);
+bool is_os_window_fullscreen(OSWindow *);
+void update_ime_position(Window* w, Screen *screen);
+bool update_ime_position_for_window(id_type window_id, bool force);
