@@ -1,21 +1,23 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
-from functools import lru_cache
-from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Tuple
-
-from .config import build_ansi_color_table
-from .fast_data_types import (
-    DECAWM, Screen, cell_size_for_window, get_options, pt_to_px,
-    set_tab_bar_render_data, viewport_for_window
+import os
+from functools import lru_cache, partial, wraps
+from typing import (
+    Any, Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple, Union
 )
-from .layout.base import Rect
-from .rgb import Color, alpha_blend, color_as_sgr, color_from_int, to_color
+
+from .borders import Border, BorderColor
+from .config import build_ansi_color_table
+from .constants import config_dir
+from .fast_data_types import (
+    DECAWM, Region, Screen, cell_size_for_window, get_options, pt_to_px,
+    set_tab_bar_render_data, viewport_for_window, Color
+)
+from .rgb import alpha_blend, color_as_sgr, color_from_int, to_color
 from .types import WindowGeometry, run_once
-from .typing import PowerlineStyle
+from .typing import EdgeLiteral, PowerlineStyle
 from .utils import color_as_int, log_error
-from .window import calculate_gl_geometry
 
 
 class TabBarData(NamedTuple):
@@ -23,8 +25,13 @@ class TabBarData(NamedTuple):
     is_active: bool
     needs_attention: bool
     num_windows: int
+    num_window_groups: int
     layout_name: str
     has_activity_since_last_focus: bool
+    active_fg: Optional[int]
+    active_bg: Optional[int]
+    inactive_fg: Optional[int]
+    inactive_bg: Optional[int]
 
 
 class DrawData(NamedTuple):
@@ -43,6 +50,25 @@ class DrawData(NamedTuple):
     active_title_template: Optional[str]
     tab_activity_symbol: Optional[str]
     powerline_style: PowerlineStyle
+    tab_bar_edge: EdgeLiteral
+
+    def tab_fg(self, tab: TabBarData) -> int:
+        if tab.is_active:
+            if tab.active_fg is not None:
+                return tab.active_fg
+            return int(self.active_fg)
+        if tab.inactive_fg is not None:
+            return tab.inactive_fg
+        return int(self.inactive_fg)
+
+    def tab_bg(self, tab: TabBarData) -> int:
+        if tab.is_active:
+            if tab.active_bg is not None:
+                return tab.active_bg
+            return int(self.active_bg)
+        if tab.inactive_bg is not None:
+            return tab.inactive_bg
+        return int(self.inactive_bg)
 
 
 def as_rgb(x: int) -> int:
@@ -51,7 +77,7 @@ def as_rgb(x: int) -> int:
 
 @lru_cache()
 def report_template_failure(template: str, e: str) -> None:
-    log_error('Invalid tab title template: "{}" with error: {}'.format(template, e))
+    log_error(f'Invalid tab title template: "{template}" with error: {e}')
 
 
 @lru_cache()
@@ -64,6 +90,9 @@ def compile_template(template: str) -> Any:
 
 class ColorFormatter:
 
+    draw_data: DrawData
+    tab_data: TabBarData
+
     def __init__(self, which: str):
         self.which = which
 
@@ -71,6 +100,9 @@ class ColorFormatter:
         q = name
         if q == 'default':
             ans = '9'
+        elif q == 'tab':
+            col = color_from_int((self.draw_data.tab_bg if self.which == '4' else self.draw_data.tab_fg)(self.tab_data))
+            ans = '8' + color_as_sgr(col)
         else:
             if name.startswith('_'):
                 q = '#' + name[1:]
@@ -92,7 +124,7 @@ class Formatter:
 
 
 @run_once
-def super_sub_maps() -> Tuple[dict, dict]:
+def super_sub_maps() -> Tuple[Dict[int, Union[None, int]], Dict[int, Union[None, int]]]:
     import string
     sup_table = str.maketrans(
         string.ascii_lowercase + string.ascii_uppercase + string.digits + '+-=()',
@@ -105,7 +137,7 @@ def super_sub_maps() -> Tuple[dict, dict]:
 
 class SupSub:
 
-    def __init__(self, data: dict, is_subscript: bool = False):
+    def __init__(self, data: Dict[str, Any], is_subscript: bool = False):
         self.__data = data
         self.__is_subscript = is_subscript
 
@@ -115,6 +147,11 @@ class SupSub:
         return name.translate(table)
 
 
+class ExtraData:
+    prev_tab: Optional[TabBarData] = None
+    next_tab: Optional[TabBarData] = None
+
+
 def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int) -> None:
     if tab.needs_attention and draw_data.bell_on_tab:
         fg = screen.cursor.fg
@@ -135,12 +172,16 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int)
             'index': index,
             'layout_name': tab.layout_name,
             'num_windows': tab.num_windows,
+            'num_window_groups': tab.num_window_groups,
             'title': tab.title,
         }
+        ColorFormatter.draw_data = draw_data
+        ColorFormatter.tab_data = tab
         eval_locals = {
             'index': index,
             'layout_name': tab.layout_name,
             'num_windows': tab.num_windows,
+            'num_window_groups': tab.num_window_groups,
             'title': tab.title,
             'fmt': Formatter,
             'sup': SupSub(data),
@@ -161,9 +202,57 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int)
         screen.draw(title)
 
 
-def draw_tab_with_separator(draw_data: DrawData, screen: Screen, tab: TabBarData, before: int, max_title_length: int, index: int, is_last: bool) -> int:
-    tab_bg = draw_data.active_bg if tab.is_active else draw_data.inactive_bg
-    screen.cursor.bg = as_rgb(color_as_int(tab_bg))
+DrawTabFunc = Callable[[DrawData, Screen, TabBarData, int, int, int, bool, ExtraData], int]
+
+
+def draw_tab_with_slant(
+    draw_data: DrawData, screen: Screen, tab: TabBarData,
+    before: int, max_title_length: int, index: int, is_last: bool,
+    extra_data: ExtraData
+) -> int:
+    orig_fg = screen.cursor.fg
+    left_sep, right_sep = ('', '') if draw_data.tab_bar_edge == 'top' else ('', '')
+    tab_bg = screen.cursor.bg
+    slant_fg = as_rgb(color_as_int(draw_data.default_bg))
+
+    def draw_sep(which: str) -> None:
+        screen.cursor.bg = tab_bg
+        screen.cursor.fg = slant_fg
+        screen.draw(which)
+        screen.cursor.bg = tab_bg
+        screen.cursor.fg = orig_fg
+
+    max_title_length += 1
+    if max_title_length <= 1:
+        screen.draw('…')
+    elif max_title_length == 2:
+        screen.draw('…|')
+    elif max_title_length < 6:
+        draw_sep(left_sep)
+        screen.draw((' ' if max_title_length == 5 else '') + '…' + (' ' if max_title_length >= 4 else ''))
+        draw_sep(right_sep)
+    else:
+        draw_sep(left_sep)
+        screen.draw(' ')
+        draw_title(draw_data, screen, tab, index)
+        extra = screen.cursor.x - before - max_title_length
+        if extra >= 0:
+            screen.cursor.x -= extra + 3
+            screen.draw('…')
+        elif extra == -1:
+            screen.cursor.x -= 2
+            screen.draw('…')
+        screen.draw(' ')
+        draw_sep(right_sep)
+
+    return screen.cursor.x
+
+
+def draw_tab_with_separator(
+    draw_data: DrawData, screen: Screen, tab: TabBarData,
+    before: int, max_title_length: int, index: int, is_last: bool,
+    extra_data: ExtraData
+) -> int:
     if draw_data.leading_spaces:
         screen.draw(' ' * draw_data.leading_spaces)
     draw_title(draw_data, screen, tab, index)
@@ -185,13 +274,18 @@ def draw_tab_with_separator(draw_data: DrawData, screen: Screen, tab: TabBarData
     return end
 
 
-def draw_tab_with_fade(draw_data: DrawData, screen: Screen, tab: TabBarData, before: int, max_title_length: int, index: int, is_last: bool) -> int:
-    tab_bg = draw_data.active_bg if tab.is_active else draw_data.inactive_bg
+def draw_tab_with_fade(
+    draw_data: DrawData, screen: Screen, tab: TabBarData,
+    before: int, max_title_length: int, index: int, is_last: bool,
+    extra_data: ExtraData
+) -> int:
+    orig_bg = screen.cursor.bg
+    tab_bg = color_from_int(orig_bg >> 8)
     fade_colors = [as_rgb(color_as_int(alpha_blend(tab_bg, draw_data.default_bg, alpha))) for alpha in draw_data.alpha]
     for bg in fade_colors:
         screen.cursor.bg = bg
         screen.draw(' ')
-    screen.cursor.bg = as_rgb(color_as_int(tab_bg))
+    screen.cursor.bg = orig_bg
     draw_title(draw_data, screen, tab, index)
     extra = screen.cursor.x - before - max_title_length
     if extra > 0:
@@ -219,30 +313,26 @@ powerline_symbols: Dict[PowerlineStyle, Tuple[str, str]] = {
 }
 
 
-def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData, before: int, max_title_length: int, index: int, is_last: bool) -> int:
-    tab_bg = as_rgb(color_as_int(draw_data.active_bg if tab.is_active else draw_data.inactive_bg))
-    tab_fg = as_rgb(color_as_int(draw_data.active_fg if tab.is_active else draw_data.inactive_fg))
-    inactive_bg = as_rgb(color_as_int(draw_data.inactive_bg))
-    default_bg = as_rgb(color_as_int(draw_data.default_bg))
+def draw_tab_with_powerline(
+    draw_data: DrawData, screen: Screen, tab: TabBarData,
+    before: int, max_title_length: int, index: int, is_last: bool,
+    extra_data: ExtraData
+) -> int:
+    tab_bg = screen.cursor.bg
+    tab_fg = screen.cursor.fg
+    default_bg = as_rgb(int(draw_data.default_bg))
+    if extra_data.next_tab:
+        next_tab_bg = as_rgb(draw_data.tab_bg(extra_data.next_tab))
+        needs_soft_separator = next_tab_bg == tab_bg
+    else:
+        next_tab_bg = default_bg
+        needs_soft_separator = False
 
-    separator_symbol, separator_alt_symbol = powerline_symbols.get(draw_data.powerline_style, ('', ''))
+    separator_symbol, soft_separator_symbol = powerline_symbols.get(draw_data.powerline_style, ('', ''))
     min_title_length = 1 + 2
-
-    if screen.cursor.x + min_title_length >= screen.columns:
-        screen.cursor.x -= 2
-        screen.cursor.bg = default_bg
-        screen.cursor.fg = inactive_bg
-        screen.draw(f'{separator_symbol}   ')
-        return screen.cursor.x
-
     start_draw = 2
-    if tab.is_active and screen.cursor.x >= 2:
-        screen.cursor.x -= 2
-        screen.cursor.fg = inactive_bg
-        screen.cursor.bg = tab_bg
-        screen.draw(f'{separator_symbol} ')
-        screen.cursor.fg = tab_fg
-    elif screen.cursor.x == 0:
+
+    if screen.cursor.x == 0:
         screen.cursor.bg = tab_bg
         screen.draw(' ')
         start_draw = 1
@@ -257,13 +347,10 @@ def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData
             screen.cursor.x -= extra + 1
             screen.draw('…')
 
-    if tab.is_active or is_last:
+    if not needs_soft_separator:
         screen.draw(' ')
         screen.cursor.fg = tab_bg
-        if is_last:
-            screen.cursor.bg = default_bg
-        else:
-            screen.cursor.bg = inactive_bg
+        screen.cursor.bg = next_tab_bg
         screen.draw(separator_symbol)
     else:
         prev_fg = screen.cursor.fg
@@ -274,7 +361,7 @@ def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData
             c2 = draw_data.inactive_bg.contrast(draw_data.inactive_fg)
             if c1 < c2:
                 screen.cursor.fg = default_bg
-        screen.draw(f' {separator_alt_symbol}')
+        screen.draw(f' {soft_separator_symbol}')
         screen.cursor.fg = prev_fg
 
     end = screen.cursor.x
@@ -283,13 +370,40 @@ def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData
     return end
 
 
+@run_once
+def load_custom_draw_tab() -> DrawTabFunc:
+    import runpy
+    import traceback
+    try:
+        m = runpy.run_path(os.path.join(config_dir, 'tab_bar.py'))
+        func: DrawTabFunc = m['draw_tab']
+    except Exception as e:
+        traceback.print_exc()
+        log_error(f'Failed to load custom draw_tab function with error: {e}')
+        return draw_tab_with_fade
+
+    @wraps(func)
+    def draw_tab(
+        draw_data: DrawData, screen: Screen, tab: TabBarData,
+        before: int, max_title_length: int, index: int, is_last: bool,
+        extra_data: ExtraData
+    ) -> int:
+        try:
+            return func(draw_data, screen, tab, before, max_title_length, index, is_last, extra_data)
+        except Exception as e:
+            log_error(f'Custom draw tab function failed with error: {e}')
+            return draw_tab_with_fade(draw_data, screen, tab, before, max_title_length, index, is_last, extra_data)
+
+    return draw_tab
+
+
 class TabBar:
 
     def __init__(self, os_window_id: int):
         self.os_window_id = os_window_id
         self.num_tabs = 1
         self.data_buffer_size = 0
-        self.blank_rects: Tuple[Rect, ...] = ()
+        self.blank_rects: Tuple[Border, ...] = ()
         self.laid_out_once = False
         self.apply_options()
 
@@ -329,40 +443,83 @@ class TabBar:
             opts.tab_bar_background or opts.background, opts.tab_title_template,
             opts.active_tab_title_template,
             opts.tab_activity_symbol,
-            opts.tab_powerline_style
+            opts.tab_powerline_style,
+            'top' if opts.tab_bar_edge == 1 else 'bottom'
         )
-        if opts.tab_bar_style == 'separator':
-            self.draw_func = draw_tab_with_separator
-        elif opts.tab_bar_style == 'powerline':
+        ts = opts.tab_bar_style
+        if ts == 'separator':
+            self.draw_func: DrawTabFunc = draw_tab_with_separator
+        elif ts == 'powerline':
             self.draw_func = draw_tab_with_powerline
+        elif ts == 'slant':
+            self.draw_func = draw_tab_with_slant
+        elif ts == 'custom':
+            self.draw_func = load_custom_draw_tab()
         else:
             self.draw_func = draw_tab_with_fade
+        if opts.tab_bar_align == 'center':
+            self.align: Callable[[], None] = partial(self.align_with_factor, 2)
+        elif opts.tab_bar_align == 'right':
+            self.align = self.align_with_factor
+        else:
+            self.align = lambda: None
 
-    def patch_colors(self, spec: Dict[str, Any]) -> None:
-        if 'active_tab_foreground' in spec:
-            self.active_fg = (spec['active_tab_foreground'] << 8) | 2
-            self.draw_data = self.draw_data._replace(active_fg=color_from_int(spec['active_tab_foreground']))
-        if 'active_tab_background' in spec:
-            self.active_bg = (spec['active_tab_background'] << 8) | 2
-            self.draw_data = self.draw_data._replace(active_bg=color_from_int(spec['active_tab_background']))
-        if 'inactive_tab_background' in spec:
-            self.draw_data = self.draw_data._replace(inactive_bg=color_from_int(spec['inactive_tab_background']))
-        if 'tab_bar_background' in spec:
-            self.draw_data = self.draw_data._replace(default_bg=color_from_int(spec['tab_bar_background']))
+    def patch_colors(self, spec: Dict[str, Optional[int]]) -> None:
         opts = get_options()
-        fg = spec.get('inactive_tab_foreground', color_as_int(opts.inactive_tab_foreground))
+        atf = spec.get('active_tab_foreground')
+        if isinstance(atf, int):
+            self.active_fg = (atf << 8) | 2
+            self.draw_data = self.draw_data._replace(active_fg=color_from_int(atf))
+        atb = spec.get('active_tab_background')
+        if isinstance(atb, int):
+            self.active_bg = (atb << 8) | 2
+            self.draw_data = self.draw_data._replace(active_bg=color_from_int(atb))
+        itb = spec.get('inactive_tab_background')
+        if isinstance(itb, int):
+            self.draw_data = self.draw_data._replace(inactive_bg=color_from_int(itb))
+        if 'tab_bar_background' in spec:
+            val = spec['tab_bar_background']
+            if val is None:
+                val = color_as_int(opts.background)
+            self.draw_data = self.draw_data._replace(default_bg=color_from_int(val))
+        elif not opts.tab_bar_background:
+            self.draw_data = self.draw_data._replace(default_bg=opts.background)
         bg = spec.get('tab_bar_background', False)
         if bg is None:
             bg = color_as_int(opts.background)
         elif bg is False:
             bg = color_as_int(opts.tab_bar_background or opts.background)
+        fg = spec.get('inactive_tab_foreground')
+        if fg is None:
+            fg = color_as_int(opts.inactive_tab_foreground)
         self.screen.color_profile.set_configured_colors(fg, bg)
 
+    def update_blank_rects(self, central: Region, tab_bar: Region, vw: int, vh: int) -> None:
+        opts = get_options()
+        blank_rects: List[Border] = []
+        bg = BorderColor.tab_bar_margin_color if opts.tab_bar_margin_color is not None else BorderColor.default_bg
+        if opts.tab_bar_margin_height:
+            if opts.tab_bar_edge == 3:  # bottom
+                if opts.tab_bar_margin_height.outer:
+                    blank_rects.append(Border(0, tab_bar.bottom + 1, vw, vh, bg))
+                if opts.tab_bar_margin_height.inner:
+                    blank_rects.append(Border(0, central.bottom + 1, vw, vh, bg))
+            else:  # top
+                if opts.tab_bar_margin_height.outer:
+                    blank_rects.append(Border(0, 0, vw, tab_bar.top, bg))
+                if opts.tab_bar_margin_height.inner:
+                    blank_rects.append(Border(0, tab_bar.bottom + 1, vw, central.top, bg))
+        g = self.window_geometry
+        if g.left > 0:
+            blank_rects.append(Border(0, g.top, g.left, g.bottom + 1, bg))
+        if g.right - 1 < vw:
+            blank_rects.append(Border(g.right - 1, g.top, vw, g.bottom + 1, bg))
+        self.blank_rects = tuple(blank_rects)
+
     def layout(self) -> None:
         central, tab_bar, vw, vh, cell_width, cell_height = viewport_for_window(self.os_window_id)
         if tab_bar.width < 2:
             return
-        opts = get_options()
         self.cell_width = cell_width
         s = self.screen
         viewport_width = max(4 * cell_width, tab_bar.width - 2 * self.margin_width)
@@ -373,25 +530,8 @@ class TabBar:
         margin = (viewport_width - ncells * cell_width) // 2 + self.margin_width
         self.window_geometry = g = WindowGeometry(
             margin, tab_bar.top, viewport_width - margin, tab_bar.bottom, s.columns, s.lines)
-        blank_rects: List[Rect] = []
-        if margin > 0:
-            blank_rects.append(Rect(0, g.top, g.left, g.bottom + 1))
-            blank_rects.append(Rect(g.right - 1, g.top, viewport_width, g.bottom + 1))
-        if opts.tab_bar_margin_height:
-            if opts.tab_bar_edge == 3:  # bottom
-                if opts.tab_bar_margin_height.outer:
-                    blank_rects.append(Rect(0, tab_bar.bottom + 1, vw, vh))
-                if opts.tab_bar_margin_height.inner:
-                    blank_rects.append(Rect(0, central.bottom + 1, vw, vh))
-            else:  # top
-                if opts.tab_bar_margin_height.outer:
-                    blank_rects.append(Rect(0, 0, vw, tab_bar.top))
-                if opts.tab_bar_margin_height.inner:
-                    blank_rects.append(Rect(0, tab_bar.bottom + 1, vw, central.top))
-
-        self.blank_rects = tuple(blank_rects)
-        self.screen_geometry = sg = calculate_gl_geometry(g, vw, vh, cell_width, cell_height)
-        set_tab_bar_render_data(self.os_window_id, sg.xstart, sg.ystart, sg.dx, sg.dy, self.screen)
+        self.update_blank_rects(central, tab_bar, vw, vh)
+        set_tab_bar_render_data(self.os_window_id, self.screen, *g[:4])
 
     def update(self, data: Sequence[TabBarData]) -> None:
         if not self.laid_out_once:
@@ -399,16 +539,19 @@ class TabBar:
         s = self.screen
         s.cursor.x = 0
         s.erase_in_line(2, False)
-        max_title_length = max(1, (self.screen_geometry.xnum // max(1, len(data))) - 1)
+        max_title_length = max(1, (self.screen.columns // max(1, len(data))) - 1)
         cr = []
         last_tab = data[-1] if data else None
+        ed = ExtraData()
 
         for i, t in enumerate(data):
-            s.cursor.bg = self.active_bg if t.is_active else 0
-            s.cursor.fg = self.active_fg if t.is_active else 0
+            ed.prev_tab = data[i - 1] if i > 0 else None
+            ed.next_tab = data[i + 1] if i + 1 < len(data) else None
+            s.cursor.bg = as_rgb(self.draw_data.tab_bg(t))
+            s.cursor.fg = as_rgb(self.draw_data.tab_fg(t))
             s.cursor.bold, s.cursor.italic = self.active_font_style if t.is_active else self.inactive_font_style
             before = s.cursor.x
-            end = self.draw_func(self.draw_data, s, t, before, max_title_length, i + 1, t is last_tab)
+            end = self.draw_func(self.draw_data, s, t, before, max_title_length, i + 1, t is last_tab, ed)
             s.cursor.bg = s.cursor.fg = 0
             cr.append((before, end))
             if s.cursor.x > s.columns - max_title_length and t is not last_tab:
@@ -419,13 +562,26 @@ class TabBar:
                 break
         s.erase_in_line(0, False)  # Ensure no long titles bleed after the last tab
         self.cell_ranges = cr
+        self.align()
+
+    def align_with_factor(self, factor: int = 1) -> None:
+        if not self.cell_ranges:
+            return
+        end = self.cell_ranges[-1][1]
+        if end < self.screen.columns - 1:
+            shift = (self.screen.columns - end) // factor
+            self.screen.cursor.x = 0
+            self.screen.insert_characters(shift)
+            self.cell_ranges = [(s + shift, e + shift) for (s, e) in self.cell_ranges]
 
     def destroy(self) -> None:
         self.screen.reset_callbacks()
         del self.screen
 
     def tab_at(self, x: int) -> Optional[int]:
-        x = (x - self.window_geometry.left) // self.cell_width
-        for i, (a, b) in enumerate(self.cell_ranges):
-            if a <= x <= b:
-                return i
+        if self.laid_out_once:
+            x = (x - self.window_geometry.left) // self.cell_width
+            for i, (a, b) in enumerate(self.cell_ranges):
+                if a <= x <= b:
+                    return i
+        return None
