@@ -1,18 +1,17 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 from typing import Tuple
 
 
 BossType = ChildType = TabType = WindowType = ScreenType = None
-BadLineType = SequenceMap = KeyActionType = None
+BadLineType = SequenceMap = KeyActionType = AliasMap = None
 AddressFamily = PopenType = Socket = StartupCtx = None
 SessionTab = SessionType = LayoutType = SpecialWindowInstance = None
 MarkType = RemoteCommandType = CoreTextFont = FontConfigPattern = None
 KeyEventType = ImageManagerType = KittyCommonOpts = HandlerType = None
 GRT_t = GRT_a = GRT_d = GRT_f = GRT_m = GRT_o = GRT_C = None
-ScreenSize = KittensKeyActionType = MouseEvent = AbstractEventLoop = None
+ScreenSize = KittensKeyActionType = MouseEvent = MouseButton = AbstractEventLoop = None
 TermManagerType = LoopType = Debug = GraphicsCommandType = None
 
 CompletedProcess = Tuple
@@ -21,5 +20,4 @@ EdgeLiteral = str
 PowerlineStyle = str
 MatchType = str
 Protocol = object
-MouseEvent = dict
 OptionsProtocol = object
