@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import atexit
@@ -15,26 +14,31 @@ from functools import lru_cache
 from time import monotonic
 from typing import (
     TYPE_CHECKING, Any, Callable, Dict, Generator, Iterable, List, Mapping,
-    Match, NamedTuple, Optional, Tuple, Union, cast
+    Match, NamedTuple, Optional, Pattern, Tuple, Union, cast
 )
 
 from .constants import (
-    appname, is_macos, is_wayland, read_kitty_resource, shell_path,
+    appname, config_dir, is_macos, is_wayland, read_kitty_resource, shell_path,
     supports_primary_selection
 )
-from .rgb import Color, to_color
+from .rgb import to_color
 from .types import run_once
 from .typing import AddressFamily, PopenType, Socket, StartupCtx
+from .fast_data_types import Color
 
 if TYPE_CHECKING:
+    from .fast_data_types import OSWindowSize
     from .options.types import Options
 else:
     Options = object
 
 
 def expandvars(val: str, env: Mapping[str, str] = {}, fallback_to_os_env: bool = True) -> str:
+    '''
+    Expand $VAR and ${VAR} Use $$ for a literal $
+    '''
 
-    def sub(m: Match) -> str:
+    def sub(m: 'Match[str]') -> str:
         key = m.group(1) or m.group(2)
         result = env.get(key)
         if result is None and fallback_to_os_env:
@@ -46,7 +50,12 @@ def expandvars(val: str, env: Mapping[str, str] = {}, fallback_to_os_env: bool =
     if '$' not in val:
         return val
 
-    return re.sub(r'\$(?:(\w+)|\{([^}]+)\})', sub, val)
+    return re.sub(r'\$(?:(\w+)|\{([^}]+)\})', sub, val.replace('$$', '\0')).replace('\0', '$')
+
+
+@run_once
+def sgr_sanitizer_pat() -> 're.Pattern[str]':
+    return re.compile('\033\\[.*?m')
 
 
 def platform_window_id(os_window_id: int) -> Optional[int]:
@@ -58,15 +67,17 @@ def platform_window_id(os_window_id: int) -> Optional[int]:
         from .fast_data_types import x11_window_id
         with suppress(Exception):
             return x11_window_id(os_window_id)
+    return None
 
 
-def load_shaders(name: str) -> Tuple[str, str]:
+def load_shaders(name: str, vertex_name: str = '', fragment_name: str = '') -> Tuple[str, str]:
     from .fast_data_types import GLSL_VERSION
 
-    def load(which: str) -> str:
-        return read_kitty_resource(f'{name}_{which}.glsl').decode('utf-8').replace('GLSL_VERSION', str(GLSL_VERSION), 1)
+    def load(which: str, lname: str = '') -> str:
+        lname = lname or name
+        return read_kitty_resource(f'{lname}_{which}.glsl').decode('utf-8').replace('GLSL_VERSION', str(GLSL_VERSION), 1)
 
-    return load('vertex'), load('fragment')
+    return load('vertex', vertex_name), load('fragment', fragment_name)
 
 
 def safe_print(*a: Any, **k: Any) -> None:
@@ -90,8 +101,8 @@ def sanitize_title(x: str) -> str:
     return re.sub(r'\s+', ' ', re.sub(r'[\0-\x19\x80-\x9f]', '', x))
 
 
-def color_as_int(val: Tuple[int, int, int]) -> int:
-    return val[0] << 16 | val[1] << 8 | val[2]
+def color_as_int(val: Color) -> int:
+    return int(val) & 0xffffff
 
 
 def color_from_int(val: int) -> Color:
@@ -113,8 +124,7 @@ def parse_color_set(raw: str) -> Generator[Tuple[int, Optional[int]], None, None
             else:
                 q = to_color(spec)
                 if q is not None:
-                    r, g, b = q
-                    yield c, r << 16 | g << 8 | b
+                    yield c, int(q) & 0xffffff
         except Exception:
             continue
 
@@ -211,7 +221,7 @@ def command_for_open(program: Union[str, List[str]] = 'default') -> List[str]:
     return cmd
 
 
-def open_cmd(cmd: Union[Iterable[str], List[str]], arg: Union[None, Iterable[str], str] = None, cwd: Optional[str] = None) -> PopenType:
+def open_cmd(cmd: Union[Iterable[str], List[str]], arg: Union[None, Iterable[str], str] = None, cwd: Optional[str] = None) -> 'PopenType[bytes]':
     import subprocess
     if arg is not None:
         cmd = list(cmd)
@@ -222,7 +232,7 @@ def open_cmd(cmd: Union[Iterable[str], List[str]], arg: Union[None, Iterable[str
     return subprocess.Popen(tuple(cmd), stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd or None)
 
 
-def open_url(url: str, program: Union[str, List[str]] = 'default', cwd: Optional[str] = None) -> PopenType:
+def open_url(url: str, program: Union[str, List[str]] = 'default', cwd: Optional[str] = None) -> 'PopenType[bytes]':
     return open_cmd(command_for_open(program), url, cwd=cwd)
 
 
@@ -280,6 +290,7 @@ def init_startup_notification(window_handle: Optional[int], startup_id: Optional
     except Exception:
         import traceback
         traceback.print_exc()
+    return None
 
 
 def end_startup_notification(ctx: Optional['StartupCtx']) -> None:
@@ -296,7 +307,7 @@ def end_startup_notification(ctx: Optional['StartupCtx']) -> None:
 
 class startup_notification_handler:
 
-    def __init__(self, do_notify: bool = True, startup_id: Optional[str] = None, extra_callback: Optional[Callable] = None):
+    def __init__(self, do_notify: bool = True, startup_id: Optional[str] = None, extra_callback: Optional[Callable[[int], None]] = None):
         self.do_notify = do_notify
         self.startup_id = startup_id
         self.extra_callback = extra_callback
@@ -376,9 +387,9 @@ class SingleInstance:
 
     def __call__(self, group_id: Optional[str] = None) -> bool:
         import socket
-        name = '{}-ipc-{}'.format(appname, os.geteuid())
+        name = f'{appname}-ipc-{os.geteuid()}'
         if group_id:
-            name += '-{}'.format(group_id)
+            name += f'-{group_id}'
 
         s = socket.socket(family=socket.AF_UNIX)
         # First try with abstract UDS
@@ -420,15 +431,20 @@ def parse_address_spec(spec: str) -> Tuple[AddressFamily, Union[Tuple[str, int],
         host, port = rest.rsplit(':', 1)
         address = host, int(port)
     else:
-        raise ValueError('Unknown protocol in --listen-on value: {}'.format(spec))
+        raise ValueError(f'Unknown protocol in --listen-on value: {spec}')
     return family, address, socket_path
 
 
-def write_all(fd: int, data: Union[str, bytes]) -> None:
+def write_all(fd: int, data: Union[str, bytes], block_until_written: bool = True) -> None:
     if isinstance(data, str):
         data = data.encode('utf-8')
     while data:
-        n = os.write(fd, data)
+        try:
+            n = os.write(fd, data)
+        except BlockingIOError:
+            if not block_until_written:
+                raise
+            continue
         if not n:
             break
         data = data[n:]
@@ -436,15 +452,28 @@ def write_all(fd: int, data: Union[str, bytes]) -> None:
 
 class TTYIO:
 
+    def __init__(self, read_with_timeout: bool = True):
+        self.read_with_timeout = read_with_timeout
+
     def __enter__(self) -> 'TTYIO':
         from .fast_data_types import open_tty
-        self.tty_fd, self.original_termios = open_tty(True)
+        self.tty_fd, self.original_termios = open_tty(self.read_with_timeout)
         return self
 
     def __exit__(self, *a: Any) -> None:
         from .fast_data_types import close_tty
         close_tty(self.tty_fd, self.original_termios)
 
+    def wait_till_read_available(self) -> bool:
+        if self.read_with_timeout:
+            raise ValueError('Cannot wait when TTY is set to read with timeout')
+        import select
+        rd = select.select([self.tty_fd], [], [])[0]
+        return bool(rd)
+
+    def read(self, limit: int) -> bytes:
+        return os.read(self.tty_fd, limit)
+
     def send(self, data: Union[str, bytes, Iterable[Union[str, bytes]]]) -> None:
         if isinstance(data, (str, bytes)):
             write_all(self.tty_fd, data)
@@ -488,7 +517,7 @@ def resolve_editor_cmd(editor: str, shell_env: Mapping[str, str]) -> Optional[st
         return ' '.join(map(shlex.quote, editor_cmd))
 
     if shell_env is os.environ:
-        q = find_exe(editor_exe)
+        q = which(editor_exe, only_system=True)
         if q:
             return patched(q)
     elif 'PATH' in shell_env:
@@ -496,6 +525,7 @@ def resolve_editor_cmd(editor: str, shell_env: Mapping[str, str]) -> Optional[st
         q = shutil.which(editor_exe, path=shell_env['PATH'])
         if q:
             return patched(q)
+    return None
 
 
 def get_editor_from_env(env: Mapping[str, str]) -> Optional[str]:
@@ -505,11 +535,11 @@ def get_editor_from_env(env: Mapping[str, str]) -> Optional[str]:
             editor = resolve_editor_cmd(editor, env)
             if editor:
                 return editor
+    return None
 
 
 def get_editor_from_env_vars(opts: Optional[Options] = None) -> List[str]:
     import shlex
-    import shutil
 
     editor = get_editor_from_env(os.environ)
     if not editor:
@@ -517,7 +547,7 @@ def get_editor_from_env_vars(opts: Optional[Options] = None) -> List[str]:
         editor = get_editor_from_env(shell_env)
 
     for ans in (editor, 'vim', 'nvim', 'vi', 'emacs', 'kak', 'micro', 'nano', 'vis'):
-        if ans and shutil.which(shlex.split(ans)[0]):
+        if ans and which(shlex.split(ans)[0], only_system=True):
             break
     else:
         ans = 'vim'
@@ -527,7 +557,12 @@ def get_editor_from_env_vars(opts: Optional[Options] = None) -> List[str]:
 def get_editor(opts: Optional[Options] = None) -> List[str]:
     if opts is None:
         from .fast_data_types import get_options
-        opts = get_options()
+        try:
+            opts = get_options()
+        except RuntimeError:
+            # we are in a kitten
+            from .cli import create_default_opts
+            opts = create_default_opts()
     if opts.editor == '.':
         return get_editor_from_env_vars()
     import shlex
@@ -552,6 +587,22 @@ def is_path_in_temp_dir(path: str) -> bool:
     return False
 
 
+def resolve_abs_or_config_path(path: str, env: Optional[Mapping[str, str]] = None, conf_dir: Optional[str] = None) -> str:
+    path = os.path.expanduser(path)
+    path = expandvars(path, env or {})
+    if not os.path.isabs(path):
+        path = os.path.join(conf_dir or config_dir, path)
+    return path
+
+
+def resolve_custom_file(path: str) -> str:
+    from .fast_data_types import get_options
+    opts: Optional[Options] = None
+    with suppress(RuntimeError):
+        opts = get_options()
+    return resolve_abs_or_config_path(path, opts.env if opts else {})
+
+
 def func_name(f: Any) -> str:
     if hasattr(f, '__name__'):
         return str(f.__name__)
@@ -571,7 +622,7 @@ def resolved_shell(opts: Optional[Options] = None) -> List[str]:
 
 
 @run_once
-def system_paths_on_macos() -> List[str]:
+def system_paths_on_macos() -> Tuple[str, ...]:
     entries, seen = [], set()
 
     def add_from_file(x: str) -> None:
@@ -593,23 +644,64 @@ def system_paths_on_macos() -> List[str]:
     for name in sorted(files):
         add_from_file(os.path.join('/etc/paths.d', name))
     add_from_file('/etc/paths')
-    return entries
+    return tuple(entries)
 
 
-@lru_cache(maxsize=32)
-def find_exe(name: str) -> Optional[str]:
+def which(name: str, only_system: bool = False) -> Optional[str]:
+    if os.sep in name:
+        return name
     import shutil
-    ans = shutil.which(name)
-    if ans is None:
-        # In case PATH is messed up
-        if is_macos:
-            paths = system_paths_on_macos()
-        else:
-            paths = ['/usr/local/bin', '/opt/bin', '/usr/bin', '/bin', '/usr/sbin', '/sbin']
-        paths.insert(0, os.path.expanduser('~/.local/bin'))
-        path = os.pathsep.join(paths) + os.pathsep + os.defpath
-        ans = shutil.which(name, path=path)
-    return ans
+    from .fast_data_types import get_options
+    opts: Optional[Options] = None
+    with suppress(RuntimeError):
+        opts = get_options()
+
+    tried_paths = set()
+    paths = []
+    append_paths = []
+    if opts and opts.exe_search_path:
+        for x in opts.exe_search_path:
+            x = x.strip()
+            if x:
+                if x[0] == '-':
+                    tried_paths.add(x[1:])
+                elif x[0] == '+':
+                    append_paths.append(x[1:])
+                else:
+                    paths.append(x)
+    ep = os.environ.get('PATH')
+    if ep:
+        paths.extend(ep.split(os.pathsep))
+    paths.append(os.path.expanduser('~/.local/bin'))
+    paths.append(os.path.expanduser('~/bin'))
+    paths.extend(append_paths)
+    ans = shutil.which(name, path=os.pathsep.join(x for x in paths if x not in tried_paths))
+    if ans:
+        return ans
+    # In case PATH is messed up try a default set of paths
+    if is_macos:
+        system_paths = system_paths_on_macos()
+    else:
+        system_paths = ('/usr/local/bin', '/opt/bin', '/usr/bin', '/bin', '/usr/sbin', '/sbin')
+    tried_paths |= set(paths)
+    system_paths = tuple(x for x in system_paths if x not in tried_paths)
+    if system_paths:
+        ans = shutil.which(name, path=os.pathsep.join(system_paths))
+        if ans:
+            return ans
+        tried_paths |= set(system_paths)
+    if only_system or opts is None:
+        return None
+    shell_env = read_shell_environment(opts)
+    for xenv in (shell_env, opts.env):
+        q = xenv.get('PATH')
+        if q:
+            paths = [x for x in xenv['PATH'].split(os.pathsep) if x not in tried_paths]
+            ans = shutil.which(name, path=os.pathsep.join(paths))
+            if ans:
+                return ans
+            tried_paths |= set(paths)
+    return None
 
 
 def read_shell_environment(opts: Optional[Options] = None) -> Dict[str, str]:
@@ -622,20 +714,23 @@ def read_shell_environment(opts: Optional[Options] = None) -> Dict[str, str]:
         shell = resolved_shell(opts)
         master, slave = openpty()
         remove_blocking(master)
+        if '-l' not in shell and '--login' not in shell:
+            shell += ['-l']
+        if '-i' not in shell and '--interactive' not in shell:
+            shell += ['-i']
         try:
-            p = subprocess.Popen(shell + ['-l', '-c', 'env'], stdout=slave, stdin=slave, stderr=slave, start_new_session=True, close_fds=True)
+            p = subprocess.Popen(shell + ['-c', 'env'], stdout=slave, stdin=slave, stderr=slave, start_new_session=True, close_fds=True)
         except FileNotFoundError:
             log_error('Could not find shell to read environment')
             return ans
         with os.fdopen(master, 'rb') as stdout, os.fdopen(slave, 'wb'):
             raw = b''
-            from subprocess import TimeoutExpired
             from time import monotonic
             start_time = monotonic()
             while monotonic() - start_time < 1.5:
                 try:
                     ret: Optional[int] = p.wait(0.01)
-                except TimeoutExpired:
+                except subprocess.TimeoutExpired:
                     ret = None
                 with suppress(Exception):
                     raw += stdout.read()
@@ -692,3 +787,68 @@ class SSHConnectionData(NamedTuple):
     binary: str
     hostname: str
     port: Optional[int] = None
+    identity_file: str = ''
+
+
+def get_new_os_window_size(
+    metrics: 'OSWindowSize', width: int, height: int, unit: str, incremental: bool = False, has_window_scaling: bool = True
+) -> Tuple[int, int]:
+    if unit == 'cells':
+        cw = metrics['cell_width']
+        ch = metrics['cell_height']
+        width *= cw
+        height *= ch
+        if has_window_scaling:
+            width = round(width / metrics['xscale'])
+            height = round(height / metrics['yscale'])
+    if incremental:
+        w = metrics['width'] + width
+        h = metrics['height'] + height
+    else:
+        w = width or metrics['width']
+        h = height or metrics['height']
+    return w, h
+
+
+def get_all_processes() -> Iterable[int]:
+    if is_macos:
+        from kitty.fast_data_types import get_all_processes as f
+        yield from f()
+    else:
+        for c in os.listdir('/proc'):
+            if c.isdigit():
+                yield int(c)
+
+
+def is_kitty_gui_cmdline(*cmd: str) -> bool:
+    if not cmd:
+        return False
+    if os.path.basename(cmd[0]) != 'kitty':
+        return False
+    if len(cmd) == 1:
+        return True
+    if '+' in cmd or '@' in cmd or cmd[1].startswith('+') or cmd[1].startswith('@'):
+        return False
+    return True
+
+
+def reload_conf_in_all_kitties() -> None:
+    import signal
+    from kitty.child import cmdline_of_process
+
+    for pid in get_all_processes():
+        try:
+            cmd = cmdline_of_process(pid)
+        except Exception:
+            continue
+        if cmd and is_kitty_gui_cmdline(*cmd):
+            os.kill(pid, signal.SIGUSR1)
+
+
+@run_once
+def control_codes_pat() -> 'Pattern[str]':
+    return re.compile('[\x00-\x09\x0b-\x1f\x7f\x80-\x9f]')
+
+
+def sanitize_control_codes(text: str, replace_with: str = '') -> str:
+    return control_codes_pat().sub(replace_with, text)
