@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -13,7 +12,6 @@ from io import BytesIO
 from itertools import cycle
 from typing import NamedTuple
 
-from kitty.constants import cache_dir
 from kitty.fast_data_types import (
     load_png_data, parse_bytes, shm_unlink, shm_write, xor_data
 )
@@ -95,7 +93,7 @@ def load_helpers(self):
 
     def pl(payload, **kw):
         kw.setdefault('i', 1)
-        cmd = ','.join('%s=%s' % (k, v) for k, v in kw.items())
+        cmd = ','.join(f'{k}={v}' for k, v in kw.items())
         res = send_command(s, cmd, payload)
         return parse_response(res)
 
@@ -176,7 +174,7 @@ def make_send_command(screen):
         if i:
             kw['i'] = i
         kw['a'] = a
-        cmd = ','.join('%s=%s' % (k, v) for k, v in kw.items())
+        cmd = ','.join(f'{k}={v}' for k, v in kw.items())
         res = send_command(screen, cmd, payload)
         return parse_full_response(res)
     return li
@@ -184,13 +182,6 @@ def make_send_command(screen):
 
 class TestGraphics(BaseTest):
 
-    def setUp(self):
-        cache_dir.set_override(tempfile.mkdtemp())
-
-    def tearDown(self):
-        os.rmdir(cache_dir())
-        cache_dir.clear_override()
-
     def test_xor_data(self):
 
         def xor(skey, data):
@@ -435,7 +426,7 @@ class TestGraphics(BaseTest):
         self.assertEqual(g.disk_cache.total_size, 0)
 
         def li(payload, **kw):
-            cmd = ','.join('%s=%s' % (k, v) for k, v in kw.items())
+            cmd = ','.join(f'{k}={v}' for k, v in kw.items())
             res = send_command(s, cmd, payload)
             return parse_response_with_ids(res)
 
@@ -470,7 +461,7 @@ class TestGraphics(BaseTest):
 
         # test put with number
         def put(**kw):
-            cmd = ','.join('%s=%s' % (k, v) for k, v in kw.items())
+            cmd = ','.join(f'{k}={v}' for k, v in kw.items())
             cmd = 'a=p,' + cmd
             return parse_response_with_ids(send_command(s, cmd))
 
@@ -483,9 +474,9 @@ class TestGraphics(BaseTest):
         def delete(ac='N', **kw):
             cmd = 'a=d'
             if ac:
-                cmd += ',d={}'.format(ac)
+                cmd += f',d={ac}'
             if kw:
-                cmd += ',' + ','.join('{}={}'.format(k, v) for k, v in kw.items())
+                cmd += ',' + ','.join(f'{k}={v}' for k, v in kw.items())
             send_command(s, cmd)
 
         count = s.grman.image_count
@@ -599,9 +590,9 @@ class TestGraphics(BaseTest):
         def delete(ac=None, **kw):
             cmd = 'a=d'
             if ac:
-                cmd += ',d={}'.format(ac)
+                cmd += f',d={ac}'
             if kw:
-                cmd += ',' + ','.join('{}={}'.format(k, v) for k, v in kw.items())
+                cmd += ',' + ','.join(f'{k}={v}' for k, v in kw.items())
             send_command(s, cmd)
 
         put_image(s, cw, ch)
