@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 from kitty.fast_data_types import (
@@ -111,12 +110,29 @@ class TestScreen(BaseTest):
         self.ae(str(s.line(0)), q)
         self.ae(s.cursor.x, 2)
 
+    def test_regional_indicators(self):
+        s = self.create_screen()
+        flag = '\U0001f1ee\U0001f1f3'
+        s.draw(flag)
+        self.ae(str(s.line(0)), flag)
+        self.ae(s.cursor.x, 2)
+        s = self.create_screen()
+        s.draw('a'), s.draw(flag[0]), s.draw('b')
+        self.ae(str(s.line(0)), 'a' + flag[0] + 'b')
+        self.ae(s.cursor.x, 4)
+
     def test_zwj(self):
         s = self.create_screen(cols=20)
         q = '\U0001f468\u200d\U0001f469\u200d\U0001f467\u200d\U0001f466'
         s.draw(q)
         self.ae(q, str(s.line(0)))
         self.ae(s.cursor.x, 8)
+        for x in '\u200b\u200c\u200d':
+            s = self.create_screen()
+            q = f'X{x}Y'
+            s.draw(q)
+            self.ae(q, str(s.line(0)))
+            self.ae(s.cursor.x, 2)
 
     def test_char_manipulation(self):
         s = self.create_screen()
@@ -513,6 +529,15 @@ class TestScreen(BaseTest):
         self.ae(s.text_for_selection(), expected)
         s.scroll(2, True)
         self.ae(s.text_for_selection(), expected)
+        s.reset()
+
+    def test_soft_hyphen(self):
+        s = self.create_screen()
+        s.draw('a\u00adb')
+        self.ae(s.cursor.x, 2)
+        s.start_selection(0, 0)
+        s.update_selection(2, 0)
+        self.ae(s.text_for_selection(), ('a\u00adb',))
 
     def test_variation_selectors(self):
         s = self.create_screen()
@@ -794,6 +819,29 @@ class TestScreen(BaseTest):
         self.ae(str(s.linebuf), '0\n5\n6\n7\n\n')
         self.ae(str(s.historybuf), '')
 
+    def test_osc_52(self):
+        s = self.create_screen()
+        c = s.callbacks
+
+        def send(what: str):
+            return parse_bytes(s, f'\033]52;p;{what}\a'.encode('ascii'))
+
+        def t(q, use_pending_mode, *expected):
+            c.clear()
+            if use_pending_mode:
+                parse_bytes(s, b'\033[?2026h')
+            send(q)
+            if use_pending_mode:
+                self.ae(c.cc_buf, [])
+                parse_bytes(s, b'\033[?2026l')
+            self.ae(c.cc_buf, list(expected))
+
+        for use_pending_mode in (False, True):
+            t('XYZ', use_pending_mode, ('p;XYZ', False))
+            t('a' * 8192, use_pending_mode, ('p;' + 'a' * (8192 - 6), True), (';' + 'a' * 6, False))
+            t('', use_pending_mode, ('p;', False))
+            t('!', use_pending_mode, ('p;!', False))
+
     def test_key_encoding_flags_stack(self):
         s = self.create_screen()
         c = s.callbacks
@@ -805,7 +853,7 @@ class TestScreen(BaseTest):
             return parse_bytes(s, f'\033[{code}{p}u'.encode('ascii'))
 
         def ac(flags):
-            parse_bytes(s, '\033[?u'.encode('ascii'))
+            parse_bytes(s, b'\033[?u')
             self.ae(c.wtcbuf, f'\033[?{flags}u'.encode('ascii'))
             c.clear()
 
@@ -889,3 +937,64 @@ class TestScreen(BaseTest):
         for trailer in '{([':
             t('http://moo.com', after=trailer)
         t('http://moo.com', x=s.columns - 9)
+
+    def test_prompt_marking(self):
+        s = self.create_screen()
+
+        def mark_prompt():
+            parse_bytes(s, b'\033]133;A\007')
+
+        def mark_output():
+            parse_bytes(s, b'\033]133;C\007')
+
+        for i in range(4):
+            mark_prompt()
+            s.draw(f'$ {i}')
+            s.carriage_return()
+            s.index(), s.index()
+        self.ae(s.scrolled_by, 0)
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 1')
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 0')
+        self.assertFalse(s.scroll_to_prompt())
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), '$ 1')
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), '$ 2')
+        self.assertFalse(s.scroll_to_prompt(1))
+
+        s = self.create_screen()
+        mark_prompt(), s.draw('$ 0')
+        s.carriage_return(), s.index()
+        mark_prompt(), s.draw('$ 1')
+        for i in range(s.lines):
+            s.carriage_return(), s.index()
+            s.draw(str(i))
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 1')
+
+        def fco():
+            a = []
+            s.cmd_output(1, a.append)
+            return ''.join(a)
+
+        def lco():
+            a = []
+            s.cmd_output(0, a.append)
+            return ''.join(a)
+
+        s = self.create_screen()
+        s.draw('abcd'), s.index(), s.carriage_return()
+        s.draw('12'), s.index(), s.carriage_return()
+        self.ae(fco(), '')
+        self.ae(lco(), 'abcd\n12')
+        s = self.create_screen()
+        mark_prompt(), s.draw('$ 0')
+        s.carriage_return(), s.index()
+        mark_output()
+        s.draw('abcd'), s.index(), s.carriage_return()
+        s.draw('12'), s.index(), s.carriage_return()
+        mark_prompt(), s.draw('$ 1')
+        self.ae(fco(), 'abcd\n12')
+        self.ae(lco(), 'abcd\n12')
