@@ -1,8 +1,8 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import argparse
+import datetime
 import io
 import json
 import mimetypes
@@ -15,8 +15,8 @@ import subprocess
 import sys
 import tempfile
 import time
-from contextlib import suppress
-from typing import IO, Any, Dict, Iterable, List, Optional, cast
+from contextlib import contextmanager, suppress
+from typing import IO, Any, Dict, Generator, Iterable, Optional, cast
 
 import requests
 
@@ -27,35 +27,55 @@ with open('kitty/constants.py') as f:
     raw = f.read()
 nv = re.search(r'^version: Version\s+=\s+Version\((\d+), (\d+), (\d+)\)', raw, flags=re.MULTILINE)
 if nv is not None:
-    version = '%s.%s.%s' % (nv.group(1), nv.group(2), nv.group(3))
+    version = f'{nv.group(1)}.{nv.group(2)}.{nv.group(3)}'
 ap = re.search(r"^appname: str\s+=\s+'([^']+)'", raw, flags=re.MULTILINE)
 if ap is not None:
     appname = ap.group(1)
 
 ALL_ACTIONS = 'man html build tag sdist upload website'.split()
+NIGHTLY_ACTIONS = 'man html build sdist upload_nightly'.split()
 
 
-def call(*cmd: str, cwd: Optional[str] = None) -> None:
+def echo_cmd(cmd: Iterable[str]) -> None:
+    isatty = sys.stdout.isatty()
+    end = '\n'
+    if isatty:
+        end = '\x1b[m' + end
+        print('\x1b[92m', end='')
+    print(shlex.join(cmd), end=end, flush=True)
+
+
+def call(*cmd: str, cwd: Optional[str] = None, echo: bool = False) -> None:
     if len(cmd) == 1:
         q = shlex.split(cmd[0])
     else:
         q = list(cmd)
+    if echo:
+        echo_cmd(cmd)
     ret = subprocess.Popen(q, cwd=cwd).wait()
     if ret != 0:
         raise SystemExit(ret)
 
 
 def run_build(args: Any) -> None:
-    call('python ../bypy linux program')
-    call('python ../bypy linux 32 program')
-    call('python ../bypy macos program --sign-installers --notarize')
+    for x in ('64', '32', 'arm64'):
+        try:
+            call(f'python ../bypy linux --arch {x} program', echo=True)
+        except (SystemExit, Exception):
+            if x != 'arm64':
+                raise
+            print('Linux ARM build failed, retrying in a few seconds...', file=sys.stderr)
+            time.sleep(15)
+            call(f'python ../bypy linux --arch {x} program', echo=True)
+        call(f'python ../bypy linux --arch {x} shutdown', echo=True)
+    call('python ../bypy macos program --sign-installers --notarize', echo=True)
     call('python ../bypy macos shutdown')
 
 
 def run_tag(args: Any) -> None:
     call('git push')
     call('git tag -s v{0} -m version-{0}'.format(version))
-    call('git push origin v{0}'.format(version))
+    call(f'git push origin v{version}')
 
 
 def run_man(args: Any) -> None:
@@ -67,16 +87,9 @@ def run_html(args: Any) -> None:
     add_old_redirects('docs/_build/dirhtml')
 
 
-def add_old_redirects(loc: str) -> None:
-    for dirpath, dirnames, filenames in os.walk(loc):
-        if dirpath != loc:
-            for fname in filenames:
-                if fname == 'index.html':
-                    bname = os.path.basename(dirpath)
-                    base = os.path.dirname(dirpath)
-                    link_name = os.path.join(base, f'{bname}.html') if base else f'{bname}.html'
-                    with open(link_name, 'w') as f:
-                        f.write(f'''
+def generate_redirect_html(link_name: str, bname: str) -> None:
+    with open(link_name, 'w') as f:
+        f.write(f'''
 <html>
 <head>
 <title>Redirecting...</title>
@@ -95,6 +108,22 @@ window.location.replace('./{bname}/' + window.location.hash);
 ''')
 
 
+def add_old_redirects(loc: str) -> None:
+    for dirpath, dirnames, filenames in os.walk(loc):
+        if dirpath != loc:
+            for fname in filenames:
+                if fname == 'index.html':
+                    bname = os.path.basename(dirpath)
+                    base = os.path.dirname(dirpath)
+                    link_name = os.path.join(base, f'{bname}.html') if base else f'{bname}.html'
+                    generate_redirect_html(link_name, bname)
+
+    old_unicode_input_path = os.path.join(loc, 'kittens', 'unicode-input')
+    os.makedirs(old_unicode_input_path, exist_ok=True)
+    generate_redirect_html(os.path.join(old_unicode_input_path, 'index.html'), '../unicode_input')
+    generate_redirect_html(f'{old_unicode_input_path}.html', 'unicode_input')
+
+
 def run_docs(args: Any) -> None:
     subprocess.check_call(['make', 'docs'])
 
@@ -117,6 +146,9 @@ def run_website(args: Any) -> None:
 
 
 def sign_file(path: str) -> None:
+    dest = path + '.sig'
+    with suppress(FileNotFoundError):
+        os.remove(dest)
     subprocess.check_call([
         os.environ['PENV'] + '/gpg-as-kovid', '--output', path + '.sig',
         '--detach-sig', path
@@ -161,7 +193,6 @@ class ReadFileWithProgressReporting(io.FileIO):  # {{{
         def write(*args: str) -> None:
             print(*args, end='')
 
-        write('\x1b[s\x1b[K')
         frac = float(self.tell()) / self._total
         mb_pos = self.tell() / float(1024**2)
         mb_tot = self._total / float(1024**2)
@@ -170,21 +201,19 @@ class ReadFileWithProgressReporting(io.FileIO):  # {{{
         bit_rate = kb_rate * 1024
         eta = int((self._total - self.tell()) / bit_rate) + 1
         eta_m, eta_s = eta / 60, eta % 60
-        write(
-            '  %.1f%%   %.1f/%.1fMB %.1f KB/sec    %d minutes, %d seconds left'
-            % (frac * 100, mb_pos, mb_tot, kb_rate, eta_m, eta_s))
-        write('\x1b[u')
+        if sys.stdout.isatty():
+            write(
+                f'\r\033[K\033[?7h {frac:%} {mb_pos:.1f}/{mb_tot:.1f}MB {kb_rate:.1f} KB/sec {eta_m} minutes, {eta_s} seconds left\033[?7l')
         if self.tell() >= self._total:
-            t = int(time.time() - self.start_time) + 1
-            print('\nUpload took %d minutes and %d seconds at %.1f KB/sec' %
-                  (t / 60, t % 60, kb_rate))
+            t = int(time.monotonic() - self.start_time) + 1
+            print(f'\nUpload took {t//60} minutes and {t%60} seconds at {kb_rate:.1f} KB/sec')
         sys.stdout.flush()
 
 
 # }}}
 
 
-class Base(object):  # {{{
+class Base:  # {{{
 
     def info(self, *args: Any, **kwargs: Any) -> None:
         print(*args, **kwargs)
@@ -215,48 +244,61 @@ class GitHub(Base):  # {{{
     ):
         self.files, self.reponame, self.version, self.username, self.password, self.replace = (
             files, reponame, version, username, password, replace)
-        self.current_tag_name = 'v' + self.version
+        self.current_tag_name = self.version if self.version == 'nightly' else ('v' + self.version)
+        self.is_nightly = self.current_tag_name == 'nightly'
         self.requests = s = requests.Session()
         s.auth = (self.username, self.password)
         s.headers.update({'Accept': 'application/vnd.github.v3+json'})
+        self.url_base = self.API + f'repos/{self.username}/{self.reponame}/releases/'
+
+    def patch(self, url: str, fail_msg: str, **data: Any) -> None:
+        rdata = json.dumps(data)
+        try:
+            r = self.requests.patch(url, data=rdata)
+        except Exception:
+            time.sleep(15)
+            r = self.requests.patch(url, data=rdata)
+        if r.status_code != 200:
+            self.fail(r, fail_msg)
+
+    def update_nightly_description(self, release_id: int) -> None:
+        url = self.url_base + str(release_id)
+        now = str(datetime.datetime.utcnow()).split('.')[0] + ' UTC'
+        with open('.git/refs/heads/master') as f:
+            commit = f.read().strip()
+        self.patch(
+            url, 'Failed to update nightly release description',
+            body=f'Nightly release, generated on: {now} from commit: {commit}.'
+            ' For how to install nightly builds, see: https://sw.kovidgoyal.net/kitty/binary/#customizing-the-installation'
+        )
 
     def __call__(self) -> None:
-        releases = self.releases()
         # self.clean_older_releases(releases)
-        release = self.create_release(releases)
+        release = self.create_release()
         upload_url = release['upload_url'].partition('{')[0]
+        asset_url = self.url_base + 'assets/{}'
         existing_assets = self.existing_assets(release['id'])
+        if self.is_nightly:
+            for fname in existing_assets:
+                self.info(f'Deleting {fname} from GitHub')
+                r = self.requests.delete(asset_url.format(existing_assets[fname]))
+                if r.status_code not in (204, 404):
+                    self.fail(r, 'Failed to delete %s from GitHub' % fname)
+            self.update_nightly_description(release['id'])
         for path, desc in self.files.items():
             self.info('')
-            url = self.API + 'repos/%s/%s/releases/assets/{}' % (self.username,
-                                                                 self.reponame)
             fname = os.path.basename(path)
+            if self.is_nightly:
+                fname = fname.replace(version, 'nightly')
             if fname in existing_assets:
-                self.info('Deleting %s from GitHub with id: %s' %
-                          (fname, existing_assets[fname]))
-                r = self.requests.delete(url.format(existing_assets[fname]))
-                if r.status_code != 204:
-                    self.fail(r, 'Failed to delete %s from GitHub' % fname)
+                self.info(f'Deleting {fname} from GitHub with id: {existing_assets[fname]}')
+                r = self.requests.delete(asset_url.format(existing_assets[fname]))
+                if r.status_code not in (204, 404):
+                    self.fail(r, f'Failed to delete {fname} from GitHub')
             r = self.do_upload(upload_url, path, desc, fname)
             if r.status_code != 201:
-                self.fail(r, 'Failed to upload file: %s' % fname)
-            try:
-                r = self.requests.patch(
-                    url.format(r.json()['id']),
-                    data=json.dumps({
-                        'name': fname,
-                        'label': desc
-                    }))
-            except Exception:
-                time.sleep(15)
-                r = self.requests.patch(
-                    url.format(r.json()['id']),
-                    data=json.dumps({
-                        'name': fname,
-                        'label': desc
-                    }))
-            if r.status_code != 200:
-                self.fail(r, 'Failed to set label for %s' % fname)
+                self.fail(r, f'Failed to upload file: {fname}')
+            self.patch(asset_url.format(r.json()['id']), f'Failed to set label for {fname}', name=fname, label=desc)
 
     def clean_older_releases(self, releases: Iterable[Dict[str, Any]]) -> None:
         for release in releases:
@@ -267,7 +309,7 @@ class GitHub(Base):  # {{{
                           release['tag_name'])
                 for asset in release['assets']:
                     r = self.requests.delete(
-                        self.API + 'repos/%s/%s/releases/assets/%s' % (
+                        self.API + 'repos/{}/{}/releases/assets/{}'.format(
                             self.username, self.reponame, asset['id']))
                     if r.status_code != 204:
                         self.fail(
@@ -277,7 +319,7 @@ class GitHub(Base):  # {{{
 
     def do_upload(self, url: str, path: str, desc: str, fname: str) -> requests.Response:
         mime_type = mimetypes.guess_type(fname)[0] or 'application/octet-stream'
-        self.info('Uploading to GitHub: %s (%s)' % (fname, mime_type))
+        self.info(f'Uploading to GitHub: {fname} ({mime_type})')
         with ReadFileWithProgressReporting(path) as f:
             return self.requests.post(
                 url,
@@ -299,28 +341,23 @@ class GitHub(Base):  # {{{
         return bool(error_code == 'already_exists')
 
     def existing_assets(self, release_id: str) -> Dict[str, str]:
-        url = self.API + 'repos/%s/%s/releases/%s/assets' % (
+        url = self.API + 'repos/{}/{}/releases/{}/assets'.format(
             self.username, self.reponame, release_id)
         r = self.requests.get(url)
         if r.status_code != 200:
             self.fail(r, 'Failed to get assets for release')
         return {asset['name']: asset['id'] for asset in r.json()}
 
-    def releases(self) -> List[Dict[str, Any]]:
-        url = self.API + 'repos/%s/%s/releases' % (self.username, self.reponame
-                                                   )
-        r = self.requests.get(url)
-        if r.status_code != 200:
-            self.fail(r, 'Failed to list releases')
-        return list(r.json())
-
-    def create_release(self, releases: Iterable[Dict[str, str]]) -> Dict[str, Any]:
+    def create_release(self) -> Dict[str, Any]:
         ' Create a release on GitHub or if it already exists, return the existing release '
-        for release in releases:
-            # Check for existing release
-            if release['tag_name'] == self.current_tag_name:
-                return release
-        url = self.API + 'repos/%s/%s/releases' % (self.username, self.reponame)
+        # Check for existing release
+        url = f'{self.API}repos/{self.username}/{self.reponame}/releases/tags/{self.current_tag_name}'
+        r = self.requests.get(url)
+        if r.status_code == 200:
+            return dict(r.json())
+        if self.is_nightly:
+            raise SystemExit('No existing nightly release found on GitHub')
+        url = self.API + f'repos/{self.username}/{self.reponame}/releases'
         r = self.requests.post(
             url,
             data=json.dumps({
@@ -346,13 +383,14 @@ def get_github_data() -> Dict[str, str]:
     return {'username': un, 'password': pw}
 
 
-def run_upload(args: Any) -> None:
+def files_for_upload() -> Dict[str, str]:
     files = {}
     signatures = {}
     for f, desc in {
         'macos/dist/kitty-{}.dmg': 'macOS dmg',
-        'linux/64/sw/dist/kitty-{}-x86_64.txz': 'Linux amd64 binary bundle',
-        'linux/32/sw/dist/kitty-{}-i686.txz': 'Linux x86 binary bundle',
+        'linux/64/dist/kitty-{}-x86_64.txz': 'Linux amd64 binary bundle',
+        'linux/32/dist/kitty-{}-i686.txz': 'Linux x86 binary bundle',
+        'linux/arm64/dist/kitty-{}-arm64.txz': 'Linux arm64 binary bundle',
     }.items():
         path = os.path.join('bypy', 'b', f.format(version))
         if not os.path.exists(path):
@@ -367,15 +405,57 @@ def run_upload(args: Any) -> None:
     for f in files:
         if not os.path.exists(f):
             raise SystemExit(f'The release artifact {f} does not exist')
+    return files
+
+
+def run_upload(args: Any) -> None:
     gd = get_github_data()
+    files = files_for_upload()
     gh = GitHub(files, appname, version, gd['username'], gd['password'])
     gh()
 
 
+def run_upload_nightly(args: Any) -> None:
+    subprocess.check_call(['git', 'tag', '-f', 'nightly'])
+    subprocess.check_call(['git', 'push', 'origin', 'nightly', '-f'])
+    gd = get_github_data()
+    files = files_for_upload()
+    gh = GitHub(files, appname, 'nightly', gd['username'], gd['password'])
+    gh()
+
+
+def current_branch() -> str:
+    return subprocess.check_output(['git', 'symbolic-ref', '--short', 'HEAD']).decode('utf-8').strip()
+
+
 def require_git_master(branch: str = 'master') -> None:
-    b = subprocess.check_output(['git', 'symbolic-ref', '--short', 'HEAD']).decode('utf-8').strip()
-    if b != branch:
-        raise SystemExit('You must be in the {} git branch'.format(branch))
+    if current_branch() != branch:
+        raise SystemExit(f'You must be in the {branch} git branch')
+
+
+def safe_read(path: str) -> str:
+    with suppress(FileNotFoundError):
+        with open(path) as f:
+            return f.read()
+    return ''
+
+
+@contextmanager
+def change_to_git_master() -> Generator[None, None, None]:
+    stash_ref_before = safe_read('.git/refs/stash')
+    subprocess.check_call(['git', 'stash'])
+    try:
+        branch_before = current_branch()
+        if branch_before != 'master':
+            subprocess.check_call(['git', 'switch', 'master'])
+        try:
+            yield
+        finally:
+            if branch_before != 'master':
+                subprocess.check_call(['git', 'switch', branch_before])
+    finally:
+        if stash_ref_before != safe_read('.git/refs/stash'):
+            subprocess.check_call(['git', 'stash', 'pop'])
 
 
 def require_penv() -> None:
@@ -383,38 +463,7 @@ def require_penv() -> None:
         raise SystemExit('The PENV env var is not present, required for uploading releases')
 
 
-def main() -> None:
-    require_git_master()
-    require_penv()
-    parser = argparse.ArgumentParser(description='Publish kitty')
-    parser.add_argument(
-        '--only',
-        default=False,
-        action='store_true',
-        help='Only run the specified action, by default the specified action and all sub-sequent actions are run')
-    parser.add_argument(
-        'action',
-        default='all',
-        nargs='?',
-        choices=list(ALL_ACTIONS) + ['all'],
-        help='The action to start with')
-    args = parser.parse_args()
-    if args.action == 'all':
-        actions = list(ALL_ACTIONS)
-    else:
-        idx = ALL_ACTIONS.index(args.action)
-        actions = ALL_ACTIONS[idx:]
-    if args.only:
-        del actions[1:]
-    else:
-        try:
-            ans = input('Publish version \033[91m{}\033[m (y/n): '.format(version))
-        except KeyboardInterrupt:
-            ans = 'n'
-        if ans.lower() != 'y':
-            return
-    if actions == ['website']:
-        actions.insert(0, 'html')
+def exec_actions(actions: Iterable[str], args: Any) -> None:
     for action in actions:
         print('Running', action)
         cwd = os.getcwd()
@@ -422,5 +471,51 @@ def main() -> None:
         os.chdir(cwd)
 
 
+def main() -> None:
+    parser = argparse.ArgumentParser(description='Publish kitty')
+    parser.add_argument(
+        '--only',
+        default=False,
+        action='store_true',
+        help='Only run the specified action, by default the specified action and all sub-sequent actions are run')
+    parser.add_argument(
+        '--nightly',
+        default=False,
+        action='store_true',
+        help='Upload a nightly release, ignores all other arguments')
+    parser.add_argument(
+        'action',
+        default='all',
+        nargs='?',
+        choices=list(ALL_ACTIONS) + ['all', 'upload_nightly'],
+        help='The action to start with')
+    args = parser.parse_args()
+    require_penv()
+    if args.nightly:
+        with change_to_git_master():
+            exec_actions(NIGHTLY_ACTIONS, args)
+        return
+    require_git_master()
+    if args.action == 'all':
+        actions = list(ALL_ACTIONS)
+    elif args.action == 'upload_nightly':
+        actions = ['upload_nightly']
+    else:
+        idx = ALL_ACTIONS.index(args.action)
+        actions = ALL_ACTIONS[idx:]
+    if args.only:
+        del actions[1:]
+    else:
+        try:
+            ans = input(f'Publish version \033[91m{version}\033[m (y/n): ')
+        except KeyboardInterrupt:
+            ans = 'n'
+        if ans.lower() != 'y':
+            return
+    if actions == ['website']:
+        actions.insert(0, 'html')
+    exec_actions(actions, args)
+
+
 if __name__ == '__main__':
     main()
